import { glob } from "glob"
import { HTMLElement, parse, TextNode } from "node-html-parser"
import fs from "fs-extra"
import { getFiles } from "../file_systems"
import { BasePlugin } from "../bugs"
import { masterUrl } from "../datasets"

export let remapJo: any = {
  RUANR5L15S2017IDS20667: "20172002",
  RUANR5L15S2018IDS20817: "20180086",
  RUANR5L15S2018IDS20864: "20180121",
  RUANR5L15S2018IDS20981: "20180174",
  RUANR5L15S2019IDS21507: "20190153",
  RUANR5L15S2019IDS21608: "20190218",
  RUANR5L15S2020IDS21840: "20200002",
  RUANR5L15S2020IDS21841: "20200003",
  RUANR5L15S2020IDS21859: "20200001",
  RUANR5L15S2020IDS21970: "20200090",
  RUANR5L15S2020IDS21983: "20200092",
  RUANR5L15S2020IDS22019: "20200117",
  RUANR5L15S2020IDS22020: "20200118",
  RUANR5L15S2020IDS22021: "20200119",
  RUANR5L15S2020IDS22054: "20200134",
  RUANR5L15S2020IDS22055: "20200135",
  RUANR5L15S2020IDS22056: "20200136",
  RUANR5L15S2020IDS22057: "20200137",
  RUANR5L15S2020IDS22087: "20200139",
  RUANR5L15S2020IDS22089: "20200140",
  RUANR5L15S2020IDS22106: "20200125",
}

export class Plugin extends BasePlugin {
  cr2filenames: any
  date2crs: any
  cr2date: any
  filename2date: any

  constructor(options: any) {
    super(options, "agenda-00002")
  }

  async preCheck(): Promise<void> {
    if (this.options.cr == undefined) throw "--cr is required"
    this.filename2date = {}
    this.cr2filenames = {}
    this.date2crs = {}
    this.cr2date = {}
    for (const cr of getFiles([`${this.options.cr}/*.asp`])) {
      this.cr2filenames[cr] = []

      const content: string = fs.readFileSync(cr, { encoding: "utf-8" })

      type ParseResult =
        | (TextNode & { valid?: boolean })
        | (HTMLElement & { valid?: boolean })
      const doc: ParseResult = parse(content)
      const root = doc as unknown as HTMLElement
      let found = false
      for (const meta of root.querySelectorAll("meta")) {
        if (meta.attributes["name"] == "QUANTIEME_SEANCE") {
          const date = meta.attributes["content"]
            .split(" ")
            .splice(-3)
            .join(" ")
          if (!(date in this.date2crs)) this.date2crs[date] = []
          this.date2crs[date].push(cr)
          this.cr2date[cr] = date
          found = true
          break
        }
      }
      if (!found) {
        console.error(`${cr} is missing the meta name="QUANTIEME_SEANCE"`)
      }
    }
  }

  static comptesRendusUrls(filename: string): string {
    const re = "^.*/comptes-rendus/(.*)"
    const match = new RegExp(re).exec(filename)
    if (match == null) return filename
    const file = match[1]
    return `[${file}](https://git.en-root.org/tricoteuses/html/comptes-rendus/-/tree/master/${file})`
  }

  async check(agenda: any, filename: any): Promise<any> {
    if (
      agenda.xsiType != "seance_type" ||
      agenda.lieu.code != "AN" ||
      agenda.cycleDeVie.etat == "Supprimé" ||
      agenda.cycleDeVie.etat == "Annulé"
    )
      return null

    if (this.options.verbose) console.log(`agenda-00002: ${filename}`)
    let status = "ok"
    let info: any = []

    const dateSeance = new Date(agenda.identifiants.dateSeance.substring(0, 10))
    const year = dateSeance.toLocaleDateString("fr", { year: "numeric" })
    const day = dateSeance.toLocaleDateString("fr", { day: "2-digit" })
    const month = dateSeance.toLocaleDateString("fr", { month: "long" })
    const dateSeanceString = `${day} ${month} ${year}`
    this.filename2date[filename] = dateSeanceString

    let idJo

    if (
      agenda.identifiants == undefined ||
      agenda.identifiants.idJo == undefined
    ) {
      const week = 7 * 24 * 60 * 60 * 1000
      if (dateSeance >= new Date(new Date().getTime() - week)) {
        return null
      } else {
        idJo = undefined
      }
    } else {
      idJo = agenda.identifiants.idJo
    }

    if (agenda.uid in remapJo) {
      if (idJo == remapJo[agenda.uid]) {
        info.push(`idJo already is ${idJo}`)
        status = "already-fixed"
      } else {
        idJo = remapJo[agenda.uid]
        status = "will-be-fixed"
      }
    }

    if (idJo == undefined) {
      return {
        status: "needs-fixing",
        info: ["idJo field is missing"],
      }
    }

    const pattern = `${this.options.cr}/*_${idJo}.asp`
    const expanded = glob.globSync(pattern)
    let cr
    if (expanded.length > 0) cr = expanded[0]
    else cr = undefined
    if (cr != undefined) {
      this.cr2filenames[cr].push(filename)

      if (dateSeanceString != this.cr2date[cr]) {
        info.push(
          `${Plugin.comptesRendusUrls(cr)} meta QUANTIEME_SEANCE contains ${
            this.cr2date[cr]
          } instead of the expected ${dateSeanceString}`,
        )
        status = "needs-fixing"
      }
    } else {
      if (status != "will-be-fixed") {
        info.push(`${pattern} does not exist`)
        status = "needs-fixing"
      }
    }
    return { status: status, info: info }
  }

  findCandidates(filename: any): any {
    const date = this.filename2date[filename]
    if (!(date in this.date2crs)) return []
    let candidates = []
    for (const cr of this.date2crs[date]) {
      if (this.cr2filenames[cr].length > 0)
        candidates.push(
          `${Plugin.comptesRendusUrls(
            cr,
          )} is not a candidate because it is referenced by ${masterUrl(
            this.cr2filenames[cr],
          )}`,
        )
      else candidates.push(`${Plugin.comptesRendusUrls(cr)} is a candidate`)
    }
    return candidates
  }

  async postCheck(analysis: any): Promise<any> {
    for (const filename of Object.keys(analysis)) {
      const result = analysis[filename]["agenda-00002"]
      if (result == null || result["status"] != "needs-fixing") continue
      const info = result["info"].join("")
      if (
        info.includes("meta QUANTIEME") ||
        info.includes("idJo field is missing")
      ) {
        result["info"] = result["info"].concat(this.findCandidates(filename))
      }
    }
  }

  async fix(agenda: any): Promise<any> {
    if (!(agenda.uid in remapJo)) {
      return null
    }

    const idJo = remapJo[agenda.uid]
    const numSeanceJo = String(Number(idJo.substring(5)))
    if (
      agenda.identifiants.idJo == idJo &&
      agenda.identifiants.numSeanceJo == numSeanceJo
    )
      return false
    agenda.identifiants.idJo = idJo
    agenda.identifiants.numSeanceJo = numSeanceJo
    return true
  }
}
