import versionCompare from "tiny-version-compare"
import { BasePlugin } from "../bugs"

// https://www.wikidata.org/wiki/Wikidata:WikiProject_France/Assembl%C3%A9e_Nationale#Legislative_session
const url =
  "https://query.wikidata.org/sparql?query=SELECT%20DISTINCT%20%3Fsession%20%3Fstart_time%20%3Fend_time%20%3Flabel%20%3Ffollows%20%3Ffollowed_by%20WHERE%20%7B%0A%0A%20%20%20%3Fsession%20wdt%3AP31%2Fwdt%3AP279*%20wd%3AQ87337481.%0A%20%20%20%3Fsession%20wdt%3AP580%20%3Fstart_time.%20%0A%20%20%20OPTIONAL%20%7B%20%3Fsession%20wdt%3AP582%20%3Fend_time%20%7D.%0A%20%20%20OPTIONAL%20%7B%20%3Fsession%20wdt%3AP155%20%3Ffollows%20%7D.%0A%20%20%20OPTIONAL%20%7B%20%3Fsession%20wdt%3AP156%20%3Ffollowed_by%20%7D.%0A%20%20%20OPTIONAL%20%7B%20%3Fsession%20rdfs%3Alabel%20%3Flabel%20filter%20(lang(%3Flabel)%20%3D%20%22fr%22)%20.%7D%0A%7D"

function compareString(a: string, b: string) {
  if (a < b) return -1
  if (a > b) return 1
  return 0
}

export class Plugin extends BasePlugin {
  sessions: any
  fetch: any

  constructor(options: any) {
    super(options, "agenda-00011")
    this.fetch = fetch
  }

  async preFix(): Promise<any> {
    return this.setSessions()
  }

  async preCheck(): Promise<any> {
    return this.setSessions()
  }

  async setSessions(): Promise<any> {
    const sessions = (await this.fetchSessions()).sort((a: any, b: any) =>
      compareString(a["start_time"]["value"], b["start_time"]["value"]),
    )
    this.sessions = sessions.map((s: any) => {
      return {
        label: s["label"]["value"].replace(/\s+en france/i, ""),
        debut: s["start_time"]["value"].substr(0, 10),
        fin: s["end_time"]["value"].substr(0, 10),
      }
    })
  }

  async get(url: string, headers: any): Promise<any> {
    return this.fetch(url, headers)
  }

  async fetchSessions(): Promise<any> {
    const response = await this.get(url, {
      headers: { Accept: "application/json" },
    })
    if (!response.ok) throw response
    const payload = await response.json()
    return payload["results"]["bindings"]
  }

  lookupSession(debut: string): any {
    const max = this.sessions.length - 1
    const min = 0
    let h = max
    let m
    let l = min
    let session
    let ok
    do {
      m = Math.floor((l + h) / 2)
      session = this.sessions[m]
      ok = session.debut <= debut && debut <= session.fin
      if (!ok) {
        if (l == m && m == h)
          // in between sessions
          return null
        if (debut < session.debut) {
          if (m <= min) return null
          h = m
        } else if (debut > session.fin) {
          if (m >= max) return null
          l = l == m ? h : m
        }
      }
    } while (!ok)
    return session
  }

  inScope(agenda: any) {
    if (
      agenda.xsiType != "seance_type" ||
      (agenda.lieu !== undefined &&
        agenda.lieu.code != "CG" &&
        agenda.lieu.code != "AN")
    )
      return false
    if (agenda.cycleDeVie.etat != "Confirmé") return false
    return true
  }

  async check(agenda: any, _filename: any): Promise<any> {
    if (!this.inScope(agenda)) return null
    const debut = agenda.timestampDebut.substr(0, 10)
    const session = this.lookupSession(debut)
    if (session == null) {
      return { status: "needs-fixing", info: [`${debut} is not in a session`] }
    } else {
      return { status: "ok", info: [""] }
    }
  }

  maybeUpgradeSchemaVersion(agenda: any) {
    if (agenda.schemaVersion !== undefined) {
      const version = agenda.schemaVersion.split("-")[1]
      if (versionCompare(version, "1.0") == 1) return
    }
    agenda.schemaVersion = "agenda-1.1"
  }

  async fix(agenda: any, _filename: any): Promise<any> {
    let status
    if ("sessionRef" in agenda) {
      // sessionRef is undefined and must be removed
      delete agenda.sessionRef
      status = true
    } else {
      // if there is no sessionRef to remove, file which is not in scope does not need to be rewritten
      status = null
    }

    if (!this.inScope(agenda)) return status

    if ("session" in agenda) delete agenda.session

    const session = this.lookupSession(agenda.timestampDebut.substr(0, 10))
    if (session == null) {
      return null
    } else {
      agenda.session = session
      this.maybeUpgradeSchemaVersion(agenda)
      return true
    }
  }
}
