import { load } from "../../file_systems"
import { assert } from "chai"
import { Plugin } from "../agenda-00011"
import fetchMock from "fetch-mock"

const id = "00011"

suite(`Plugin${id}`)

test(`#Plugin${id}FindSession`, async function () {
  const plugin = new Plugin({})

  const sessions = [
    {
      debut: "2019-09-10",
      fin: "2019-09-27",
    },
    {
      debut: "2019-10-01",
      fin: "2020-06-30",
    },
    {
      debut: "2020-06-31",
      fin: "2020-07-31",
    },
    {
      debut: "2020-08-01",
      fin: "2020-08-31",
    },
    {
      debut: "2020-09-01",
      fin: "2020-09-30",
    },
  ]

  plugin.sessions = sessions.slice(0, 2)
  // between sessions
  assert.strictEqual(plugin.lookupSession("2019-09-28"), null)
  assert.strictEqual(plugin.lookupSession("2017-01-01"), null)
  assert.deepEqual(plugin.lookupSession("2019-09-15"), plugin.sessions[0])
  assert.deepEqual(plugin.lookupSession("2020-02-02"), plugin.sessions[1])
  assert.strictEqual(plugin.lookupSession("2022-01-01"), null)

  plugin.sessions = sessions.slice(0, 1)
  assert.strictEqual(plugin.lookupSession("2017-01-01"), null)
  assert.deepEqual(plugin.lookupSession("2019-09-15"), plugin.sessions[0])
  assert.strictEqual(plugin.lookupSession("2020-02-02"), null)
  assert.strictEqual(plugin.lookupSession("2022-01-01"), null)

  plugin.sessions = sessions
  assert.strictEqual(plugin.lookupSession("2017-01-01"), null)
  assert.deepEqual(plugin.lookupSession("2019-09-15"), plugin.sessions[0])
  assert.deepEqual(plugin.lookupSession("2020-02-02"), plugin.sessions[1])
  assert.deepEqual(plugin.lookupSession("2020-09-01"), plugin.sessions[4])
  assert.deepEqual(plugin.lookupSession("2020-09-30"), plugin.sessions[4])
  assert.strictEqual(plugin.lookupSession("2022-01-01"), null)
})

test(`#Plugin${id}Load`, async function () {
  const plugin = new Plugin({})
  const d = `src/bugs/agenda-${id}`

  const wikidata = load(`${d}/wikidata.json`)
  plugin.fetch = fetchMock.route(/query\.wikidata\.org/, {
    body: wikidata,
    status: 200,
  })
  await plugin.preFix()

  const expected = [
    {
      debut: "2019-09-10",
      fin: "2019-09-27",
      label: "SESSION02",
    },
    {
      debut: "2019-10-01",
      fin: "2020-06-30",
      label: "SESSION01",
    },
  ]
  assert.deepEqual(plugin.sessions, expected)
})

test(`#Plugin${id}Fix`, async function () {
  const plugin = new Plugin({})
  const d = `src/bugs/agenda-${id}`

  const wikidata = load(`${d}/wikidata.json`)
  plugin.fetch = fetchMock.route(/query\.wikidata\.org/, {
    body: wikidata,
    status: 200,
  })

  await plugin.preFix()
  assert(plugin.sessions != undefined)

  {
    const f = `${d}/ok.json`
    const content = load(f)
    assert("sessionRef" in content, content)
    assert.equal(await plugin.fix(content, f), true, f)
    assert(!("sessionRef" in content), content)
    assert(content.session.label, "SESSION01")
  }

  {
    const f = `${d}/oldschema.json`
    const content = load(f)
    assert.equal(content.schemaVersion, "agenda-1.0")
    assert.equal(await plugin.fix(content, f), true, f)
    assert.equal(content.schemaVersion, "agenda-1.1")
  }

  {
    const f = `${d}/noschema.json`
    const content = load(f)
    assert.strictEqual(content.schemaVersion, undefined)
    assert.equal(await plugin.fix(content, f), true, f)
    assert.equal(content.schemaVersion, "agenda-1.1")
  }

  {
    const f = `${d}/nosessionref.json`
    const content = load(f)
    assert.equal(await plugin.fix(content, f), true, f)
    assert.equal(content.session.label, "SESSION01")
  }

  {
    const f = `${d}/session.json`
    const content = load(f)
    assert.equal(content.session.label, "SOMETHING")
    assert.equal(await plugin.fix(content, f), true, f)
    assert.equal(content.session.label, "SESSION01")
  }

  {
    const f = `${d}/not_seance_type.json`
    const content = load(f)
    assert.equal(await plugin.fix(content, f), null, f)
  }
})

test(`#Plugin${id}Check`, async function () {
  const plugin = new Plugin({})
  const d = `src/bugs/agenda-${id}`

  const wikidata = load(`${d}/wikidata.json`)
  plugin.fetch = fetchMock.route(/query\.wikidata\.org/, {
    body: wikidata,
    status: 200,
  })

  await plugin.preCheck()
  assert(plugin.sessions != undefined)

  {
    const f = `${d}/ok.json`
    const content = load(f)
    const result = await plugin.check(content, f)
    assert.deepEqual(result, { status: "ok", info: [""] }, f)
  }

  {
    const f = `${d}/badtimestamp.json`
    const content = load(f)
    const result = await plugin.check(content, f)
    assert.deepEqual(
      result,
      { status: "needs-fixing", info: ["2017-06-20 is not in a session"] },
      f,
    )
  }
})

test(`#Plugin${id}FetchFail`, async function () {
  const plugin = new Plugin({})

  plugin.fetch = fetchMock.route(/query\.wikidata\.org/, 500)

  let status = "unexpected"
  await plugin
    .preFix()
    .then(() => (status = "exception not thrown"))
    .catch((e) => (status = String(e.status)))
  assert.equal(status, "500")
})
