import assert from "node:assert"

// Move constant to module scope to avoid recreation
const BOOLEAN_MAP = {
  'false': false,
  'true': true,
} as const;

const XML_NAMESPACE = "http://schemas.assemblee-nationale.fr/referentiel";
const XSI_NAMESPACE = "http://www.w3.org/2001/XMLSchema-instance";

export function cleanBooleanAttribute(
  o: { [key: string]: any },
  key: string,
): boolean | undefined {
  const stringValue = o[key]
  if (stringValue === undefined || stringValue === null) {
    delete o[key]
    return undefined
  }
  const booleanValue: boolean = BOOLEAN_MAP[stringValue as keyof typeof BOOLEAN_MAP]
  assert.notStrictEqual(
    booleanValue,
    undefined,
    `Invalid boolean value: ${key} = ${stringValue}`,
  )
  o[key] = booleanValue
  return booleanValue
}

export function cleanIntegerAttribute(
  o: { [key: string]: any },
  key: string,
): number | undefined {
  const stringValue = o[key]
  if (stringValue === undefined || stringValue === null) {
    delete o[key]
    return undefined
  }
  const integerValue = Number.parseInt(stringValue)
  assert(!Number.isNaN(integerValue), `Invalid number value: ${key} = ${stringValue}`)
  o[key] = integerValue
  return integerValue
}

/// Remove XML fields that have no use in JSON.
export function cleanXmlArtefacts(o: { [key: string]: any }): void {
  // Optimize namespace checks with direct property access
  const xmlns = o["@xmlns"] ?? o["xmlns"];
  if (xmlns !== undefined) {
    assert.strictEqual(xmlns, XML_NAMESPACE);
    delete o["@xmlns"];
    delete o["xmlns"];
  }

  if (o["@xmlns:xsi"] === XSI_NAMESPACE) {
    delete o["@xmlns:xsi"];
  }

  // Optimize empty checks with direct property access
  for (const key in o) {
    const value = o[key];
    if (value === null) {
      delete o[key];
      continue;
    }
    
    if (value === undefined) continue;

    if (
      value["@xsi:nil"] === "true" ||
      (Array.isArray(value) && value.length === 0) ||
      (!Array.isArray(value) && value.constructor === Object && Object.keys(value).length === 0)
    ) {
      delete o[key];
    }
  }
}
