import { createHash } from "crypto"
import assert from "assert"
import fs from "fs-extra"
import path from "path"

import { Legislature } from "./types/legislatures"

export interface Dataset {
  filename: string
  ignoreForWeb?: boolean
  legislature: Legislature
  repairZip?: (dataset: Dataset, dataDir: string) => void
  structure: DatasetStructure
  title: string
  url?: string
}

export interface Datasets {
  acteursEtOrganes: Dataset[]
  agendas: Dataset[]
  amendements: Dataset[]
  dossiersLegislatifs: Dataset[]
  scrutins: Dataset[]
  questions: Dataset[]
  comptesRendusSeances: Dataset[]
}

export enum DatasetStructure {
  SingleFile,
  SegmentedFiles,
}

export enum EnabledDatasets {
  None = 0,
  ActeursEtOrganes = 1 << 0,
  Agendas = 1 << 1,
  Amendements = 1 << 2,
  DossiersLegislatifs = 1 << 3,
  Photos = 1 << 4,
  Scrutins = 1 << 5,
  Questions = 1 << 6,
  ComptesRendusSeances = 1 << 7,
  All = ActeursEtOrganes |
    Agendas |
    Amendements |
    DossiersLegislatifs |
    Photos |
    Scrutins |
    Questions |
    ComptesRendusSeances,
}

// The zip files (named NAME.json.zip) usually contain a single subfolder named "json"
// containing the actual contents
// We'd rather rename the folder NAME.json
function repairForJsonSubdirectory(dataset: Dataset, dataDir: string) {
  const targetDir = path.join(dataDir, dataset.filename)
  if (fs.existsSync(targetDir)) {
    fs.removeSync(targetDir)
  }

  if (fs.existsSync(path.join(dataDir, "json"))) {
    fs.moveSync(path.join(dataDir, "json"), targetDir)
  } else {
    fs.mkdirSync(targetDir)
  }
}

export const datasets: Datasets = {
  acteursEtOrganes: [
    {
      // AMO10 (XVII)
      //
      // Contient les députés actifs
      // * avec tous les mandats en cours
      // * y compris avec leur parti politique (typeOrgane: "PARPOL")
      // * mais sans les mandats achevés (y compris dans cette législature)
      //
      // Organes :
      // * Contient seulement les organes correspondants aux mandats en cours
      // * avec le xsiType de l'organe
      // * avec des commissions sénatoriales (permanentes et spéciales) en cours
      // * avec des délégations sénatoriales en cours
      // * avec des groupes sénatoriaux en cours
      // * avec des ministères en cours
      // * avec les Présidences de la République (y compris Sarkozy)
      // * avec le Conseil constitutionnel
      //
      // Ce fichier est un sous-ensemble à tous les niveaux du fichier AMO20
      // (qui est lui même un sous-ensemble de l'AMO30).
      // Il n'apporte donc absolument rien par rapport à AMO20 ou l'AMO30.
      filename: "AMO10_deputes_actifs_mandats_actifs_organes_XVII.json",
      ignoreForWeb: true,
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Députés actifs et organes de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/amo/deputes_actifs_mandats_actifs_organes/AMO10_deputes_actifs_mandats_actifs_organes.json.zip",
    },
    {
      // AMO10 (XVI)
      //
      // Contient les députés actifs
      // * avec tous les mandats en cours
      // * y compris avec leur parti politique (typeOrgane: "PARPOL")
      // * mais sans les mandats achevés (y compris dans cette législature)
      //
      // Organes :
      // * Contient seulement les organes correspondants aux mandats en cours
      // * avec le xsiType de l'organe
      // * avec des commissions sénatoriales (permanentes et spéciales) en cours
      // * avec des délégations sénatoriales en cours
      // * avec des groupes sénatoriaux en cours
      // * avec des ministères en cours
      // * avec les Présidences de la République (y compris Sarkozy)
      // * avec le Conseil constitutionnel
      //
      // Ce fichier est un sous-ensemble à tous les niveaux du fichier AMO20
      // (qui est lui même un sous-ensemble de l'AMO30).
      // Il n'apporte donc absolument rien par rapport à AMO20 ou l'AMO30.
      filename: "AMO10_deputes_actifs_mandats_actifs_organes_XVI.json",
      ignoreForWeb: true,
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Députés actifs et organes de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/amo/deputes_actifs_mandats_actifs_organes/AMO10_deputes_actifs_mandats_actifs_organes.json.zip",
    },
    {
      // AMO10 (XV)
      //
      // Contient les députés actifs
      // * avec tous les mandats en cours
      // * y compris avec leur parti politique (typeOrgane: "PARPOL")
      // * mais sans les mandats achevés (y compris dans cette législature)
      //
      // Organes :
      // * Contient seulement les organes correspondants aux mandats en cours
      // * avec le xsiType de l'organe
      // * avec des commissions sénatoriales (permanentes et spéciales) en cours
      // * avec des délégations sénatoriales en cours
      // * avec des groupes sénatoriaux en cours
      // * avec des ministères en cours
      // * avec les Présidences de la République (y compris Sarkozy)
      // * avec le Conseil constitutionnel
      //
      // Ce fichier est un sous-ensemble à tous les niveaux du fichier AMO20
      // (qui est lui même un sous-ensemble de l'AMO30).
      // Il n'apporte donc absolument rien par rapport à AMO20 ou l'AMO30.
      filename: "AMO10_deputes_actifs_mandats_actifs_organes_XV.json",
      ignoreForWeb: true,
      legislature: Legislature.Quinze,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Députés actifs et organes de la XVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/amo/deputes_actifs_mandats_actifs_organes/AMO10_deputes_actifs_mandats_actifs_organes_XV.json.zip",
    },
    {
      // AMO20 (XVII)
      //
      // Contient les députés de la législature, actifs ou non (invalidés, etc), les sénateurs actuels
      // Pour les députés :
      // * avec tous leurs mandats de la législature
      // * y compris les mandats achevés
      // * y compris leurs partis politiques  (typeOrgane: "PARPOL")
      // * mais sans les mandats des législatures précédentes (qui sont dans AMO40)
      //
      // Contient tous les organes référencés par les mandats
      // * avec le xsiType de l'organe
      // * avec des commissions sénatoriales (permanentes et spéciales) en cours
      // * avec des délégations sénatoriales en cours
      // * avec des groupes sénatoriaux en cours
      // * avec des ministères en cours
      // * avec les Présidences de la République (y compris Sarkozy)
      // * avec le Conseil constitutionnel
      //
      // Ce fichier est un sous-ensemble à tous les niveaux du fichier AMO30.
      // Il n'apporte donc absolument rien par rapport à l'AMO30.
      filename: "AMO20_dep_sen_min_tous_mandats_et_organes_XVII.json",
      ignoreForWeb: true,
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Acteurs et organes de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/amo/deputes_senateurs_ministres_legislature/AMO20_dep_sen_min_tous_mandats_et_organes.json.zip",
    },
    {
      // AMO20 (XVI)
      //
      // Contient les députés de la législature, actifs ou non (invalidés, etc), les sénateurs actuels
      // Pour les députés :
      // * avec tous leurs mandats de la législature
      // * y compris les mandats achevés
      // * y compris leurs partis politiques  (typeOrgane: "PARPOL")
      // * mais sans les mandats des législatures précédentes (qui sont dans AMO40)
      //
      // Contient tous les organes référencés par les mandats
      // * avec le xsiType de l'organe
      // * avec des commissions sénatoriales (permanentes et spéciales) en cours
      // * avec des délégations sénatoriales en cours
      // * avec des groupes sénatoriaux en cours
      // * avec des ministères en cours
      // * avec les Présidences de la République (y compris Sarkozy)
      // * avec le Conseil constitutionnel
      //
      // Ce fichier est un sous-ensemble à tous les niveaux du fichier AMO30.
      // Il n'apporte donc absolument rien par rapport à l'AMO30.
      filename: "AMO20_dep_sen_min_tous_mandats_et_organes_XVI.json",
      ignoreForWeb: true,
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Acteurs et organes de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/amo/deputes_senateurs_ministres_legislature/AMO20_dep_sen_min_tous_mandats_et_organes.json.zip",
    },
    {
      // AMO20 (XV)
      //
      // Contient les députés de la législature, actifs ou non (invalidés, etc), les sénateurs actuels
      // Pour les députés :
      // * avec tous leurs mandats de la législature
      // * y compris les mandats achevés
      // * y compris leurs partis politiques  (typeOrgane: "PARPOL")
      // * mais sans les mandats des législatures précédentes (qui sont dans AMO40)
      //
      // Contient tous les organes référencés par les mandats
      // * avec le xsiType de l'organe
      // * avec des commissions sénatoriales (permanentes et spéciales) en cours
      // * avec des délégations sénatoriales en cours
      // * avec des groupes sénatoriaux en cours
      // * avec des ministères en cours
      // * avec les Présidences de la République (y compris Sarkozy)
      // * avec le Conseil constitutionnel
      //
      // Ce fichier est un sous-ensemble à tous les niveaux du fichier AMO30.
      // Il n'apporte donc absolument rien par rapport à l'AMO30.
      filename: "AMO20_dep_sen_min_tous_mandats_et_organes_XV.json",
      ignoreForWeb: true,
      legislature: Legislature.Quinze,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Acteurs et organes de la XVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/amo/deputes_senateurs_ministres_legislature/AMO20_dep_sen_min_tous_mandats_et_organes_XV.json.zip",
    },
    {
      // AMO20 (XIV)
      //
      // Comme AMO20 (XV) mais pour la XIVème législature
      // Tous les acteurs sauf un (PA429892) sont présents en plus à jour dans AMO30.
      // Cet acteur PA429892 (ancienne Sénatrice) n'est présent dans aucun des autres
      // fichiers AMO.
      // Tous les organes sont présents en plus complets dans AMO30.
      //
      // Ce fichier est donc à un détail prêt (PA429892) un sous-ensemble d'AMO30.
      // Il n'apporte donc quasiment rien par rapport à l'AMO30.
      filename: "AMO20_dep_sen_min_tous_mandats_et_organes_XIV.json",
      ignoreForWeb: true,
      legislature: Legislature.Quatorze,
      structure: DatasetStructure.SingleFile,
      title: "Acteurs et organes de la XIVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/14/amo/deputes_senateurs_ministres_legislatures_XIV/AMO20_dep_sen_min_tous_mandats_et_organes_XIV.json.zip",
    },
    {
      // AMO30 (XVII)
      //
      // Ce fichier contient tous les acteurs et organes.
      // C'est un sur-ensemble des fichiers AMO20.
      filename: "AMO30_tous_acteurs_tous_mandats_tous_organes_historique.json",
      legislature: Legislature.All,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title:
        "Historique de tous les acteurs et organes depuis la XIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/amo/tous_acteurs_mandats_organes_xi_legislature/AMO30_tous_acteurs_tous_mandats_tous_organes_historique.json.zip",
    },
    // {
    //   // AMO30 (XVI)
    //   //
    //   // Ce fichier contient tous les acteurs et organes.
    //   // C'est un sur-ensemble des fichiers AMO20.
    //   filename: "AMO30_tous_acteurs_tous_mandats_tous_organes_historique.json",
    //   legislature: Legislature.All,
    //   repairZip: repairForJsonSubdirectory,
    //   structure: DatasetStructure.SegmentedFiles,
    //   title:
    //     "Historique de tous les acteurs et organes depuis la XIème législature",
    //   url: "http://data.assemblee-nationale.fr/static/openData/repository/16/amo/tous_acteurs_mandats_organes_xi_legislature/AMO30_tous_acteurs_tous_mandats_tous_organes_historique.json.zip",
    // },
    // {
    //   // AMO30 (XV)
    //   //
    //   // Ce fichier contient tous les acteurs et organes.
    //   // C'est un sur-ensemble des fichiers AMO20.
    //   filename: "AMO30_tous_acteurs_tous_mandats_tous_organes_historique.json",
    //   legislature: Legislature.All,
    //   repairZip: (dataset: Dataset, dataDir: string) => {
    //     const targetDir = path.join(dataDir, dataset.filename)
    //     if (fs.existsSync(targetDir)) {
    //       fs.removeSync(targetDir)
    //     }
    //     fs.moveSync(path.join(dataDir, "json"), targetDir)
    //   },
    //   structure: DatasetStructure.SegmentedFiles,
    //   title:
    //     "Historique de tous les acteurs et organes depuis la XIème législature",
    //   url: "http://data.assemblee-nationale.fr/static/openData/repository/15/amo/tous_acteurs_mandats_organes_xi_legislature/AMO30_tous_acteurs_tous_mandats_tous_organes_historique.json.zip",
    // },
    {
      // AMO40 (XVII)
      //
      // Contient les députés actifs (les mêmes qu'AMO20)
      // * avec presque tous leurs mandats
      // * y compris les mandats achevés et ceux des législatures précédentes
      // * y compris leurs partis politiques précédents
      // * mais sans leur parti politique actuel (typeOrgane: "PARPOL")
      //
      // Contient beaucoup d'organes référencés, y compris ceux des législatures
      // précédentes
      // * mais il manque le xsiType des organes
      // * ne contient pas les organes sénatoriaux
      // * ne contient pas les ministères
      // * ne contient pas la Présidence de la République
      // * ne contient pas le Conseil constitutionnel
      //
      // Pour les acteurs, ce dépôt est un sous-ensemble strict de AMO30 :
      // Il a moins d'acteurs et il sont moins complets.
      //
      // Pour les organes :
      // Quand l'organe est aussi présent dans AMO30, il vaut mieux prendre celui
      // d'AMO30, car il contient le xsiType (c'est la seule différence semble-t-i).
      // Sinon, il faut le prendre seulement si on est intéressé par les législatures
      // précédentes.
      filename: "AMO40_deputes_actifs_mandats_actifs_organes_divises_XVII.json",
      ignoreForWeb: true,
      legislature: Legislature.DixSept,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.mkdirSync(targetDir)
        if (fs.pathExistsSync(path.join(dataDir, "acteur"))) {
          // Note; Le répertoire acteur peut disparaître en fin de législature
          // quand les mandats des députés s'interrompent.
          fs.moveSync(
            path.join(dataDir, "acteur"),
            path.join(targetDir, "acteur"),
          )
        }
        fs.moveSync(
          path.join(dataDir, "organe"),
          path.join(targetDir, "organe"),
        )
      },
      structure: DatasetStructure.SegmentedFiles,
      title: "Députés actifs et organes de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/amo/deputes_actifs_mandats_actifs_organes_divises/AMO40_deputes_actifs_mandats_actifs_organes_divises.json.zip",
    },
    {
      // AMO40 (XVI)
      //
      // Contient les députés actifs (les mêmes qu'AMO20)
      // * avec presque tous leurs mandats
      // * y compris les mandats achevés et ceux des législatures précédentes
      // * y compris leurs partis politiques précédents
      // * mais sans leur parti politique actuel (typeOrgane: "PARPOL")
      //
      // Contient beaucoup d'organes référencés, y compris ceux des législatures
      // précédentes
      // * mais il manque le xsiType des organes
      // * ne contient pas les organes sénatoriaux
      // * ne contient pas les ministères
      // * ne contient pas la Présidence de la République
      // * ne contient pas le Conseil constitutionnel
      //
      // Pour les acteurs, ce dépôt est un sous-ensemble strict de AMO30 :
      // Il a moins d'acteurs et il sont moins complets.
      //
      // Pour les organes :
      // Quand l'organe est aussi présent dans AMO30, il vaut mieux prendre celui
      // d'AMO30, car il contient le xsiType (c'est la seule différence semble-t-i).
      // Sinon, il faut le prendre seulement si on est intéressé par les législatures
      // précédentes.
      filename: "AMO40_deputes_actifs_mandats_actifs_organes_divises_XVI.json",
      ignoreForWeb: true,
      legislature: Legislature.Seize,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.mkdirSync(targetDir)
        if (fs.pathExistsSync(path.join(dataDir, "acteur"))) {
          // Note; Le répertoire acteur peut disparaître en fin de législature
          // quand les mandats des députés s'interrompent.
          fs.moveSync(
            path.join(dataDir, "acteur"),
            path.join(targetDir, "acteur"),
          )
        }
        fs.moveSync(
          path.join(dataDir, "organe"),
          path.join(targetDir, "organe"),
        )
      },
      structure: DatasetStructure.SegmentedFiles,
      title: "Députés actifs et organes de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/amo/deputes_actifs_mandats_actifs_organes_divises/AMO40_deputes_actifs_mandats_actifs_organes_divises.json.zip",
    },
    {
      // AMO40 (XV)
      //
      // Contient les députés actifs (les mêmes qu'AMO20)
      // * avec presque tous leurs mandats
      // * y compris les mandats achevés et ceux des législatures précédentes
      // * y compris leurs partis politiques précédents
      // * mais sans leur parti politique actuel (typeOrgane: "PARPOL")
      //
      // Contient beaucoup d'organes référencés, y compris ceux des législatures
      // précédentes
      // * mais il manque le xsiType des organes
      // * ne contient pas les organes sénatoriaux
      // * ne contient pas les ministères
      // * ne contient pas la Présidence de la République
      // * ne contient pas le Conseil constitutionnel
      //
      // Pour les acteurs, ce dépôt est un sous-ensemble strict de AMO30 :
      // Il a moins d'acteurs et il sont moins complets.
      //
      // Pour les organes :
      // Quand l'organe est aussi présent dans AMO30, il vaut mieux prendre celui
      // d'AMO30, car il contient le xsiType (c'est la seule différence semble-t-i).
      // Sinon, il faut le prendre seulement si on est intéressé par les législatures
      // précédentes.
      filename: "AMO40_deputes_actifs_mandats_actifs_organes_divises_XV.json",
      ignoreForWeb: true,
      legislature: Legislature.Quinze,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.mkdirSync(targetDir)
        fs.moveSync(
          path.join(dataDir, "acteur"),
          path.join(targetDir, "acteur"),
        )
        fs.moveSync(
          path.join(dataDir, "organe"),
          path.join(targetDir, "organe"),
        )
      },
      structure: DatasetStructure.SegmentedFiles,
      title: "Députés actifs et organes de la XVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/amo/deputes_actifs_mandats_actifs_organes_divises/AMO40_deputes_actifs_mandats_actifs_organes_divises_XV.json.zip",
    },
    {
      // AMO50 (XVII)
      //
      // Attention, ce fichier n'est pas mis à jour régulièrement !
      //
      // Pour les acteurs :
      // * c'est un sous ensemble obsolète d'AMO30
      // * il manque les xsiType aux mandats
      //
      // Pour les organes :
      // * c'est un sous ensemble obsolète d'AMO30
      // * il manque les xsiType
      //
      // Ce fichier n'apporte donc absolument rien par rapport à l'AMO30,
      // dont il est un sous ensemble obsolète.
      filename: "AMO50_acteurs_mandats_organes_divises_XVII.json",
      ignoreForWeb: true,
      legislature: Legislature.DixSept,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.mkdirSync(targetDir)
        fs.moveSync(
          path.join(dataDir, "mandat"),
          path.join(targetDir, "mandat"),
        )
        fs.moveSync(
          path.join(dataDir, "acteur"),
          path.join(targetDir, "acteur"),
        )
        fs.moveSync(
          path.join(dataDir, "organe"),
          path.join(targetDir, "organe"),
        )
      },
      structure: DatasetStructure.SegmentedFiles,
      title:
        "Acteurs, mandats et organes de la XVIIème législature (fichiers séparés, y compris mandats)",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/amo/acteurs_mandats_organes_divises/AMO50_acteurs_mandats_organes_divises.json.zip",
    },
    {
      // AMO50 (XVI)
      //
      // Attention, ce fichier n'est pas mis à jour régulièrement !
      //
      // Pour les acteurs :
      // * c'est un sous ensemble obsolète d'AMO30
      // * il manque les xsiType aux mandats
      //
      // Pour les organes :
      // * c'est un sous ensemble obsolète d'AMO30
      // * il manque les xsiType
      //
      // Ce fichier n'apporte donc absolument rien par rapport à l'AMO30,
      // dont il est un sous ensemble obsolète.
      filename: "AMO50_acteurs_mandats_organes_divises_XVI.json",
      ignoreForWeb: true,
      legislature: Legislature.Seize,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.mkdirSync(targetDir)
        fs.moveSync(
          path.join(dataDir, "mandat"),
          path.join(targetDir, "mandat"),
        )
        fs.moveSync(
          path.join(dataDir, "acteur"),
          path.join(targetDir, "acteur"),
        )
        fs.moveSync(
          path.join(dataDir, "organe"),
          path.join(targetDir, "organe"),
        )
      },
      structure: DatasetStructure.SegmentedFiles,
      title:
        "Acteurs, mandats et organes de la XVIème législature (fichiers séparés, y compris mandats)",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/amo/acteurs_mandats_organes_divises/AMO50_acteurs_mandats_organes_divises.json.zip",
    },
    {
      // AMO50 (XV)
      //
      // Attention, ce fichier n'est pas mis à jour régulièrement !
      //
      // Pour les acteurs :
      // * c'est un sous ensemble obsolète d'AMO30
      // * il manque les xsiType aux mandats
      //
      // Pour les organes :
      // * c'est un sous ensemble obsolète d'AMO30
      // * il manque les xsiType
      //
      // Ce fichier n'apporte donc absolument rien par rapport à l'AMO30,
      // dont il est un sous ensemble obsolète.
      filename: "AMO50_acteurs_mandats_organes_divises_XV.json",
      ignoreForWeb: true,
      legislature: Legislature.Quinze,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.mkdirSync(targetDir)
        fs.moveSync(
          path.join(dataDir, "mandat"),
          path.join(targetDir, "mandat"),
        )
        fs.moveSync(
          path.join(dataDir, "acteur"),
          path.join(targetDir, "acteur"),
        )
        fs.moveSync(
          path.join(dataDir, "organe"),
          path.join(targetDir, "organe"),
        )
      },
      structure: DatasetStructure.SegmentedFiles,
      title:
        "Acteurs, mandats et organes de la XVème législature (fichiers séparés, y compris mandats)",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/amo/acteurs_mandats_organes_divises/AMO50_acteurs_mandats_organes_divises_XV.json.zip",
    },
  ],
  agendas: [
    {
      filename: "Agenda_XVII.json",
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Agenda de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/vp/reunions/Agenda.json.zip",
    },
    {
      filename: "Agenda_XVI.json",
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Agenda de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/vp/reunions/Agenda.json.zip",
    },
    {
      filename: "Agenda_XV.json",
      legislature: Legislature.Quinze,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Agenda de la XVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/vp/reunions/Agenda_XV.json.zip",
    },
    {
      filename: "Agenda_XIV.json",
      legislature: Legislature.Quatorze,
      structure: DatasetStructure.SingleFile,
      title: "Agenda de la XIVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/14/vp/reunions/Agenda_XIV.json.zip",
    },
  ],
  amendements: [
    {
      filename: "Amendements_XVII.json",
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Amendements de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/loi/amendements_div_legis/Amendements.json.zip",
    },
    {
      filename: "Amendements_XVI.json",
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Amendements de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/loi/amendements_div_legis/Amendements.json.zip",
    },
    {
      filename: "Amendements_XV.json",
      legislature: Legislature.Quinze,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Amendements de la XVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/loi/amendements_legis/Amendements_XV.json.zip",
    },
    // {
    //   filename: "Amendements_XIV.json",
    //   ignoreForWeb: true,
    //   legislature: Legislature.Quatorze,
    //   structure: DatasetStructure.SingleFile,
    //   title: "Amendements de la XIVème législature",
    //   url:
    //     "http://data.assemblee-nationale.fr/static/openData/repository/14/loi/amendements_legis_XIV/Amendements_XIV.json.zip",
    // },
  ],
  dossiersLegislatifs: [
    {
      filename: "Dossiers_Legislatifs_XVII.json",
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Dossiers législatifs de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/loi/dossiers_legislatifs/Dossiers_Legislatifs.json.zip",
    },
    {
      filename: "Dossiers_Legislatifs_XVI.json",
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Dossiers législatifs de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/loi/dossiers_legislatifs/Dossiers_Legislatifs.json.zip",
    },
    {
      filename: "Dossiers_Legislatifs_XV.json",
      legislature: Legislature.Quinze,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Dossiers législatifs de la XVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/loi/dossiers_legislatifs/Dossiers_Legislatifs_XV.json.zip",
    },
    {
      filename: "Dossiers_Legislatifs_XIV.json",
      legislature: Legislature.Quatorze,
      structure: DatasetStructure.SingleFile,
      title: "Dossiers législatifs de la XIVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/14/loi/dossiers_legislatifs/Dossiers_Legislatifs_XIV.json.zip",
    },
  ],
  scrutins: [
    {
      filename: "Scrutins_XVII.json",
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Scrutins de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/loi/scrutins/Scrutins.json.zip",
    },
    {
      filename: "Scrutins_XVI.json",
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Scrutins de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/loi/scrutins/Scrutins.json.zip",
    },
    {
      filename: "Scrutins_XV.json",
      legislature: Legislature.Quinze,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Scrutins de la XVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/15/loi/scrutins/Scrutins_XV.json.zip",
    },
    {
      filename: "Scrutins_XIV.json",
      legislature: Legislature.Quatorze,
      structure: DatasetStructure.SingleFile,
      title: "Scrutins de la XIVème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/14/loi/scrutins/Scrutins_XIV.json.zip",
    },
  ],
  questions: [
    {
      filename: "Questions_ecrites_XVII.json",
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Questions écrites de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/questions/questions_ecrites/Questions_ecrites.json.zip",
    },
    {
      filename: "Questions_ecrites_XVI.json",
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Questions écrites de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/questions/questions_ecrites/Questions_ecrites.json.zip",
    },
    {
      filename: "Questions_gouvernement_XVII.json",
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Questions au gouvernement de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/questions/questions_gouvernement/Questions_gouvernement.json.zip",
    },
    {
      filename: "Questions_gouvernement_XVI.json",
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Questions au gouvernement de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/questions/questions_gouvernement/Questions_gouvernement.json.zip",
    },
    {
      filename: "Questions_orales_XVII.json",
      legislature: Legislature.DixSept,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Questions orales sans débats de la XVIIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/17/questions/questions_orales_sans_debat/Questions_orales_sans_debat.json.zip",
    },
    {
      filename: "Questions_orales_XVI.json",
      legislature: Legislature.Seize,
      repairZip: repairForJsonSubdirectory,
      structure: DatasetStructure.SegmentedFiles,
      title: "Questions orales sans débats de la XVIème législature",
      url: "http://data.assemblee-nationale.fr/static/openData/repository/16/questions/questions_orales_sans_debat/Questions_orales_sans_debat.json.zip",
    },
  ],
  comptesRendusSeances: [
    {
      filename: "Comptes_Rendus_Seances_XVII.xml",
      legislature: Legislature.DixSept,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.moveSync(path.join(dataDir, "xml"), targetDir)
      },
      structure: DatasetStructure.SegmentedFiles,
      title: "Comptes rendus des séances publiques de la XVIIème législature",
      url: "https://data.assemblee-nationale.fr/static/openData/repository/17/vp/syceronbrut/syseron.xml.zip",
    },
    {
      filename: "Comptes_Rendus_Seances_XVI.xml",
      legislature: Legislature.Seize,
      repairZip: (dataset: Dataset, dataDir: string) => {
        const targetDir = path.join(dataDir, dataset.filename)
        if (fs.existsSync(targetDir)) {
          fs.removeSync(targetDir)
        }
        fs.moveSync(path.join(dataDir, "xml"), targetDir)
      },
      structure: DatasetStructure.SegmentedFiles,
      title: "Comptes rendus des séances publiques de la XVIème législature",
      url: "https://data.assemblee-nationale.fr/static/openData/repository/16/vp/syceronbrut/syseron.xml.zip",
    },
  ],
}

const directory2schema: any = {
  ".*Agenda_.*": "agenda",
  ".*Scrutins_.*": "scrutin",
  ".*Questions_.*": "question",
  ".*Amendements_.*": "amendement",
  ".*Dossiers_Legislatifs_.*/documents": "document",
  ".*Dossiers_Legislatifs_.*/dossiers": "dossier",
  ".*acteurs_mandats_organes/organes": "organe",
  ".*acteurs_mandats_organes/acteurs": "acteur",
  ".*AMO30_tous_acteurs_tous_mandats_tous_organes_historique.*/organes":
    "organe",
  ".*AMO30_tous_acteurs_tous_mandats_tous_organes_historique.*/acteurs":
    "acteur",
  ".*Comptes_Rendus_Seances_.*": "debat",
}

export function getDatasets(): any {
  return [
    "Agenda_XIV",
    "Agenda_XV",
    "Agenda_XVI",
    "Agenda_XVII",
    "Scrutins_XIV",
    "Scrutins_XV",
    "Scrutins_XVI",
    "Scrutins_XVII",
    "Questions_Ecrites_XVI",
    "Questions_Ecrites_XVII",
    "Questions_Gouvernement_XVI",
    "Questions_Gouvernement_XVII",
    "Amendements_XIV",
    "Amendements_XV",
    "Amendements_XVI",
    "Amendements_XVII",
    "Dossiers_Legislatifs_XIV",
    "Dossiers_Legislatifs_XV",
    "Dossiers_Legislatifs_XVI",
    "Dossiers_Legislatifs_XVII",
    "AMO30_tous_acteurs_tous_mandats_tous_organes_historique",
    "acteurs_mandats_organes",
    "Comptes_Rendus_Seances_XVI",
    "Comptes_Rendus_Seances_XVII",
  ]
}

export function validDataset(dir: any): any {
  for (const dataset of getDatasets()) {
    if (dir.includes(dataset)) return true
  }
  return false
}

export function getSchemas(): any {
  return Object.values(directory2schema)
}

export function datasetDirectorySchema(dataset: string): any {
  let directories: any
  if (dataset.includes("Dossiers_Legislatifs_")) {
    directories = [`${dataset}/documents`, `${dataset}/dossiers`]
  } else if (
    dataset.includes("acteurs_mandats_organes") ||
    dataset.includes("AMO30_tous_acteurs_tous_mandats_tous_organes_historique")
  ) {
    directories = [`${dataset}/organes`, `${dataset}/acteurs`]
  } else {
    directories = [dataset]
  }

  let results = []
  for (const directory of directories) {
    for (const re of Object.keys(directory2schema)) {
      if (new RegExp(re).exec(directory))
        results.push([`${directory}/**/*.json`, directory2schema[re]])
    }
  }
  return results
}

export function fileDatasetUrl(filename: string): [string, string] | null {
  const re = "^.*/(.*)_nettoye/(.*)"
  const match = new RegExp(re).exec(filename)
  if (match == null) return null
  let [dataset, file] = [match[1], match[2]]
  if (dataset == "AMO30_tous_acteurs_tous_mandats_tous_organes_historique")
    dataset = "acteurs_mandats_organes"
  return [dataset, file]
}

export function diffUpstream(filename: string): string {
  const datasetUrl = fileDatasetUrl(filename)
  if (datasetUrl == null) return filename
  const [dataset, file] = datasetUrl
  const sum = createHash("sha1").update(file).digest("hex").toString()
  return `[${file}](https://git.en-root.org/tricoteuses/data.tricoteuses.fr/${dataset}/-/compare/upstream...master#${sum})`
}

export function masterUrl(filename: string): string {
  const datasetUrl = fileDatasetUrl(filename)
  if (datasetUrl == null) return filename
  const [dataset, file] = datasetUrl
  return `[${file}](https://git.en-root.org/tricoteuses/data.tricoteuses.fr/${dataset}/-/tree/master/${file})`
}

export function filterByLegislature(
  datasets: Datasets,
  legislature: any,
): Datasets {
  // When a given legislature is 0, it is the same as giving no legislature:
  // It means "every legislatures".
  if (legislature !== 0 && legislature !== undefined) {
    for (let [key, value] of Object.entries(datasets)) {
      datasets[key as keyof typeof datasets] = value.filter(
        (dataset: { legislature: any }) =>
          dataset.legislature === Legislature.All ||
          legislature === dataset.legislature,
      )
    }
  }
  return datasets
}

export function getEnabledDatasets(categories: any): EnabledDatasets {
  categories.map((datasetName: string) =>
    assert.notStrictEqual(
      (EnabledDatasets as any as { [name: string]: EnabledDatasets })[
        datasetName
      ],
      undefined,
      `Unknown name of dataset: ${datasetName}`,
    ),
  )

  return categories.reduce(
    (enabledDatasets: EnabledDatasets, datasetName: string): EnabledDatasets =>
      enabledDatasets |
      (EnabledDatasets as any as { [name: string]: EnabledDatasets })[
        datasetName
      ],
    EnabledDatasets.None,
  )
}
