import { format } from "date-fns/format"
import { fr as frLocale } from "date-fns/locale/fr"
import { parse } from "date-fns/parse"
import { parseISO } from "date-fns/parseISO"

// Patch Date.prototype.toJSON to ensure that dates keep their timezone when converted to strings.
// https://stackoverflow.com/questions/31096130/how-to-json-stringify-a-javascript-date-and-preserve-timezone
export const existingDateToJson = Date.prototype.toJSON
export const patchedDateToJson = function (this: Date): string {
  try {
    const hours = this.getHours()
    const minutes = this.getMinutes()
    const seconds = this.getSeconds()
    const milliseconds = this.getMilliseconds()

    const timezoneOffsetInHours = -(this.getTimezoneOffset() / 60) //UTC minus local time
    // When year <= 1900, timezoneOffsetInHours is 0.15, minutes is 9 and seconds is 21!
    if (
      (timezoneOffsetInHours === 0.15 &&
        hours === 0 &&
        minutes === 9 &&
        seconds === 21) ||
      (hours - timezoneOffsetInHours === 0 &&
        minutes === 0 &&
        seconds === 0 &&
        milliseconds === 0)
    ) {
      return this.toISOString().split("T")[0]
    }

    // It's a bit unfortunate that we need ttimezoneOffsetInHours = -(this.getTimezoneOffset() / 60) o construct a new Date instance
    // (we don't want _this_ Date instance to be modified).
    const correctedDate = new Date(
      this.getFullYear(),
      this.getMonth(),
      this.getDate(),
      hours,
      minutes,
      seconds,
      milliseconds,
    )
    correctedDate.setHours(this.getHours() + timezoneOffsetInHours)
    // Remove milliseconds & trailing "Z".
    const iso = correctedDate.toISOString().replace(/(\.0+)?Z$/, "")
    const sign = timezoneOffsetInHours >= 0 ? "+" : "-"
    const leadingZero = Math.abs(timezoneOffsetInHours) < 10 ? "0" : ""
    return (
      iso +
      sign +
      leadingZero +
      Math.abs(timezoneOffsetInHours).toString().split(".")[0] +
      ":00"
    )
  } catch (e) {
    Date.prototype.toJSON = existingDateToJson
    const jsonDate = this.toJSON()
    Date.prototype.toJSON = patchedDateToJson
    return jsonDate
  }
}

export const frenchDateFormat = new Intl.DateTimeFormat("fr-FR", {
  day: "numeric",
  month: "long",
  weekday: "long",
  year: "numeric",
})
export const frenchDateSameYearShortFormat = new Intl.DateTimeFormat("fr-FR", {
  day: "numeric",
  month: "short",
})
export const frenchTimeFormat = new Intl.DateTimeFormat("fr-FR", {
  hour: "numeric",
  minute: "2-digit",
})

export function dateFromNullableDateOrIsoString(
  date: Date | string | undefined | null,
  defaultValue: Date | undefined | null,
): Date | undefined | null {
  if (date instanceof Date) {
    return date
  }
  if (typeof date === "string") {
    return parseISO(date)
  }
  return defaultValue
}

/// Convert strings like "2017-11-24+01:00" to dates.
export function dateFromYearMonthDayTimezoneString(dateString: string): Date {
  return parse(dateString, "yyyy-MM-ddXXX", new Date())
}

export function displayDateFromLocalIsoString(dateString: string): string {
  const date = parseISO(dateString)
  dateString = format(date, "EEEE d LLLL yyyy H'h'mm", { locale: frLocale })
  if (dateString.endsWith(" 0h00")) {
    dateString = dateString.substring(0, dateString.length - 5)
  }
  return dateString
}

export function formatRelativeFrenchTime(date: Date, now: Date) {
  if (
    date.getFullYear() === now.getFullYear() &&
    date.getMonth() === now.getMonth() &&
    date.getDate() === now.getDate()
  ) {
    return frenchTimeFormat.format(date)
  }
  return frenchDateSameYearShortFormat.format(date)
}

export function frenchDateWithoutDayOfWeekFromDateOrString(
  dateOrString: Date | string,
): string {
  const date: Date =
    dateOrString instanceof Date ? dateOrString : parseISO(dateOrString)
  let dateString = format(date, "d LLLL yyyy H'h'mm", { locale: frLocale })
  if (dateString.endsWith(" 0h00")) {
    dateString = dateString.substring(0, dateString.length - 5)
  }
  return dateString
}

export function frenchTimeFromDateOrString(
  dateOrString: Date | string,
): string {
  const date: Date =
    dateOrString instanceof Date ? dateOrString : parseISO(dateOrString)
  return format(date, "H'h'mm", { locale: frLocale })
}

export function localIso8601StringFromDate(date: Date) {
  return new Date(date.getTime() - date.getTimezoneOffset() * 60000)
    .toISOString()
    .split("T")[0]
}
