import path from "path"
import fs from "fs-extra"
import { getFiles } from "./file_systems"
import { ChildProcess, execSync } from "child_process"

const MAXBUFFER = 50 * 1024 * 1024

export function commit(repositoryDir: string, message: string): boolean {
  execSync("git add .", {
    cwd: repositoryDir,
    env: process.env,
    encoding: "utf-8",
    stdio: ["ignore", "ignore", "pipe"],
    maxBuffer: MAXBUFFER,
  })
  try {
    execSync(`git commit -m "${message}" --quiet`, {
      cwd: repositoryDir,
      env: process.env,
      encoding: "utf-8",
      stdio: ["ignore", "pipe", "pipe"],
    })
    return true
  } catch (childProcess) {
    if (
      (childProcess as ChildProcess).stdout === null ||
      !/nothing to commit/.test((childProcess as any).stdout)
    ) {
      console.error((childProcess as any).output)
      throw childProcess
    }
    return false
  }
}

export function commitAndPush(
  repositoryDir: string,
  message: string,
  remotes?: string[],
): number {
  let exitCode = 0
  if (commit(repositoryDir, message)) {
    for (const remote of remotes || []) {
      try {
        execSync(`git push ${remote} master`, {
          cwd: repositoryDir,
          env: process.env,
          encoding: "utf-8",
          stdio: ["ignore", "ignore", "pipe"],
        })
      } catch (childProcess) {
        // Don't stop when push fails.
        console.error((childProcess as any).output)
        exitCode = (childProcess as any).status
      }
    }
  } else {
    // There was nothing to commit.
    exitCode = 10
  }
  return exitCode
}

export function resetAndPull(gitDir: string): boolean {
  execSync(`git reset --hard origin/master`, {
    cwd: gitDir,
    env: process.env,
    encoding: "utf-8",
    stdio: ["ignore", "ignore", "pipe"],
  })

  execSync(`git pull --rebase`, {
    cwd: gitDir,
    env: process.env,
    encoding: "utf-8",
    stdio: ["ignore", "ignore", "pipe"],
  })

  return true
}

export function clone(
  gitGroupUrl: string | undefined,
  gitName: string,
  workingDir: string,
): void {
  if (gitGroupUrl !== undefined) {
    execSync(`git clone ${gitGroupUrl}/${gitName}.git`, {
      cwd: workingDir,
      env: process.env,
      encoding: "utf-8",
      stdio: ["ignore", "ignore", "pipe"],
    })
  }
}

export function run(
  repositoryDir: string,
  args: string,
  verbose?: boolean,
): string {
  try {
    if (verbose) console.log(`git -C ${repositoryDir} ${args}`)
    const output = execSync(`git ${args}`, {
      cwd: repositoryDir,
      maxBuffer: MAXBUFFER,
    })
      .toString()
      .trim()
    if (verbose) console.log(output)
    return output
  } catch (childProcess) {
    for (const output of ["stdout", "stderr"])
      console.error(`${output}: ${(childProcess as any)[output]}`)
    throw childProcess
  }
}

export function test(
  repositoryDir: string,
  args: string,
  verbose?: boolean,
): boolean {
  try {
    if (verbose) console.log(`git -C ${repositoryDir} ${args}`)
    const output = execSync(`git ${args}`, {
      cwd: repositoryDir,
      stdio: ["ignore", "pipe", "pipe"],
      maxBuffer: MAXBUFFER,
    })
      .toString()
      .trim()
    if (verbose) console.log(output)
    return true
  } catch (childProcess) {
    if ((childProcess as any).status != 0) return false
    throw childProcess
  }
}

export function uidBranchUpdate(
  repositoryDir: string,
  subdirectories: any,
  notHard: boolean,
): boolean {
  const previousBranch = run(repositoryDir, "rev-parse --abbrev-ref HEAD")
  const branch = "uid"
  const cwd = process.cwd()
  process.chdir(repositoryDir)
  for (const subdirectory of subdirectories) {
    const uidDirectory = `${subdirectory}/uid`
    fs.ensureDirSync(uidDirectory)
    for (const pathname of getFiles([`${subdirectory}/**/*.json`])) {
      const filename = path.basename(pathname)
      fs.renameSync(pathname, `${uidDirectory}/${filename}`)
    }
  }
  process.chdir(cwd)
  if (test(repositoryDir, `rev-parse remotes/origin/${branch}`)) {
    // if a remote branch exists, it must be used even if a local branch exists
    run(repositoryDir, `branch --force ${branch} origin/${branch}`)
    run(repositoryDir, `symbolic-ref HEAD refs/heads/${branch}`)
    run(repositoryDir, "reset")
  } else if (test(repositoryDir, `rev-parse refs/heads/${branch}`)) {
    // if no remote branch exists but a local one exists, use it
    run(repositoryDir, `symbolic-ref HEAD refs/heads/${branch}`)
    run(repositoryDir, "reset")
  } else {
    // if there are no remote or local branches, create a new one
    run(repositoryDir, `branch --force ${branch} master`)
    run(repositoryDir, `symbolic-ref HEAD refs/heads/${branch}`)
    run(repositoryDir, "reset")
    run(repositoryDir, "rm -r .")
  }
  for (const subdirectory of subdirectories) {
    const uidDirectory = `${subdirectory}/uid`
    run(repositoryDir, `add ${uidDirectory}`)
  }
  let status = false
  if (!test(repositoryDir, "diff --quiet --cached")) {
    run(repositoryDir, `commit -m 'Update'`)
    status = true
  }
  run(repositoryDir, `symbolic-ref HEAD refs/heads/${previousBranch}`)
  const resetOption = notHard ? "" : "--hard"
  run(repositoryDir, `reset ${resetOption}`)
  return status
}
