/// Load data from files

import assert from "assert"
import fs from "fs-extra"
import path from "path"

export { EnabledDatasets } from "./datasets"
import { datasets, EnabledDatasets, Dataset } from "./datasets"
import type { DocumentFilesIndex } from "./dossiers_legislatifs"
export { walkDir } from "./file_systems"
import { walkDir } from "./file_systems"
import {
  Acteur,
  Convert as ActeursEtOrganesConvert,
  Organe,
  Photo,
} from "./types/acteurs_et_organes"
import { Convert as AgendasConvert, Reunion } from "./types/agendas"
import { Amendement, Convert as AmendementsConvert } from "./types/amendements"
import {
  Convert as DossiersLegislatifsConvert,
  Document,
  DossierParlementaire,
} from "./types/dossiers_legislatifs"
import { Legislature } from "./types/legislatures"
import { Convert as ScrutinsConvert, Scrutin } from "./types/scrutins"
import { Convert as QuestionsConvert, Question } from "./types/questions"
import { Convert as ComptesRendusConvert, CompteRendu } from "./types/debats"
import { DocumentUrlFormat } from "./urls"

export interface Data {
  readonly acteurByUid: { [uid: string]: Acteur }
  readonly amendementByUid: { [uid: string]: Amendement }
  readonly documentByUid: { [uid: string]: Document }
  readonly dossierParlementaireBySenatChemin: {
    [uid: string]: DossierParlementaire
  }
  readonly dossierParlementaireByTitreChemin: {
    [uid: string]: DossierParlementaire
  }
  readonly dossierParlementaireByUid: { [uid: string]: DossierParlementaire }
  readonly organeByUid: { [uid: string]: Organe }
  readonly photoByUid: { [uid: string]: Photo }
  readonly reunionByUid: { [uid: string]: Reunion }
  readonly reunionsByDay: { [day: string]: Reunion[] }
  readonly reunionsByDossierUid: { [dossierUid: string]: Reunion[] }
  readonly scrutinByUid: { [uid: string]: Scrutin }
  readonly questionByUid: { [uid: string]: Question }
  readonly compteRenduByUid: { [uid: string]: CompteRendu }
}

const documentUidRegex =
  /^((?<documentType>ACIN|ALCN|AVCE|AVIS|DECL|ETDI|LETT|MESS|MION|PION|PNRE|PRJL|RAPP|RINF|RION)(?<chambre>AN|CG|SN)(?<republique>R\d+)[LS](?<legislature>\d+)(?<etat>B|BTA|BTC|TAP)(?<numero>\d+)(-.*)?)|((?<documentType2>MION)(?<chambre2>AN|SN)(?<republique2>R\d+)[LS](?<legislature2>\d+)-N(?<numero2>\d+))$/
const dossierUidRegex = /^DL(R\d+)(L\d+)N(\d+)$/
const reunionUidRegex =
  /^RU(AN|CG|SN)(R\d+)(L\d+)([AS]\d{4})(IDC|IDFL?|IDS)(\d+)$/
const scrutinUidRegex = /^VT(AN|CG)(R\d+)(L\d+)V(\d+)$/
const compteRenduUidRegex: RegExp =
  /^CRS(AN|CG)(R\d+)(L\d+)(S\d+)(O|E|D)(\d+)(N\d+)$/
const questionUidRegex: RegExp = /^QAN(R\d+)(L\d+)Q(E|G|OSD)(\d+)$/

type ItemConverter<T> = (json: string) => T

function* iterLoadAssembleeItems<T extends { uid: string }>(
  dataDir: string,
  legislature: Legislature,
  datasets: Dataset[],
  convertToItem: ItemConverter<T>,
  subfolderName: string[],
  fileExtension: string = "json",
  { noValidation = false, log = false } = {},
): Generator<{ item: T; datasetLegislature: Legislature }, void, unknown> {
  const itemUid = new Set<string>()
  for (const dataset of datasets) {
    if (dataset.ignoreForWeb) {
      continue
    }
    if (
      legislature !== undefined &&
      legislature !== Legislature.All &&
      dataset.legislature !== Legislature.All &&
      legislature !== dataset.legislature
    ) {
      continue
    }
    let fileExtensionRegex = new RegExp(`\\.${fileExtension}$`)
    const datasetDir = path.join(
      dataDir,
      dataset.filename.replace(fileExtensionRegex, "_nettoye"),
    )

    for (const itemSplitPath of walkDir(datasetDir, subfolderName)) {
      const itemFilename = itemSplitPath[itemSplitPath.length - 1]
      if (!itemFilename.endsWith(".json")) {
        continue
      }
      const itemFilePath = path.join(datasetDir, ...itemSplitPath)
      if (log) {
        console.log(`Loading file: ${itemFilePath}…`)
      }

      let itemJson: any = fs.readFileSync(itemFilePath, { encoding: "utf8" })
      let item: T = noValidation
        ? JSON.parse(itemJson)
        : convertToItem(itemJson)

      if (itemUid.has(item.uid)) {
        continue
      }
      itemUid.add(item.uid)
      yield { item: item, datasetLegislature: dataset.legislature }

      itemJson = null
    }
  }
}

export function* iterLoadAssembleeActeurs(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { acteur: Acteur; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: acteur,
    datasetLegislature,
  } of iterLoadAssembleeItems<Acteur>(
    dataDir,
    legislature,
    datasets.acteursEtOrganes,
    ActeursEtOrganesConvert.toActeur,
    ["acteurs"],
    "json",
    options,
  )) {
    yield { acteur, datasetLegislature }
  }
}

export function* iterLoadAssembleeOrganes(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { organe: Organe; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: organe,
    datasetLegislature,
  } of iterLoadAssembleeItems<Organe>(
    dataDir,
    legislature,
    datasets.acteursEtOrganes,
    ActeursEtOrganesConvert.toOrgane,
    ["organes"],
    "json",
    options,
  )) {
    yield { organe, datasetLegislature }
  }
}

export function* iterLoadAssembleeAmendements(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { amendement: Amendement; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: amendement,
    datasetLegislature,
  } of iterLoadAssembleeItems<Amendement>(
    dataDir,
    legislature,
    datasets.amendements,
    AmendementsConvert.toAmendement,
    [],
    "json",
    options,
  )) {
    yield { amendement, datasetLegislature }
  }
}

export function* iterLoadAssembleeDocuments(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { document: Document; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: document,
    datasetLegislature,
  } of iterLoadAssembleeItems<Document>(
    dataDir,
    legislature,
    datasets.dossiersLegislatifs,
    DossiersLegislatifsConvert.toDocument,
    ["documents"],
    "json",
    options,
  )) {
    yield { document, datasetLegislature }
  }
}

export function* iterLoadAssembleeDossiersParlementaires(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  {
    dossierParlementaire: DossierParlementaire
    datasetLegislature: Legislature
  },
  void,
  unknown
> {
  for (const {
    item: dossierParlementaire,
    datasetLegislature,
  } of iterLoadAssembleeItems<DossierParlementaire>(
    dataDir,
    legislature,
    datasets.dossiersLegislatifs,
    DossiersLegislatifsConvert.toDossierParlementaire,
    ["dossiers"],
    "json",
    options,
  )) {
    yield { dossierParlementaire, datasetLegislature }
  }
}

export function* iterLoadAssembleeReunions(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { reunion: Reunion; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: reunion,
    datasetLegislature,
  } of iterLoadAssembleeItems<Reunion>(
    dataDir,
    legislature,
    datasets.agendas,
    AgendasConvert.toReunion,
    [],
    "json",
    options,
  )) {
    yield { reunion, datasetLegislature }
  }
}

export function* iterLoadAssembleeScrutins(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { scrutin: Scrutin; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: scrutin,
    datasetLegislature,
  } of iterLoadAssembleeItems<Scrutin>(
    dataDir,
    legislature,
    datasets.scrutins,
    ScrutinsConvert.toScrutin,
    [],
    "json",
    options,
  )) {
    yield { scrutin, datasetLegislature }
  }
}

export function* iterLoadAssembleeQuestions(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { question: Question; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: question,
    datasetLegislature,
  } of iterLoadAssembleeItems<Question>(
    dataDir,
    legislature,
    datasets.questions,
    QuestionsConvert.toQuestion,
    [],
    "json",
    options,
  )) {
    yield { question, datasetLegislature }
  }
}

export function* iterLoadAssembleeComptesRendus(
  dataDir: string,
  legislature: Legislature,
  options = {},
): Generator<
  { compteRendu: CompteRendu; datasetLegislature: Legislature },
  void,
  unknown
> {
  for (const {
    item: compteRendu,
    datasetLegislature,
  } of iterLoadAssembleeItems<CompteRendu>(
    dataDir,
    legislature,
    datasets.comptesRendusSeances,
    ComptesRendusConvert.toCompteRendu,
    [],
    "xml",
    options,
  )) {
    yield { compteRendu, datasetLegislature }
  }
}

/// Load data from split files.
/// Deprecated: Use iterLoad... iterators instead.
export function loadAssembleeData(
  dataDir: string,
  enabledDatasets: EnabledDatasets,
  legislature: Legislature,
  { log = false } = {},
): Data {
  const acteurByUid: { [uid: string]: Acteur } = {}
  const organeByUid: { [uid: string]: Organe } = {}
  if (enabledDatasets & EnabledDatasets.ActeursEtOrganes) {
    for (const dataset of datasets.acteursEtOrganes) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const datasetDir = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye"),
      )

      for (const acteurSplitPath of walkDir(datasetDir, ["acteurs"])) {
        const acteurFilename = acteurSplitPath[acteurSplitPath.length - 1]
        if (!acteurFilename.endsWith(".json")) {
          continue
        }
        const acteurFilePath = path.join(datasetDir, ...acteurSplitPath)
        if (log) {
          console.log(`Loading file: ${acteurFilePath}…`)
        }
        const acteurJson: string = fs.readFileSync(acteurFilePath, {
          encoding: "utf8",
        })
        const acteur = ActeursEtOrganesConvert.toActeur(acteurJson)
        // Keep the first existing "acteur", because the first wrapper is the most complete.
        if (acteurByUid[acteur.uid] === undefined) {
          acteurByUid[acteur.uid] = acteur
        }
      }

      for (const organeSplitPath of walkDir(datasetDir, ["organes"])) {
        const organeFilename = organeSplitPath[organeSplitPath.length - 1]
        if (!organeFilename.endsWith(".json")) {
          continue
        }
        const organeFilePath = path.join(datasetDir, ...organeSplitPath)
        if (log) {
          console.log(`Loading file: ${organeFilePath}…`)
        }
        const organeJson: string = fs.readFileSync(organeFilePath, {
          encoding: "utf8",
        })
        const organe = ActeursEtOrganesConvert.toOrgane(organeJson)
        // Keep the first existing "organe", because the first wrapper is the most complete.
        if (organeByUid[organe.uid] === undefined) {
          organeByUid[organe.uid] = organe
        }
      }
    }
  }

  const reunionByUid: { [uid: string]: Reunion } = {}
  const reunionsByDay: { [day: string]: Reunion[] } = {}
  const reunionsByDossierUid: { [dossierUid: string]: Reunion[] } = {}
  if (enabledDatasets & EnabledDatasets.Agendas) {
    for (const dataset of datasets.agendas) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const datasetDir = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye"),
      )
      for (const reunionSplitPath of walkDir(datasetDir)) {
        const reunionFilename = reunionSplitPath[reunionSplitPath.length - 1]
        if (!reunionFilename.endsWith(".json")) {
          continue
        }
        const reunionFilePath = path.join(datasetDir, ...reunionSplitPath)
        if (log) {
          console.log(`Loading file: ${reunionFilePath}…`)
        }
        const reunionJson: string = fs.readFileSync(reunionFilePath, {
          encoding: "utf8",
        })
        const reunion = AgendasConvert.toReunion(reunionJson)
        if (reunionByUid[reunion.uid] === undefined) {
          reunionByUid[reunion.uid] = reunion
        }

        const day = reunion.timestampDebut.toISOString().split("T")[0]
        let reunionsAtDay = reunionsByDay[day]
        if (reunionsAtDay === undefined) {
          reunionsAtDay = reunionsByDay[day] = []
        }
        reunionsAtDay.push(reunion)

        const odj = reunion.odj
        if (odj !== undefined) {
          for (const pointOdj of odj.pointsOdj || []) {
            for (const dossierLegislatifRef of pointOdj.dossiersLegislatifsRefs ||
              []) {
              let reunionsAtDossierUid =
                reunionsByDossierUid[dossierLegislatifRef]
              if (reunionsAtDossierUid === undefined) {
                reunionsAtDossierUid = reunionsByDossierUid[
                  dossierLegislatifRef
                ] = []
              }
              reunionsAtDossierUid.push(reunion)
            }
          }
        }
      }
    }
  }

  const amendementByUid: { [uid: string]: Amendement } = {}
  if (enabledDatasets & EnabledDatasets.Amendements) {
    for (const dataset of datasets.amendements) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const datasetDir = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye"),
      )

      for (const amendementSplitPath of walkDir(datasetDir)) {
        const amendementFilename =
          amendementSplitPath[amendementSplitPath.length - 1]
        if (!amendementFilename.endsWith(".json")) {
          continue
        }
        const amendementFilePath = path.join(datasetDir, ...amendementSplitPath)
        if (log) {
          console.log(`Loading file: ${amendementFilePath}…`)
        }
        const amendementJson: string = fs.readFileSync(amendementFilePath, {
          encoding: "utf8",
        })
        const amendement = AmendementsConvert.toAmendement(amendementJson)
        if (amendementByUid[amendement.uid] === undefined) {
          amendementByUid[amendement.uid] = amendement
        }
      }
    }
  }

  const documentByUid: { [uid: string]: Document } = {}
  const dossierParlementaireBySenatChemin: {
    [uid: string]: DossierParlementaire
  } = {}
  const dossierParlementaireByTitreChemin: {
    [uid: string]: DossierParlementaire
  } = {}
  const dossierParlementaireByUid: { [uid: string]: DossierParlementaire } = {}
  if (enabledDatasets & EnabledDatasets.DossiersLegislatifs) {
    for (const dataset of datasets.dossiersLegislatifs) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const datasetDir = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye"),
      )

      for (const documentSplitPath of walkDir(datasetDir, ["documents"])) {
        const documentFilename = documentSplitPath[documentSplitPath.length - 1]
        if (!documentFilename.endsWith(".json")) {
          continue
        }
        const documentFilePath = path.join(datasetDir, ...documentSplitPath)
        if (log) {
          console.log(`Loading file: ${documentFilePath}…`)
        }
        const documentJson: string = fs.readFileSync(documentFilePath, {
          encoding: "utf8",
        })
        const document = DossiersLegislatifsConvert.toDocument(documentJson)
        if (documentByUid[document.uid] === undefined) {
          documentByUid[document.uid] = document
        }
      }

      for (const dossierSplitPath of walkDir(datasetDir, ["dossiers"])) {
        const dossierFilename = dossierSplitPath[dossierSplitPath.length - 1]
        if (!dossierFilename.endsWith(".json")) {
          continue
        }
        const dossierFilePath = path.join(datasetDir, ...dossierSplitPath)
        if (log) {
          console.log(`Loading file: ${dossierFilePath}…`)
        }
        const dossierJson: string = fs.readFileSync(dossierFilePath, {
          encoding: "utf8",
        })
        const dossier =
          DossiersLegislatifsConvert.toDossierParlementaire(dossierJson)
        const titreDossier = dossier.titreDossier
        if (
          titreDossier.senatChemin &&
          dossierParlementaireBySenatChemin[titreDossier.senatChemin] ===
            undefined
        ) {
          dossierParlementaireBySenatChemin[titreDossier.senatChemin] = dossier
        }
        if (
          titreDossier.titreChemin &&
          dossierParlementaireByTitreChemin[titreDossier.titreChemin] ===
            undefined
        ) {
          dossierParlementaireByTitreChemin[titreDossier.titreChemin] = dossier
        }
        if (dossierParlementaireByUid[dossier.uid] === undefined) {
          dossierParlementaireByUid[dossier.uid] = dossier
        }
      }
    }
  }

  let photoByUid: { [uid: string]: Photo } = {}
  if (enabledDatasets & EnabledDatasets.Photos) {
    {
      const jsonFilePath: string = path.join(
        dataDir,
        `photos_deputes_${
          // TODO: Handle Legislature.All.
          legislature === Legislature.All ? Legislature.DixSept : legislature
        }`,
        "deputes.json",
      )
      if (fs.existsSync(jsonFilePath)) {
        if (log) {
          console.log(`Loading file: ${jsonFilePath}`)
        }
        const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
        Object.assign(photoByUid, JSON.parse(json))
      } else if (log) {
        console.log(`Ignoring missing file: ${jsonFilePath}`)
      }
    }

    {
      const jsonFilePath: string = path.join(
        dataDir,
        "photos_senateurs",
        "senateurs.json",
      )
      if (fs.existsSync(jsonFilePath)) {
        if (log) {
          console.log(`Loading file: ${jsonFilePath}`)
        }
        const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
        Object.assign(photoByUid, JSON.parse(json))
      } else if (log) {
        console.log(`Ignoring missing file: ${jsonFilePath}`)
      }
    }
  }

  const scrutinByUid: { [uid: string]: Scrutin } = {}
  if (enabledDatasets & EnabledDatasets.Scrutins) {
    for (const dataset of datasets.scrutins) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const datasetDir = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye"),
      )

      for (const scrutinSplitPath of walkDir(datasetDir)) {
        const scrutinFilename = scrutinSplitPath[scrutinSplitPath.length - 1]
        if (!scrutinFilename.endsWith(".json")) {
          continue
        }
        const scrutinFilePath = path.join(datasetDir, ...scrutinSplitPath)
        if (log) {
          console.log(`Loading file: ${scrutinFilePath}…`)
        }
        const scrutinJson: string = fs.readFileSync(scrutinFilePath, {
          encoding: "utf8",
        })
        const scrutin = ScrutinsConvert.toScrutin(scrutinJson)
        if (scrutinByUid[scrutin.uid] === undefined) {
          scrutinByUid[scrutin.uid] = scrutin
        }
      }
    }
  }

  const questionByUid: { [uid: string]: Question } = {}
  if (enabledDatasets & EnabledDatasets.Questions) {
    for (const dataset of datasets.questions) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const datasetDir = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye"),
      )

      for (const questionSplitPath of walkDir(datasetDir)) {
        const questionFilename = questionSplitPath[questionSplitPath.length - 1]
        if (!questionFilename.endsWith(".json")) {
          continue
        }
        const questionFilePath = path.join(datasetDir, ...questionSplitPath)
        if (log) {
          console.log(`Loading file: ${questionFilePath}…`)
        }
        const questionJson: string = fs.readFileSync(questionFilePath, {
          encoding: "utf8",
        })
        const question = QuestionsConvert.toQuestion(questionJson)
        if (questionByUid[question.uid] === undefined) {
          questionByUid[question.uid] = question
        }
      }
    }
  }

  const compteRenduByUid: { [uid: string]: CompteRendu } = {}
  if (enabledDatasets & EnabledDatasets.ComptesRendusSeances) {
    for (const dataset of datasets.comptesRendusSeances) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const datasetDir = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye"),
      )

      for (const compteRenduPath of walkDir(datasetDir)) {
        const compteRenduFilename = compteRenduPath[compteRenduPath.length - 1]
        if (!compteRenduFilename.endsWith(".json")) {
          continue
        }
        const compteRenduFilePath = path.join(datasetDir, ...compteRenduPath)
        if (log) {
          console.log(`Loading file: ${compteRenduFilePath}…`)
        }
        const compteRenduJson: string = fs.readFileSync(compteRenduFilePath, {
          encoding: "utf8",
        })
        const compteRendu = ComptesRendusConvert.toCompteRendu(compteRenduJson)
        if (compteRenduByUid[compteRendu.uid] === undefined) {
          compteRenduByUid[compteRendu.uid] = compteRendu
        }
      }
    }
  }

  return {
    acteurByUid,
    amendementByUid,
    documentByUid,
    dossierParlementaireBySenatChemin,
    dossierParlementaireByTitreChemin,
    dossierParlementaireByUid,
    organeByUid,
    photoByUid,
    reunionByUid,
    reunionsByDay,
    reunionsByDossierUid,
    scrutinByUid,
    questionByUid,
    compteRenduByUid,
  }
}

/// Load data from big monolytic files.
/// Deprecated: Assemblée nationale open data no more uses monolytic files.
export function loadAssembleeDataFromBigFiles(
  dataDir: string,
  enabledDatasets: EnabledDatasets,
  legislature: Legislature,
  { log = false } = {},
): Data {
  const acteurByUid: { [uid: string]: Acteur } = {}
  const organeByUid: { [uid: string]: Organe } = {}
  if (enabledDatasets & EnabledDatasets.ActeursEtOrganes) {
    for (const dataset of datasets.acteursEtOrganes) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const jsonFilePath: string = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye.json"),
      )
      if (log) {
        console.log(`Loading file: ${jsonFilePath}`)
      }
      const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
      const acteursEtOrganes = ActeursEtOrganesConvert.toActeursEtOrganes(json)
      for (const acteur of acteursEtOrganes.acteurs) {
        // Keep the first existing "acteur", because the first wrapper is the most complete.
        if (acteurByUid[acteur.uid] === undefined) {
          acteurByUid[acteur.uid] = acteur
        }
      }
      for (const organe of acteursEtOrganes.organes) {
        // Keep the first existing "organe", because the first wrapper is the most complete.
        if (organeByUid[organe.uid] === undefined) {
          organeByUid[organe.uid] = organe
        }
      }
    }
  }

  const reunionByUid: { [uid: string]: Reunion } = {}
  const reunionsByDay: { [day: string]: Reunion[] } = {}
  const reunionsByDossierUid: { [dossierUid: string]: Reunion[] } = {}
  if (enabledDatasets & EnabledDatasets.Agendas) {
    for (const dataset of datasets.agendas) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const jsonFilePath: string = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye.json"),
      )
      if (log) {
        console.log(`Loading file: ${jsonFilePath}`)
      }
      const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
      const agendas = AgendasConvert.toAgendas(json)
      for (const reunion of agendas.reunions) {
        if (reunionByUid[reunion.uid] === undefined) {
          reunionByUid[reunion.uid] = reunion
        }

        const day = reunion.timestampDebut.toISOString().split("T")[0]
        let reunionsAtDay = reunionsByDay[day]
        if (reunionsAtDay === undefined) {
          reunionsAtDay = reunionsByDay[day] = []
        }
        reunionsAtDay.push(reunion)

        const odj = reunion.odj
        if (odj !== undefined) {
          for (const pointOdj of odj.pointsOdj || []) {
            for (const dossierLegislatifRef of pointOdj.dossiersLegislatifsRefs ||
              []) {
              let reunionsAtDossierUid =
                reunionsByDossierUid[dossierLegislatifRef]
              if (reunionsAtDossierUid === undefined) {
                reunionsAtDossierUid = reunionsByDossierUid[
                  dossierLegislatifRef
                ] = []
              }
              reunionsAtDossierUid.push(reunion)
            }
          }
        }
      }
    }
  }

  const amendementByUid: { [uid: string]: Amendement } = {}
  if (enabledDatasets & EnabledDatasets.Amendements) {
    for (const dataset of datasets.amendements) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const jsonFilePath: string = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye.json"),
      )
      if (log) {
        console.log(`Loading file: ${jsonFilePath}`)
      }
      const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
      const amendements = AmendementsConvert.toAmendements(json)
      for (const texteLegislatif of amendements.textesLegislatifs) {
        for (const amendement of texteLegislatif.amendements) {
          if (amendementByUid[amendement.uid] === undefined) {
            amendementByUid[amendement.uid] = amendement
          }
        }
      }
    }
  }

  const documentByUid: { [uid: string]: Document } = {}
  const dossierParlementaireBySenatChemin: {
    [uid: string]: DossierParlementaire
  } = {}
  const dossierParlementaireByTitreChemin: {
    [uid: string]: DossierParlementaire
  } = {}
  const dossierParlementaireByUid: { [uid: string]: DossierParlementaire } = {}
  if (enabledDatasets & EnabledDatasets.DossiersLegislatifs) {
    for (const dataset of datasets.dossiersLegislatifs) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const jsonFilePath: string = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye.json"),
      )
      if (log) {
        console.log(`Loading file: ${jsonFilePath}`)
      }
      const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
      const dossiersLegislatifs =
        DossiersLegislatifsConvert.toDossiersLegislatifs(json)
      for (const document of dossiersLegislatifs.textesLegislatifs) {
        if (documentByUid[document.uid] === undefined) {
          documentByUid[document.uid] = document
        }
      }
      for (const dossier of dossiersLegislatifs.dossiersParlementaires) {
        const titreDossier = dossier.titreDossier
        if (
          titreDossier.senatChemin &&
          dossierParlementaireBySenatChemin[titreDossier.senatChemin] ===
            undefined
        ) {
          dossierParlementaireBySenatChemin[titreDossier.senatChemin] = dossier
        }
        if (
          titreDossier.titreChemin &&
          dossierParlementaireByTitreChemin[titreDossier.titreChemin] ===
            undefined
        ) {
          dossierParlementaireByTitreChemin[titreDossier.titreChemin] = dossier
        }
        if (dossierParlementaireByUid[dossier.uid] === undefined) {
          dossierParlementaireByUid[dossier.uid] = dossier
        }
      }
    }
  }

  const photoByUid: { [uid: string]: Photo } = {}
  if (enabledDatasets & EnabledDatasets.Photos) {
    const photoByUid = {}

    {
      const jsonFilePath: string = path.join(
        dataDir,
        `photos_deputes_${
          // TODO: Handle Legislature.All.
          legislature === Legislature.All ? Legislature.DixSept : legislature
        }`,
        "deputes.json",
      )
      if (fs.existsSync(jsonFilePath)) {
        if (log) {
          console.log(`Loading file: ${jsonFilePath}`)
        }
        const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
        Object.assign(photoByUid, JSON.parse(json))
      } else if (log) {
        console.log(`Ignoring missing file: ${jsonFilePath}`)
      }
    }

    {
      const jsonFilePath: string = path.join(
        dataDir,
        "photos_senateurs",
        "senateurs.json",
      )
      if (fs.existsSync(jsonFilePath)) {
        if (log) {
          console.log(`Loading file: ${jsonFilePath}`)
        }
        const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
        Object.assign(photoByUid, JSON.parse(json))
      } else if (log) {
        console.log(`Ignoring missing file: ${jsonFilePath}`)
      }
    }
  }

  const scrutinByUid: { [uid: string]: Scrutin } = {}
  if (enabledDatasets & EnabledDatasets.Scrutins) {
    for (const dataset of datasets.scrutins) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const jsonFilePath: string = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye.json"),
      )
      if (log) {
        console.log(`Loading file: ${jsonFilePath}`)
      }
      const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
      const scrutins = ScrutinsConvert.toScrutins(json)
      for (const scrutin of scrutins.scrutins) {
        if (scrutinByUid[scrutin.uid] === undefined) {
          scrutinByUid[scrutin.uid] = scrutin
        }
      }
    }
  }

  const questionByUid: { [uid: string]: Question } = {}
  if (enabledDatasets & EnabledDatasets.Questions) {
    for (const dataset of datasets.questions) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const jsonFilePath: string = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye.json"),
      )
      if (log) {
        console.log(`Loading file: ${jsonFilePath}`)
      }
      const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
      const questions = QuestionsConvert.toQuestions(json)
      for (const question of questions.questions) {
        if (questionByUid[question.uid] === undefined) {
          questionByUid[question.uid] = question
        }
      }
    }
  }

  const compteRenduByUid: { [uid: string]: CompteRendu } = {}
  if (enabledDatasets & EnabledDatasets.ComptesRendusSeances) {
    for (const dataset of datasets.comptesRendusSeances) {
      if (dataset.ignoreForWeb) {
        continue
      }
      if (
        legislature !== Legislature.All &&
        dataset.legislature !== Legislature.All &&
        legislature !== dataset.legislature
      ) {
        continue
      }
      const jsonFilePath: string = path.join(
        dataDir,
        dataset.filename.replace(/\.json$/, "_nettoye.json"),
      )
      if (log) {
        console.log(`Loading file: ${jsonFilePath}`)
      }
      const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
      const comptesRendus = ComptesRendusConvert.toComptesRendus(json)
      for (const compteRendu of comptesRendus.comptesRendus) {
        if (compteRenduByUid[compteRendu.uid] === undefined) {
          compteRenduByUid[compteRendu.uid] = compteRendu
        }
      }
    }
  }

  return {
    acteurByUid,
    amendementByUid,
    documentByUid,
    dossierParlementaireBySenatChemin,
    dossierParlementaireByTitreChemin,
    dossierParlementaireByUid,
    organeByUid,
    photoByUid,
    reunionByUid,
    reunionsByDay,
    reunionsByDossierUid,
    scrutinByUid,
    questionByUid,
    compteRenduByUid,
  }
}

/**
 * Read infos of a document or a division of a document from file system
 *
 * @param dataDir
 * @param uid UID of a document or a division
 * @param formats requested formats of the document in decreasing preference order
 */
export function loadAssembleeDocumentFile(
  dataDir: string,
  uid: string,
  formats: DocumentUrlFormat[],
):
  | {
      format: DocumentUrlFormat
      path: string
      url: string
    }
  | undefined {
  const documentsDir = path.join(dataDir, "Documents")
  const documentDir = pathFromDocumentUid(documentsDir, uid)
  const indexPath = path.join(documentDir, "index.json")
  if (!fs.pathExistsSync(indexPath)) {
    return undefined
  }
  const index = fs.readJsonSync(indexPath) as DocumentFilesIndex
  for (const format of formats) {
    for (const { filename, url } of index[format] ?? []) {
      if (filename === undefined) {
        continue
      }
      return { format, path: path.join(documentDir, filename), url }
    }
  }
  return undefined
}

export function loadAssembleePhotosDeputes(
  dataDir: string,
  legislature: Legislature,
  { log = false } = {},
): { [uid: string]: Photo } {
  const jsonFilePath: string = path.join(
    dataDir,
    `photos_deputes_${
      // TODO: Handle Legislature.All.
      legislature === Legislature.All ? Legislature.DixSept : legislature
    }`,
    "deputes.json",
  )
  if (log) {
    console.log(`Loading file: ${jsonFilePath}`)
  }
  const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
  return JSON.parse(json) as { [uid: string]: Photo }
}

export function loadAssembleePhotosSenateurs(
  dataDir: string,
  { log = false } = {},
): { [uid: string]: Photo } {
  const jsonFilePath: string = path.join(
    dataDir,
    "photos_senateurs",
    "senateurs.json",
  )
  if (log) {
    console.log(`Loading file: ${jsonFilePath}`)
  }
  const json: string = fs.readFileSync(jsonFilePath, { encoding: "utf8" })
  return JSON.parse(json) as { [uid: string]: Photo }
}

export function pathFromDocumentUid(documentsDir: string, uid: string): string {
  const match = uid.match(documentUidRegex)
  assert.notStrictEqual(
    match,
    null,
    `Unexpected structure for document UID: ${uid}`,
  )
  const groups = match!.groups as { [key: string]: string }
  const documentType = path.join(
    documentsDir,
    groups.documentType ?? groups.documentType2,
  )
  const chambreDir = path.join(documentType, groups.chambre ?? groups.chambre2)
  const republiqueDir = path.join(
    chambreDir,
    groups.republique ?? groups.republique2,
  )
  const legislatureDir = path.join(
    republiqueDir,
    groups.legislature ?? groups.legislature2,
  )
  const etat = groups.etat
  if (etat === undefined) {
    fs.ensureDirSync(legislatureDir)
    return path.join(legislatureDir, uid)
  }
  const etatDir = path.join(legislatureDir, etat)
  let numero = "000000" + (groups.numero ?? groups.numero2)
  numero = numero.substring(numero.length - 6)
  const numeroMilliersDir = path.join(etatDir, numero.substring(0, 3))
  fs.ensureDirSync(numeroMilliersDir)
  return path.join(numeroMilliersDir, uid)
}

export function pathFromDossierParlementaireUid(
  dossiersDir: string,
  uid: string,
): string {
  return pathFromUid(dossiersDir, uid, dossierUidRegex)
}

export function pathFromReunionUid(reunionsDir: string, uid: string): string {
  const match = reunionUidRegex.exec(uid)
  assert.notStrictEqual(
    match,
    null,
    `Unexpected structure for reunion UID: ${uid}`,
  )
  const chambreDir = path.join(reunionsDir, match![1])
  const republiqueDir = path.join(chambreDir, match![2])
  const legislatureDir = path.join(republiqueDir, match![3])
  const sessionDir = path.join(legislatureDir, match![4])
  const idDir = path.join(sessionDir, match![5])
  let numero = "000000000" + match![6]
  numero = numero.substring(numero.length - 9)
  const numeroMillionsDir = path.join(idDir, numero.substring(0, 3))
  const numeroMilliersDir = path.join(numeroMillionsDir, numero.substring(3, 3))
  fs.ensureDirSync(numeroMilliersDir)
  return path.join(numeroMilliersDir, uid)
}

export function pathFromScrutinUid(scrutinsDir: string, uid: string): string {
  const match = scrutinUidRegex.exec(uid)
  assert.notStrictEqual(
    match,
    null,
    `Unexpected structure for scrutin UID: ${uid}`,
  )
  const chambreDir = path.join(scrutinsDir, match![1])
  const republiqueDir = path.join(chambreDir, match![2])
  const legislatureDir = path.join(republiqueDir, match![3])
  let numero = "000000" + match![4]
  numero = numero.substring(numero.length - 6)
  const numeroMilliersDir = path.join(legislatureDir, numero.substring(0, 3))
  fs.ensureDirSync(numeroMilliersDir)
  return path.join(numeroMilliersDir, uid)
}

export function pathFromQuestionUid(dataDir: string, uid: string): string {
  return pathFromUid(dataDir, uid, questionUidRegex)
}

export function pathFromCompteRenduUid(dataDir: string, uid: string): string {
  return pathFromUid(dataDir, uid, compteRenduUidRegex)
}

// Get standard path for object with uid : republique > legislature > group objects by thousands
function pathFromUid(dataDir: string, uid: string, dataRegexp: RegExp): string {
  const match = dataRegexp.exec(uid)
  assert.notStrictEqual(
    match,
    null,
    `Unexpected structure for object with UID: ${uid}`,
  )
  const republiqueDir = path.join(dataDir, match![1])
  const legislatureDir = path.join(republiqueDir, match![2])
  let numero = "000000" + match![3]
  numero = numero.substring(numero.length - 6)
  const numeroMilliersDir = path.join(legislatureDir, numero.substring(0, 3))
  fs.ensureDirSync(numeroMilliersDir)
  return path.join(numeroMilliersDir, uid)
}
