import assert from "assert"
import { formatISO } from "date-fns/formatISO"
import { fr as frLocale } from "date-fns/locale/fr"
import { parse as parseDate } from "date-fns/parse"
import deepEqual from "deep-equal"
import fs from "fs-extra"
import frontMatter from "front-matter"
import jsYaml from "js-yaml"
import { parse as parseHtml } from "node-html-parser"
import path from "path"

import { cleanAmendement } from "../cleaners"
import { Amendement as RawAmendement } from "../raw_types/amendements"
import { Amendement, Convert as AmendementsConvert } from "../types/amendements"

export async function* iterRechercheAmendements(
  amendementsSearchCacheDir: string,
  url: string,
  incremental: boolean,
  minDateDepot?: string | null,
  minDateExamen?: string | null,
  options: { silent?: boolean; verbose?: boolean } = {},
): AsyncGenerator<[string, Amendement], void, unknown> {
  for await (const [
    amendementUrlPath,
    amendement,
  ] of iterRechercheRawAmendements(
    amendementsSearchCacheDir,
    url,
    incremental,
    minDateDepot,
    minDateExamen,
    options,
  )) {
    cleanAmendement(amendement)
    // Double conversion, to convert strings to dates.
    yield [
      amendementUrlPath,
      AmendementsConvert.toAmendement(JSON.stringify(amendement)),
    ]
  }
}

export async function* iterRechercheRawAmendements(
  amendementsSearchCacheDir: string,
  url: string,
  incremental: boolean,
  minDateDepot?: string | null,
  minDateExamen?: string | null,
  options: { silent?: boolean; verbose?: boolean } = {},
): AsyncGenerator<[string, RawAmendement], void, unknown> {
  await fs.ensureDir(amendementsSearchCacheDir)

  assert.strictEqual(
    /[?&]date_depot=/.exec(url),
    null,
    `URL ${url} already contains a deposit date`,
  )
  assert.strictEqual(
    /[?&]order=/.exec(url),
    null,
    `URL ${url} already contains a sort order`,
  )
  assert.strictEqual(
    /[?&]page=/.exec(url),
    null,
    `URL ${url} already contains a page number`,
  )
  if (incremental) {
    assert.notStrictEqual(
      /[?&]etat=/.exec(url),
      null,
      `In incremental mode, URL ${url} must contain an "etat" query parameter`,
    )
  }
  url += (url.includes("?") ? "&" : "?") + "order=date_depot,desc&page=1"

  if (minDateDepot != null) {
    assert.notStrictEqual(
      /\d{4}-\d{2}-\d{2}/.exec(minDateDepot),
      null,
      `Invalid format for minimum date: ${minDateDepot}`,
    )
  }

  let amendementNumber = 0
  iterSearchPages: while (true) {
    if (options.verbose) {
      console.log(`Fetching amendements search page at ${url}…`)
    }
    const response = await fetch(url)
    assert(
      response.ok,
      `Retrieval of search page at ${url} failed with error: ${response.status} ${response.statusText}`,
    )
    const page = await response.text()
    const html = parseHtml(page)
    if (html.querySelector("div.no-result") !== null) {
      break
    }
    const amendementsCountDiv = html.querySelector("div.mirror-card-subtitle")
    assert.notStrictEqual(amendementsCountDiv, null)
    const amendementsCount = parseInt(/\d+/.exec(amendementsCountDiv!.text)![0])
    const amendementsDiv = html.querySelector(
      "div.amendement-list--results-table",
    )
    assert.notStrictEqual(amendementsDiv, null)
    const amendementsTable = amendementsDiv!.querySelector("table")
    assert.notStrictEqual(amendementsTable, null)
    const tfootTr = amendementsTable!.querySelector("tfoot > tr")
    assert.notStrictEqual(tfootTr, null)
    const tfootTrCells = tfootTr!
      .querySelectorAll("td, th")
      .map((cell) => cell.text)
    assert(
      deepEqual(tfootTrCells, [
        "",
        "n°",
        "Dossier législatif",
        "Emplacement",
        "Auteur",
        "État",
        "Sort",
        "Date d'examen",
        "Examiné par",
        "Texte visé",
        "Date de dépôt",
      ]),
      `Unexpected columns in ${JSON.stringify(tfootTrCells, null, 2)}`,
    )
    const amendementsTr = amendementsTable!.querySelectorAll("tbody > tr")
    assert.notStrictEqual(amendementsTr.length, 0)
    for (const amendementTr of amendementsTr) {
      amendementNumber++

      if (minDateDepot != null || minDateExamen != null) {
        const amendementTdList = amendementTr.querySelectorAll("td")
        assert.strictEqual(
          amendementTdList.length,
          11,
          `Unexpected number of columns in amendment row: ${amendementTr.outerHTML}`,
        )
        if (minDateDepot != null) {
          const dateDepotTd = amendementTdList[10]
          const dateDepot = formatISO(
            parseDate(dateDepotTd.text, "d MMMM y", new Date(), {
              locale: frLocale,
            }),
            { representation: "date" },
          )
          if (dateDepot < minDateDepot) {
            // Since the amendments are sorted by deposit date, once an amendment
            // with an old deposit date is encountered, the full search can be stopped.
            break iterSearchPages
          }
        }
        if (minDateExamen != null) {
          const dateExamenTd = amendementTdList[7]
          if (dateExamenTd.text) {
            const dateExamen = formatISO(
              parseDate(dateExamenTd.text, "d MMMM y", new Date(), {
                locale: frLocale,
              }),
              { representation: "date" },
            )
            if (dateExamen < minDateExamen) {
              // This amendment has been discussed since a few days.
              // => Ignore it, because it should be in Assemblée open data.
              continue
            }
          }
        }
      }

      const amendementUrlPath = amendementTr.getAttribute("data-href")
      assert.notStrictEqual(amendementUrlPath, undefined)
      const amendementUrlPathSplitted = amendementUrlPath!.split("/")
      assert.strictEqual(
        amendementUrlPathSplitted[0],
        "",
        `Unexpected URL path for amendement: ${amendementUrlPath}`,
      )
      assert.strictEqual(
        amendementUrlPathSplitted[1],
        "dyn",
        `Unexpected URL path for amendement: ${amendementUrlPath}`,
      )
      const amendementSearchCacheFilePath =
        path.join(
          amendementsSearchCacheDir,
          ...amendementUrlPathSplitted.slice(2),
        ) + ".html"
      await fs.ensureDir(path.dirname(amendementSearchCacheFilePath))
      const existingAmendementSearchCache = (await fs.pathExists(
        amendementSearchCacheFilePath,
      ))
        ? await fs.readFile(amendementSearchCacheFilePath, "utf8")
        : null
      const amendementTrOuterHtml = amendementTr.outerHTML
      const amendementSearchCache = `---\n${jsYaml.dump(
        {
          position: amendementsCount - amendementNumber,
          search: url.replace(/&page=[\d]+/, ""),
        },
        {
          sortKeys: true,
        },
      )}---\n\n${amendementTrOuterHtml}`
      if (amendementSearchCache === existingAmendementSearchCache) {
        if (incremental) {
          // The same <tr> is already in cache and in the same position.
          // Assume that following amendments are not changed and stop looking
          // for amendments to upsert.
          break iterSearchPages
        }
        continue
      }
      if (
        existingAmendementSearchCache !== null &&
        frontMatter(existingAmendementSearchCache).body ===
          amendementTrOuterHtml
      ) {
        // The same <tr> is already in cache, but in a different position.
        // Assume the amendment is unchanged.
        continue
      }
      if (options.verbose) {
        if (existingAmendementSearchCache === null) {
          console.log(
            `Adding amendement search cache: ${amendementSearchCacheFilePath}…`,
          )
        } else {
          console.log(
            `Updating amendement search cache: ${amendementSearchCacheFilePath}…`,
          )
        }
      }
      const amendementHtmlUrl = new URL(
        amendementUrlPath!,
        "https://www.assemblee-nationale.fr/",
      ).toString()
      const amendementHtmlResponse = await fetch(amendementHtmlUrl)
      if (amendementHtmlResponse.status === 404) {
        console.log(
          `Amendement HTML page not found at ${amendementHtmlUrl}. Skipping.`,
        )
        continue
      }
      assert(
        amendementHtmlResponse.ok,
        `Retrieval of amendement HTML page at ${amendementHtmlUrl} failed with error: ${amendementHtmlResponse.status} ${amendementHtmlResponse.statusText}`,
      )
      const amendementHtmlPage = await amendementHtmlResponse.text()
      const amendementHtml = parseHtml(amendementHtmlPage)
      const formatsLi = amendementHtml.querySelectorAll(
        "li.mirror-card-header--options--content--item",
      )
      assert.notStrictEqual(formatsLi.length, 0)
      const jsonLi = formatsLi.find((formatLi) => {
        const formatSpan = formatLi.querySelector("a > span")
        assert.notStrictEqual(
          formatSpan,
          null,
          `No <span> in <a> found in ${formatLi.toString()}`,
        )
        return formatSpan!.text === "Version JSON"
      })
      assert.notStrictEqual(
        jsonLi,
        undefined,
        `No JSON version found for amendement at ${amendementHtmlUrl}`,
      )
      const jsonA = jsonLi!.querySelector("a")
      assert.notStrictEqual(
        jsonA,
        null,
        `No <a> found in ${jsonLi!.toString()}`,
      )
      const amendementJsonUrlPath = jsonA!.getAttribute("href")
      assert.notStrictEqual(
        amendementJsonUrlPath,
        undefined,
        `No URL found for JSON version of amendement: ${jsonLi!.toString()}`,
      )
      const amendementJsonUrl = new URL(
        amendementJsonUrlPath!,
        "https://www.assemblee-nationale.fr/",
      ).toString()
      const amendementJsonResponse = await fetch(amendementJsonUrl)
      assert(
        amendementJsonResponse.ok,
        `Retrieval of amendement JSON page at ${amendementJsonUrl} failed with error: ${amendementJsonResponse.status} ${amendementJsonResponse.statusText}`,
      )
      const amendement = await amendementJsonResponse.json()
      yield [amendementUrlPath!, amendement]

      // SearchCache must be written at the end to ensure that everything is in a consistent state,
      // even if script fails abruptly.
      await fs.writeFile(
        amendementSearchCacheFilePath,
        amendementSearchCache,
        "utf8",
      )
    }

    const paginationA = html.querySelector(
      "div.an-pagination--item > i.an-icons-chevron-right ~ a",
    )
    if (paginationA === null) {
      break
    }
    const urlPath = paginationA.getAttribute("href")
    assert.notStrictEqual(urlPath, undefined)
    url = new URL(urlPath!, "https://www.assemblee-nationale.fr/").toString()
  }
}
