import assert from "assert"
import { JSDOM } from "jsdom"

function parseHeader(header: string): [number | null, string] {
  const headersMapping: { regex: RegExp; level: number; name: string }[] = [
    { regex: /^(RAPPORT_)?ANNEXE(_|$)|^ETAT_/, level: 0, name: "Annexe" },
    { regex: /^TOME_/, level: 1, name: "Tome" },
    {
      regex:
        /^PARTIE_|^(PREMIERE|SECONDE|DEUXIEME|TROISIEME|QUATRIEME)_PARTIE(_|$)/,
      level: 2,
      name: "Partie",
    },
    { regex: /^LIVRE_/, level: 3, name: "Livre" },
    { regex: /^TITRE_/, level: 4, name: "Titre" },
    { regex: /^SOUSTITRE_/, level: 5, name: "SousTitre" },
    { regex: /^CHAPITRE_/, level: 6, name: "Chapitre" },
    { regex: /^SECTION_/, level: 7, name: "Section" },
    { regex: /^SOUSSECTION_/, level: 8, name: "SousSection" },
    { regex: /^ARTICLES?_/, level: 9, name: "Article" },
    { regex: /^EXPOSE_DES_MOTIFS$/, level: 9, name: "ExposeMotifs" },
  ]

  for (let { regex, level, name } of headersMapping) {
    if (header.match(regex)) {
      return [level, name]
    }
  }
  return [null, ""]
}

function cleanText(text: string | null): string {
  return text?.replace(/[\n\t]+/g, "").trim() || ""
}

const excludedAlineas = [
  /^Délibéré en séance publique/,
  /^Fait le/,
  /^Le Président,$/,
  /^Signé/,
]

export function parseTexte(assembleeUrl: string, page: string) {
  // Repair HTML.
  let html = page.replace(/(<style[\w\W]+style>)/g, "")

  // Extract subdivisions from HTML.
  const { window } = new JSDOM(html)
  const { document } = window
  assert.strictEqual(document.children.length, 1)
  const htmlElement = document.children[0]
  assert.strictEqual(htmlElement.children.length, 2)
  const bodyElement = htmlElement.children[1]

  if (bodyElement.children.length < 3) {
    // Occurs in http://www.assemblee-nationale.fr/15/textes/0326.asp.
    window.close() // Free memory.
    return {
      error: { code: -1, message: "Texte de loi sans contenu" },
      html,
      page,
    }
  }
  assert.strictEqual(bodyElement.children[0].tagName, "DIV")
  // First child is a DIV describing the document (Assemblée's header). Skip it for now.
  let bodyChild = bodyElement.children[1]

  assert.strictEqual(bodyChild.tagName, "BR")
  let alineaElement: Element | null = null
  let isMultiLinesHeader: boolean = false
  let level: number | null = null
  let levels: number[] = []
  let nextParentState = null
  let state: string | null = "nextBodyChild"
  const subdivisions = []
  let subdivisionAlineas: object[] | null = null
  let subdivisionHeaders: object[] | null = null
  while (state !== null) {
    switch (state) {
      case "nextBodyChild":
        if (bodyChild.nextElementSibling === null) {
          // The document has been fully parsed.
          state = null
        } else {
          bodyChild = bodyChild.nextElementSibling
          if (bodyChild.tagName === "DIV") {
            alineaElement = bodyChild.children[0]

            if (alineaElement === undefined || bodyChild.id.includes("ftn")) {
              // No alinea in current bodyChild: go to next bodyChild.
              state = "nextBodyChild"
            } else {
              state = "alineaElement"
            }
          } else {
            // <P/> is for footnotes
            assert(
              ["BR", "HR", "P"].includes(bodyChild.tagName),
              `Unexpected tag name "${bodyChild.tagName}" for body child`,
            )
            // Stay in the same state to go to next bodyChild.
          }
        }
        break
      case "alineaElement":
        switch (alineaElement!.tagName) {
          case "DIV":
            state = "nextAlineaElement"
            break
          case "H4":
            // TODO
            nextParentState = "nextAlineaElement"
            state = "firstParagraphChild"
            break
          case "P":
            nextParentState = "nextAlineaElement"
            state = "firstParagraphChild"
            break
          case "OL":
            // TODO
            state = "nextAlineaElement"
            break
          case "TABLE":
            nextParentState = "nextAlineaElement"
            state = "firstParagraphChild"
            break
          default:
            return {
              error: {
                code: -2,
                message: `Unexpected tag name for alinea element: ${
                  alineaElement!.tagName
                }`,
              },
              html,
              page,
            }
        }
        break
      case "firstParagraphChild":
        const headerText = alineaElement!.textContent
        const nameComputed = (headerText || "")
          .normalize("NFD")
          .replace(/[\u0300-\u036f]/g, "")
          .replace(/\(nouveau\)/, "")
          .replace(/\(Pour coordination\)/, "")
          .replace(/\(Supprimés?\)/, "")
          .replace(/ /g, " ")
          .replace(/[\-,.…]/g, "")
          .trim()
          .replace(/ {1,}/g, "_")

        const nameUpper = nameComputed.toUpperCase()
        const [nextLevel, paragraphType] = parseHeader(nameUpper)

        if (
          !nameUpper ||
          nameUpper.match(/^(PROJET|PROPOSITION)_DE_LOI(_|$)/) !== null
        ) {
          // Occurs in:
          // * http://www.assemblee-nationale.fr/15/textes/0232.asp
          // * http://www.assemblee-nationale.fr/15/textes/0626.asp
          // * http://www.assemblee-nationale.fr/15/textes/0676.asp
          state = nextParentState
          break
        } else if (nameUpper === "JEUX_OLYMPIQUES_ET_PARALYMPIQUES_DE_2024") {
          // Occurs in http://www.assemblee-nationale.fr/15/textes/0676.asp
          state = nextParentState
          break
        } else if (nameUpper === "TEXTE_DE_LA_COMMISSION_MIXTE_PARITAIRE") {
          // Occurs in http://www.assemblee-nationale.fr/15/textes/1294.asp
          state = nextParentState
          break
        }

        const lineHtml = cleanText(alineaElement!.outerHTML)
        const lineText = cleanText(alineaElement!.textContent)

        if (nextLevel === null) {
          // It is a regular alinea
          // Exclude some alineas
          let excludeAlinea = false
          for (let regex of excludedAlineas) {
            if (regex.test(lineText)) {
              excludeAlinea = true
            }
          }
          if (excludeAlinea) {
            state = nextParentState
            break
          }
          if (
            subdivisionHeaders!.length === 0 ||
            (isMultiLinesHeader && subdivisionAlineas!.length === 0)
          ) {
            subdivisionHeaders!.push({
              texte: lineText,
              html: lineHtml,
            })
          } else {
            subdivisionAlineas!.push({
              texte: lineText,
              html: lineHtml,
            })
          }
          if (isMultiLinesHeader && subdivisionHeaders!.length >= 2) {
            isMultiLinesHeader = false
          }
          // Skip paragraph.
          state = nextParentState
          break
        } else if (nextLevel !== null) {
          // It is a header
          level = nextLevel
          while (levels.length > 0 && level < levels[levels.length - 1]) {
            levels.pop()
          }
          if (levels.length === 0 || level > levels[levels.length - 1]) {
            levels.push(level)
          }

          subdivisionAlineas = []
          subdivisionHeaders = [{ texte: lineText, html: lineHtml }]

          // Articles & "Exposé des motifs" are the only divisions without second title.
          isMultiLinesHeader =
            nameUpper.match(/^ARTICLES?_/) === null &&
            nameUpper.match(/^EXPOSE_DES_MOTIFS$/) === null

          subdivisions.push({
            id: "D_" + nameComputed,
            type: paragraphType,
            niveau: level + 1,
            niveauRelatif: levels.length,
            titres: subdivisionHeaders,
            alineas: subdivisionAlineas,
          })
        }

        state = nextParentState
        break
      case "nextAlineaElement":
        alineaElement = alineaElement!.nextElementSibling
        if (alineaElement === null) {
          // The bodyChild has been fully parsed. Go to next bodyChild
          state = "nextBodyChild"
        } else {
          state = "alineaElement"
        }
        break
      default:
        throw `Unexpected state: ${state}`
    }
  }

  window.close() // Free memory.

  return {
    error: null,
    html,
    page,
    subdivisions,
    url: assembleeUrl,
  }
}
