// To parse this data:
//
//   import { Convert, ComptesRendus, CompteRenduWrapper } from "./file";
//
//   const compteRendu = Convert.toCompteRendu(json);
//   const compteRenduWrapper = Convert.toCompteRenduWrapper(json);
//
// These functions will throw an error if the JSON doesn't
// match the expected interface, even if the JSON is valid.

export interface ComptesRendus {
    compteRendu: CompteRenduClass;
}

export interface CompteRenduClass {
    compteRendu: CompteRendu[];
}

export interface CompteRenduWrapper {
    compteRendu: CompteRendu;
}

export interface CompteRendu {
    xmlns: string;
    uid: string;
    seanceRef: string;
    sessionRef: string;
    metadonnees: Metadonnees;
    contenu: Contenu;
}

export interface Contenu {
    quantiemes: Quantiemes;
    ouvertureSeance: Point[] | Point;
    point: Point[] | Point;
    finSeance?: FinSeance;
    paragraphe?: Point[];
}

export interface FinSeance {
    point: Point;
}

export interface Orateurs {
    orateur: OrateurElement[] | OrateurElement;
}

export interface OrateurElement {
    nom: string | undefined;
    id: string | undefined;
    qualite: string | undefined;
}

export interface Texte {
    _?: string
    id_syceron?: string
    stime?: string
    sup?: string
    lienAdt?: Texte[] | Texte
}

export enum Decision {
    AdoptionAmendementMainLevéeAdt = "ADOPTION amendement (main levée) adt",
    AdoptionDUnArticleVoteÀMainLevée = "ADOPTION d’un article (vote à main levée)",
    AdoptionDUnSousAmendementMainLevéeSsAdt = "ADOPTION d’un sous-amendement (main levée) ss-adt",
    AdoptionDePlusieursAmendementsMainLevéeAdts = "ADOPTION de plusieurs amendements (main levée) adts",
    AdoptionDePlusieursSousAmendementsMainLevéeSsAdts = "ADOPTION de plusieurs sous-amendements (main levée) ss-adts",
    Empty = "",
    RejetDUnAmendementMainLevéeAdt = "REJET d’un amendement (main levée) adt",
    RejetDUnArticleVoteÀMainLevée = "REJET d’un article (vote à main levée)",
    RejetDUnSousAmendementMainLevéeSsAdt = "REJET d’un sous-amendement (main levée) ss-adt",
    RejetDePlusieursAmendementsMainLevéeAdts = "REJET de plusieurs amendements (main levée) adts",
    RejetDePlusieursSousAmendementsMainLevéeSsAdts = "REJET de plusieurs sous-amendements (main levée) ss-adts",
    RejetMotionDeRejetPréalableMainLevée = "REJET motion de rejet préalable (main levée)",
}

export interface Point {
    id_preparation?: string;
    ordre_absolu_seance?: string;
    code_grammaire?: string;
    sommaire?: string;
    id_syceron?: string;
    valeur_ptsodj?: string;
    nivpoint?: string;
    ordinal_prise?: string;
    code_style?: string;
    code_parole?: string;
    valeur?: string;
    orateurs?: Orateurs | string;
    texte?: Texte | string;
    paragraphe?: Point[] | Point;
    interExtraction?: Point[] | Point;
    point?: Point[] | Point;
    structure?: string;
    id_acteur?: string;
    id_mandat?: string;
    id_nomination_oe?: string;
    id_nomination_op?: string;
    orateur?: string;
    bibard?: string;
    art?: string;
    adt?: string;
    ssadt?: string;
    roledebat?: string;
    changementPresidence?: Point;
    type_debat?: string;
    decision?: Decision;
    ouvertureSeance?: Point;
    qualite?: string;
    nom_orateur?: string;
}

export interface Quantiemes {
    journee: string;
    session: string;
}

export interface Metadonnees {
    dateSeance: string;
    dateSeanceJour: string;
    numSeanceJour: string;
    numSeance: string;
    typeAssemblee: string;
    legislature: string;
    session: string;
    nomFichierJo: string;
    validite: string;
    etat: string;
    diffusion: string;
    version: string;
    environnement: string;
    heureGeneration: Date;
    sommaire: Sommaire;
}

export interface Sommaire {
    presidentSeance: Texte;
    sommaire1: SommaireElement[] | SommaireElement;
    sommaire3?: SommaireElement[] | SommaireElement;
    sommaire2?: SommaireElement[] | SommaireElement;
    para?: Texte[] | Texte;
}

export interface SommaireElement {
    valeur_pts_odj: string | undefined;
    titreStruct: TitreStruct;
    para?: Array<Texte | string> | Texte;
    sommaire2?: SommaireElement[] | SommaireElement;
    sommaire3?: SommaireElement[] | SommaireElement;
    presidentSeance?: Texte[] | Texte;
    type_debat?: string;
}

export interface TitreStruct {
    id_syceron: string;
    intitule?: string;
    sousIntitule?: string;
    type_debat?: string;
}

// Converts JSON strings to/from your types
// and asserts the results of JSON.parse at runtime
export class Convert {
    public static toComptesRendus(json: string): ComptesRendus {
        return cast(JSON.parse(json), r("ComptesRendus"));
    }

    public static comptesRendusToJson(value: ComptesRendus): string {
        return JSON.stringify(uncast(value, r("ComptesRendus")), null, 2);
    }

    public static toCompteRenduWrapper(json: string): CompteRenduWrapper {
        return cast(JSON.parse(json), r("CompteRenduWrapper"));
    }

    public static compteRenduWrapperToJson(value: CompteRenduWrapper): string {
        return JSON.stringify(uncast(value, r("CompteRenduWrapper")), null, 2);
    }
}

function invalidValue(typ: any, val: any, key: any, parent: any = ''): never {
    const prettyTyp = prettyTypeName(typ);
    const parentText = parent ? ` on ${parent}` : '';
    const keyText = key ? ` for key "${key}"` : '';
    throw Error(`Invalid value${keyText}${parentText}. Expected ${prettyTyp} but got ${JSON.stringify(val)}`);
}

function prettyTypeName(typ: any): string {
    if (Array.isArray(typ)) {
        if (typ.length === 2 && typ[0] === undefined) {
            return `an optional ${prettyTypeName(typ[1])}`;
        } else {
            return `one of [${typ.map(a => { return prettyTypeName(a); }).join(", ")}]`;
        }
    } else if (typeof typ === "object" && typ.literal !== undefined) {
        return typ.literal;
    } else {
        return typeof typ;
    }
}

function jsonToJSProps(typ: any): any {
    if (typ.jsonToJS === undefined) {
        const map: any = {};
        typ.props.forEach((p: any) => map[p.json] = { key: p.js, typ: p.typ });
        typ.jsonToJS = map;
    }
    return typ.jsonToJS;
}

function jsToJSONProps(typ: any): any {
    if (typ.jsToJSON === undefined) {
        const map: any = {};
        typ.props.forEach((p: any) => map[p.js] = { key: p.json, typ: p.typ });
        typ.jsToJSON = map;
    }
    return typ.jsToJSON;
}

function transform(val: any, typ: any, getProps: any, key: any = '', parent: any = ''): any {
    function transformPrimitive(typ: string, val: any): any {
        if (typeof typ === typeof val) return val;
        return invalidValue(typ, val, key, parent);
    }

    function transformUnion(typs: any[], val: any): any {
        // val must validate against one typ in typs
        const l = typs.length;
        for (let i = 0; i < l; i++) {
            const typ = typs[i];
            try {
                return transform(val, typ, getProps);
            } catch (_) { }
        }
        return invalidValue(typs, val, key, parent);
    }

    function transformEnum(cases: string[], val: any): any {
        if (cases.indexOf(val) !== -1) return val;
        return invalidValue(cases.map(a => { return l(a); }), val, key, parent);
    }

    function transformArray(typ: any, val: any): any {
        // val must be an array with no invalid elements
        if (!Array.isArray(val)) return invalidValue(l("array"), val, key, parent);
        return val.map(el => transform(el, typ, getProps));
    }

    function transformDate(val: any): any {
        if (val === null) {
            return null;
        }
        const d = new Date(val);
        if (isNaN(d.valueOf())) {
            return invalidValue(l("Date"), val, key, parent);
        }
        return d;
    }

    function transformObject(props: { [k: string]: any }, additional: any, val: any): any {
        if (val === null || typeof val !== "object" || Array.isArray(val)) {
            return invalidValue(l(ref || "object"), val, key, parent);
        }
        const result: any = {};
        Object.getOwnPropertyNames(props).forEach(key => {
            const prop = props[key];
            const v = Object.prototype.hasOwnProperty.call(val, key) ? val[key] : undefined;
            result[prop.key] = transform(v, prop.typ, getProps, key, ref);
        });
        Object.getOwnPropertyNames(val).forEach(key => {
            if (!Object.prototype.hasOwnProperty.call(props, key)) {
                result[key] = transform(val[key], additional, getProps, key, ref);
            }
        });
        return result;
    }

    if (typ === "any") return val;
    if (typ === null) {
        if (val === null) return val;
        return invalidValue(typ, val, key, parent);
    }
    if (typ === false) return invalidValue(typ, val, key, parent);
    let ref: any = undefined;
    while (typeof typ === "object" && typ.ref !== undefined) {
        ref = typ.ref;
        typ = typeMap[typ.ref];
    }
    if (Array.isArray(typ)) return transformEnum(typ, val);
    if (typeof typ === "object") {
        return typ.hasOwnProperty("unionMembers") ? transformUnion(typ.unionMembers, val)
            : typ.hasOwnProperty("arrayItems") ? transformArray(typ.arrayItems, val)
                : typ.hasOwnProperty("props") ? transformObject(getProps(typ), typ.additional, val)
                    : invalidValue(typ, val, key, parent);
    }
    // Numbers can be parsed by Date but shouldn't be.
    if (typ === Date && typeof val !== "number") return transformDate(val);
    return transformPrimitive(typ, val);
}

function cast<T>(val: any, typ: any): T {
    return transform(val, typ, jsonToJSProps);
}

function uncast<T>(val: T, typ: any): any {
    return transform(val, typ, jsToJSONProps);
}

function l(typ: any) {
    return { literal: typ };
}

function a(typ: any) {
    return { arrayItems: typ };
}

function u(...typs: any[]) {
    return { unionMembers: typs };
}

function o(props: any[], additional: any) {
    return { props, additional };
}

// function m(additional: any) {
//     return { props: [], additional };
// }

function r(name: string) {
    return { ref: name };
}

const typeMap: any = {
    "ComptesRendus": o([
        { json: "comptesRendus", js: "comptesRendus", typ: r("CompteRenduClass") },
    ], false),
    "CompteRenduClass": o([
        { json: "compteRendu", js: "compteRendu", typ: a(r("CompteRendu")) },
    ], false),
    "CompteRenduWrapper": o([
        { json: "compteRendu", js: "compteRendu", typ: r("CompteRendu") },
    ], false),
    "CompteRendu": o([
        { json: "xmlns", js: "xmlns", typ: "" },
        { json: "uid", js: "uid", typ: "" },
        { json: "seanceRef", js: "seanceRef", typ: "" },
        { json: "sessionRef", js: "sessionRef", typ: "" },
        { json: "metadonnees", js: "metadonnees", typ: r("Metadonnees") },
        { json: "contenu", js: "contenu", typ: r("Contenu") },
    ], false),
    "Contenu": o([
        { json: "quantiemes", js: "quantiemes", typ: r("Quantiemes") },
        { json: "ouvertureSeance", js: "ouvertureSeance", typ: u(a(r("Point")), r("Point")) },
        { json: "point", js: "point", typ: u(a(r("Point")), r("Point")) },
        { json: "finSeance", js: "finSeance", typ: u(undefined, r("FinSeance")) },
        { json: "paragraphe", js: "paragraphe", typ: u(undefined, a(r("Point"))) },
    ], false),
    "FinSeance": o([
        { json: "point", js: "point", typ: r("Point") },
    ], false),
    "Orateurs": o([
        { json: "orateur", js: "orateur", typ: u(a(r("OrateurElement")), r("OrateurElement")) },
    ], false),
    "OrateurElement": o([
        { json: "nom", js: "nom", typ: u(undefined, "") },
        { json: "id", js: "id", typ: u(undefined, "") },
        { json: "qualite", js: "qualite", typ: u(undefined, "") },
    ], false),
    "Texte": o([
        { json: "_", js: "_", typ: u(undefined, "") },
        { json: "id_syceron", js: "id_syceron", typ: u(undefined, "") },
        { json: "stime", js: "stime", typ: u(undefined, "") },
        { json: "sup", js: "sup", typ: u(undefined, "") },
        { json: "lienAdt", js: "lienAdt", typ: u(undefined, u(a(r("Texte")), r("Texte"))) },
    ], false),
    "Point": o([
        { json: "nivpoint", js: "nivpoint", typ: u(undefined, "") },
        { json: "valeur_ptsodj", js: "valeur_ptsodj", typ: u(undefined, "") },
        { json: "ordinal_prise", js: "ordinal_prise", typ: u(undefined, "") },
        { json: "id_preparation", js: "id_preparation", typ: u(undefined, "") },
        { json: "ordre_absolu_seance", js: "ordre_absolu_seance", typ: u(undefined, "") },
        { json: "code_grammaire", js: "code_grammaire", typ: u(undefined, "") },
        { json: "code_style", js: "code_style", typ: u(undefined, "") },
        { json: "code_parole", js: "code_parole", typ: u(undefined, "") },
        { json: "sommaire", js: "sommaire", typ: u(undefined, "") },
        { json: "id_syceron", js: "id_syceron", typ: u(undefined, "") },
        { json: "valeur", js: "valeur", typ: u(undefined, "") },
        { json: "orateurs", js: "orateurs", typ: u(r("Orateurs"), u("", undefined)) },
        { json: "texte", js: "texte", typ: u(r("Texte"), u("", undefined)) },
        { json: "paragraphe", js: "paragraphe", typ: u(undefined, u(a(r("Point")), r("Point"))) },
        { json: "interExtraction", js: "interExtraction", typ: u(undefined, u(a(r("Point")), r("Point"))) },
        { json: "point", js: "point", typ: u(undefined, u(a(r("Point")), r("Point"))) },
        { json: "structure", js: "structure", typ: u(undefined, "") },
        { json: "id_acteur", js: "id_acteur", typ: u(undefined, "") },
        { json: "id_mandat", js: "id_mandat", typ: u(undefined, "") },
        { json: "id_nomination_oe", js: "id_nomination_oe", typ: u(undefined, "") },
        { json: "id_nomination_op", js: "id_nomination_op", typ: u(undefined, "") },
        { json: "orateur", js: "orateur", typ: u(undefined, "") },
        { json: "bibard", js: "bibard", typ: u(undefined, "") },
        { json: "art", js: "art", typ: u(undefined, "") },
        { json: "adt", js: "adt", typ: u(undefined, "") },
        { json: "ssadt", js: "ssadt", typ: u(undefined, "") },
        { json: "roledebat", js: "roledebat", typ: u(undefined, "") },
        { json: "type_debat", js: "type_debat", typ: u(undefined, "") },
        { json: "changementPresidence", js: "changementPresidence", typ: u(undefined, r("Point")) },
        { json: "decision", js: "decision", typ: u(undefined, r("Decision")) },
        { json: "ouvertureSeance", js: "ouvertureSeance", typ: u(undefined, r("Point")) },
        { json: "nom_orateur", js: "nom_orateur", typ: u(undefined, "") },
        { json: "qualite", js: "qualite", typ: u(undefined, "") },
    ], false),
    "Quantiemes": o([
        { json: "journee", js: "journee", typ: "" },
        { json: "session", js: "session", typ: "" },
    ], false),
    "Metadonnees": o([
        { json: "dateSeance", js: "dateSeance", typ: "" },
        { json: "dateSeanceJour", js: "dateSeanceJour", typ: "" },
        { json: "numSeanceJour", js: "numSeanceJour", typ: "" },
        { json: "numSeance", js: "numSeance", typ: "" },
        { json: "typeAssemblee", js: "typeAssemblee", typ: "" },
        { json: "legislature", js: "legislature", typ: "" },
        { json: "session", js: "session", typ: "" },
        { json: "nomFichierJo", js: "nomFichierJo", typ: "" },
        { json: "validite", js: "validite", typ: "" },
        { json: "etat", js: "etat", typ: "" },
        { json: "diffusion", js: "diffusion", typ: "" },
        { json: "version", js: "version", typ: "" },
        { json: "environnement", js: "environnement", typ: "" },
        { json: "heureGeneration", js: "heureGeneration", typ: Date },
        { json: "sommaire", js: "sommaire", typ: r("Sommaire") },
    ], false),
    "Sommaire": o([
        { json: "presidentSeance", js: "presidentSeance", typ: r("Texte") },
        { json: "sommaire1", js: "sommaire1", typ: u(a(r("SommaireElement")), r("SommaireElement")) },
        { json: "sommaire3", js: "sommaire3", typ: u(undefined, u(a(r("SommaireElement")), r("SommaireElement"))) },
        { json: "sommaire2", js: "sommaire2", typ: u(undefined, u(a(r("SommaireElement")), r("SommaireElement"))) },
        { json: "para", js: "para", typ: u(undefined, u(a(r("Texte")), r("Texte"))) },
    ], false),
    "SommaireElement": o([
        { json: "valeur_pts_odj", js: "valeur_pts_odj", typ: u(undefined, "") },
        { json: "titreStruct", js: "titreStruct", typ: u(undefined, r("TitreStruct")) },
        { json: "para", js: "para", typ: u(undefined, u(a(u(r("Texte"), "")), r("Texte"))) },
        { json: "sommaire2", js: "sommaire2", typ: u(undefined, u(a(r("SommaireElement")), r("SommaireElement"))) },
        { json: "sommaire3", js: "sommaire3", typ: u(undefined, u(a(r("SommaireElement")), r("SommaireElement"))) },
        { json: "presidentSeance", js: "presidentSeance", typ: u(undefined, u(a(r("Texte")), r("Texte"))) },
        { json: "type_debat", js: "type_debat", typ: u(undefined, "") },
    ], false),
    "TitreStruct": o([
        { json: "id_syceron", js: "id_syceron", typ: "" },
        { json: "intitule", js: "intitule", typ: u(undefined, "") },
        { json: "sousIntitule", js: "sousIntitule", typ: u(undefined, "") },
        { json: "type_debat", js: "type_debat", typ: u(undefined, "") },
    ], false),
    "Decision": [
        "ADOPTION amendement (main levée) adt",
        "ADOPTION d’un article (vote à main levée)",
        "ADOPTION d’un sous-amendement (main levée) ss-adt",
        "ADOPTION de plusieurs amendements (main levée) adts",
        "ADOPTION de plusieurs sous-amendements (main levée) ss-adts",
        "",
        "REJET d’un amendement (main levée) adt",
        "REJET d’un article (vote à main levée)",
        "REJET d’un sous-amendement (main levée) ss-adt",
        "REJET de plusieurs amendements (main levée) adts",
        "REJET de plusieurs sous-amendements (main levée) ss-adts",
        "REJET motion de rejet préalable (main levée)",
    ],
};