import fs from "fs-extra"
import commandLineArgs from "command-line-args"
import commandLineUsage from "command-line-usage"
import { load, getFiles, GetFileNotFoundError } from "../file_systems"
import { SchemaBugs } from "../../src/bugs"
import * as git from "../../src/git"
import {
  validDataset,
  getDatasets,
  getSchemas,
  datasetDirectorySchema,
  diffUpstream,
  masterUrl,
} from "../../src/datasets"

function parseArgs(argv: string[]): any {
  const optionsDefinitions = [
    {
      description: "pathname to the directory containing the dataset",
      defaultOption: true,
      name: "data",
      type: String,
    },
    {
      description:
        "pathname to the directory containing the HTML comptes rendus de seance",
      name: "cr",
      type: String,
    },
    {
      description: "pathname to the directory to commit",
      name: "commit",
      type: String,
    },
    {
      description: "only consider this bug",
      name: "bug",
      type: String,
    },
    {
      description:
        "pathname to the directory where the check outputs will be stored",
      name: "output",
      type: String,
    },
    {
      description: `dataset (possible values must include exactly one of ${getDatasets()})`,
      name: "dataset",
      type: String,
    },
    {
      description: `schema (possible values are ${getSchemas()})`,
      name: "schema",
      type: String,
    },
    {
      description: "directory in which bugs are found (default .)",
      name: "bugs-dir",
      type: String,
    },
    {
      description: "increase verbosity",
      name: "verbose",
      type: Boolean,
    },
    {
      description: "debug information",
      name: "debug",
      type: Boolean,
    },
    {
      description: "check",
      name: "check",
      type: Boolean,
    },
    {
      description: "fix",
      name: "fix",
      type: Boolean,
    },
    {
      name: "help",
      description: "Print this usage guide.",
    },
  ]
  const sections = [
    {
      header: "Cross check data consistency and fix problems",
      content: "",
    },
    {
      header: "Options",
      optionList: optionsDefinitions,
    },
  ]
  const options = commandLineArgs(optionsDefinitions, {
    argv: argv,
  })

  if (options.dataset) {
    if (!validDataset(options.dataset)) {
      console.error(`--dataset ${options.dataset}`)
      options.help = true
    }
  }

  if (options.schema) {
    if (!getSchemas().includes(options.schema)) {
      console.error(`--schema ${options.schema}`)
      options.help = true
    }
  }

  if (options.commit) {
    if (!fs.existsSync(options.commit)) {
      console.error(`${options.commit} does not exist`)
      options.help = true
    }
  }

  if (options.debug) options.verbose = true

  if ("help" in options) {
    const usage = commandLineUsage(sections)
    console.warn(usage)
    return null
  }
  return options
}

export class BugsHelper {
  options: any
  bugs: any

  constructor(options: any) {
    this.options = options
    this.bugs = new SchemaBugs(this.options)
  }

  async check(): Promise<number> {
    const analysis: any = {}
    await this.bugs.preCheck()
    if (this.options.verbose) console.log(`check ${this.options.data}`)
    for (const file of getFiles([this.options.data])) {
      const content = load(file)
      analysis[file] = await this.bugs.check(content, file)
    }
    await this.bugs.postCheck(analysis)
    await this.show(analysis)
    if (this.options.output) await this.store(analysis)
    return 0
  }

  async show(analysis: any): Promise<void> {
    for (const filename of Object.keys(analysis)) {
      const bug = analysis[filename]
      for (const name of Object.keys(bug)) {
        const result = bug[name]
        if (result != null) {
          let log
          if (result["status"] == "needs-fixing") {
            log = console.error
          } else if (this.options.verbose) {
            log = console.log
          }
          if (log != undefined) {
            log(`${name} finds ${filename} ${result["status"]}`)
            log(result["info"].join("\n"))
          }
        }
      }
    }
  }

  async store(analysis: any): Promise<void> {
    const name2reports: any = {}
    for (const filename of Object.keys(analysis)) {
      const bug = analysis[filename]
      for (const name of Object.keys(bug)) {
        if (!(name in name2reports)) name2reports[name] = {}
        name2reports[name][filename] = bug[name]
      }
    }
    for (const name of Object.keys(name2reports)) {
      const reports = name2reports[name]
      const n = name.split("-")[1]
      let lines: any = [
        `## Bug [${name}](https://git.en-root.org/tricoteuses/tricoteuses-assemblee-qa/issues/${Number(
          n,
        )})`,
        `This report was created by the **check** function of [${name}](https://git.en-root.org/tricoteuses/tricoteuses-assemblee/-/blob/master/src/bugs/${name}.ts).`,
      ]
      const status2lines: any = {
        "needs-fixing": [],
        "will-be-fixed": [],
        "already-fixed": [],
      }
      for (const filename of Object.keys(reports)) {
        const report = reports[filename]
        if (report == null || report["status"] == "ok") continue
        if (!(report["status"] in status2lines))
          console.error(`${filename} status ${report["status"]} unknown`)
        const status = report["status"]
        let l = status2lines[status]
        if (status == "will-be-fixed") l.push(`* ${diffUpstream(filename)}`)
        else l.push(`* ${masterUrl(filename)}`)
        for (const info of report["info"]) l.push(`  * ${info}`)
      }
      for (const [status, label] of [
        ["needs-fixing", "The following files must be fixed"],
        ["already-fixed", "The following files no longer need a fix"],
        ["will-be-fixed", "The following files have been successfully fixed"],
      ]) {
        lines.push(`### ${label}`)
        lines = lines.concat(status2lines[status])
      }
      const output = `${this.options.output}/${name}.md`
      fs.writeFileSync(output, lines.join("\n"), {
        encoding: "utf8",
      })
    }
  }

  async fix(): Promise<number> {
    await this.bugs.preFix()
    await this.bugs.fix(getFiles([this.options.data]))
    return 0
  }

  async main(): Promise<number> {
    if (this.options.check) return this.check()
    else return this.fix()
  }
}

export async function checkoutFix(options: any): Promise<void> {
  const c = options.commit
  if (c == undefined) return
  const v = options.verbose
  git.run(c, "checkout --quiet master", v)
  if (/tmp/.test(git.run(c, "branch", v))) git.run(c, "branch -D tmp", v)
  git.run(c, "checkout --quiet -b tmp master", v)
  if (/Merge bugfixes/.test(git.run(c, "log --oneline -1", v)))
    git.run(c, "revert --mainline 1 --no-edit HEAD", v)
  git.run(c, "merge --no-ff -m 'Merge upstream' upstream/uid", v)
  if (/bugfixes/.test(git.run(c, "branch", v)))
    git.run(c, "branch -D bugfixes", v)
  git.run(c, "checkout --quiet -b bugfixes", v)
}

export async function mergeFix(options: any): Promise<void> {
  const c = options.commit
  if (c == undefined) return
  const v = options.verbose
  git.run(c, "checkout --quiet tmp", v)
  git.run(c, "merge --no-ff -m 'Merge bugfixes' bugfixes", v)
  git.run(c, "branch -D bugfixes", v)
  git.run(c, "checkout --quiet master", v)
  if (!git.test(c, "diff --quiet tmp master", v))
    git.run(c, "merge --ff-only tmp", v)
  git.run(c, "branch -D tmp", v)
}

export async function main(argv: any): Promise<number> {
  const options = parseArgs(argv)
  if (options === null) return 1
  let directorySchema = []
  if (options.schema) {
    directorySchema = [[options.data, options.schema]]
  } else if (options.dataset) {
    directorySchema = datasetDirectorySchema(options.dataset)
  } else {
    console.error("either --schema or --dataset is required")
    return 1
  }
  await checkoutFix(options)
  for (const d of directorySchema) {
    options.schema = d[1]
    options.data = d[0]
    if (options.debug)
      console.log(`working on ${options.data} with schema ${options.schema}`)
    let helper
    try {
      helper = new BugsHelper(options)
    } catch (e) {
      if (!(e instanceof GetFileNotFoundError)) {
        throw e
      }
      continue
    }
    if (options.verbose)
      console.log(
        `working on files ${options.data} with schema ${options.schema}`,
      )
    const status = await helper.main()
    if (status != 0) return status
  }
  await mergeFix(options)
  return 0
}

/* istanbul ignore if */
if (process.argv[1].endsWith("bugs_helper.ts"))
  main(process.argv)
    .then((exitCode) => {
      process.exit(exitCode)
    })
    .catch((error) => {
      console.error(error)
      process.exit(1)
    })
