import commandLineArgs from "command-line-args"
import fs from "fs-extra"
import path from "path"
import {
  cleanActeur,
  cleanAmendement,
  cleanDocumentOrDivision,
  cleanDossierParlementaire,
  cleanOrgane,
  cleanReunion,
  cleanScrutin,
  cleanQuestion,
  cleanCompteRendu,
} from "../cleaners"
import {
  datasets,
  EnabledDatasets,
  getEnabledDatasets,
  filterByLegislature,
} from "../datasets"
import { downloadAndParse } from "./retrieve_documents"
import { existingDateToJson, patchedDateToJson } from "../dates"
import { walkDir } from "../file_systems"
import * as git from "../git"
import {
  Acteur,
  Convert as ActeursEtOrganesConvert,
  Organe,
} from "../types/acteurs_et_organes"
import { Convert as AgendasConvert, Reunion } from "../types/agendas"
import { Amendement, Convert as AmendementsConvert } from "../types/amendements"
import {
  Convert as DossiersLegislatifsConvert,
  Document,
  DossierParlementaire,
} from "../types/dossiers_legislatifs"
import { Convert as ScrutinsConvert, Scrutin } from "../types/scrutins"
import { Convert as QuestionsConvert, Question } from "../types/questions"
import { Convert as ComptesRendusConvert, CompteRendu } from "../types/debats"
import {
  cloneOption,
  commitOption,
  remoteOption,
  legislatureOption,
  categoriesOption,
  dataDirDefaultOption,
  silentOption,
  verboseOption,
  pullOption,
  fetchDocumentsOption,
  parseDocumentsOption,
  documentFormatOption,
} from "./shared/cli_helpers"

function parseArgs(argv: string[]): any {
  const optionsDefinitions = [
    cloneOption,
    commitOption,
    remoteOption,
    legislatureOption,
    categoriesOption,
    silentOption,
    verboseOption,
    dataDirDefaultOption,
    pullOption,
    fetchDocumentsOption,
    parseDocumentsOption,
    documentFormatOption,
    {
      alias: "d",
      help: "single dataset to reorganize",
      name: "dataset",
      type: String,
    },
    {
      help: "do not reset --hard after commit to save I/O if the working tree is to be discarded",
      name: "no-reset-after-commit",
      type: Boolean,
    },
    {
      alias: "V",
      help: "don't validate data",
      name: "no-validate",
      type: Boolean,
    },
  ]
  const options = commandLineArgs(optionsDefinitions, {
    argv: argv,
  })
  return options
}

function cloneGit(
  gitGroupUrl: string | undefined,
  datasetName: string,
  dataDir: string,
): void {
  git.clone(gitGroupUrl, `${datasetName}_nettoye`, dataDir)
}

function commitGit(
  datasetCleanDir: string,
  options: any,
  exitCode: number,
  subdirectories: any,
): number {
  if (options.commit) {
    const errorCode = git.commitAndPush(
      datasetCleanDir,
      "Nouvelle moisson",
      options.remote,
    )
    if (
      (exitCode === 10 && errorCode !== 10) ||
      (exitCode === 0 && errorCode !== 0 && errorCode !== 10)
    ) {
      exitCode = errorCode
    }
  }
  if (
    commitAndPushUid(datasetCleanDir, options, subdirectories) &&
    (exitCode === 0 || exitCode === 10)
  ) {
    exitCode = 1
  }
  return exitCode
}

function cleanAndValidateData<T>(
  data: any,
  cleaner: (data: T) => void,
  fromJson: (json: string) => T,
  toJson: (data: T) => string,
  noValidate: Boolean,
): string | null {
  cleaner(data)
  if (typeof data === "object" && Object.keys(data).length === 0) {
    // By convention: an empty object must be deleted.
    return null
  }
  const jsonData = JSON.stringify(data, null, 2)
  if (!noValidate) {
    return jsonData
  } else {
    const validatedData: T = fromJson(jsonData)
    return toJson(validatedData)
  }
}

async function processFiles<T>(
  datasetRawDir: string,
  datasetCleanDir: string,
  walkDirParams: any[],
  cleaner: any,
  dataConverter: { fromJson: any; toJson: any },
  options: any,
  postProcessor?: any,
): Promise<void> {
  for (const filePathArray of walkDir(datasetRawDir, walkDirParams)) {
    const filename = filePathArray[filePathArray.length - 1]
    if (!filename.endsWith(".json")) continue
    const rawFilePath = path.join(datasetRawDir, ...filePathArray)
    const rawJson = fs.readFileSync(rawFilePath, { encoding: "utf8" })
    if (options.verbose) console.log(`Cleaning file: ${rawFilePath}...`)
    const data = JSON.parse(rawJson)
    const validateData = !options["no-validate"]
    const cleanJson = cleanAndValidateData<T>(
      data,
      cleaner,
      dataConverter.fromJson,
      dataConverter.toJson,
      validateData,
    )

    if (cleanJson !== null) {
      const cleanFilePath = path.join(datasetCleanDir, ...filePathArray)
      fs.ensureDirSync(path.dirname(cleanFilePath))
      fs.writeFileSync(cleanFilePath, cleanJson, { encoding: "utf8" })

      if (postProcessor) {
        await postProcessor(data, datasetCleanDir, options)
      }
    }
  }
}

async function cleanReorganizedData(options: any): Promise<number> {
  const dataDir: string = options.dataDir
  Date.prototype.toJSON = patchedDateToJson
  let exitCode = options.commit ? 10 : 0
  const gitGroupUrl = options.clone?.trim().replace(/\/+$/, "")
  const enabledDatasets = getEnabledDatasets(options.categories)
  let chosenDatasets = filterByLegislature(datasets, options.legislature)

  if (enabledDatasets & EnabledDatasets.ActeursEtOrganes) {
    for (const dataset of chosenDatasets.acteursEtOrganes) {
      const datasetName = path.basename(dataset.filename, ".json")
      if (options.dataset && datasetName !== options.dataset) {
        continue
      }
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetRawDir: string = path.join(dataDir, datasetName)
      const datasetCleanDir: string = `${datasetRawDir}_nettoye`
      if (options.pull) {
        git.resetAndPull(datasetCleanDir)
      }

      analyzeAndRemove(datasetName, options)
      await processFiles<Acteur>(
        datasetRawDir,
        datasetCleanDir,
        ["acteurs"],
        cleanActeur,
        {
          fromJson: ActeursEtOrganesConvert.toActeur,
          toJson: ActeursEtOrganesConvert.acteurToJson,
        },
        options,
      )
      await processFiles<Organe>(
        datasetRawDir,
        datasetCleanDir,
        ["organes"],
        cleanOrgane,
        {
          fromJson: ActeursEtOrganesConvert.toOrgane,
          toJson: ActeursEtOrganesConvert.organeToJson,
        },
        options,
      )

      exitCode = commitGit(datasetCleanDir, options, exitCode, [
        "acteurs",
        "organes",
      ])
    }
  }

  if (enabledDatasets & EnabledDatasets.Agendas) {
    for (const dataset of chosenDatasets.agendas) {
      const datasetName = path.basename(dataset.filename, ".json")
      if (options.dataset && datasetName !== options.dataset) {
        continue
      }

      cloneGit(gitGroupUrl, datasetName, dataDir)
      analyzeAndRemove(datasetName, options)
      const datasetRawDir: string = path.join(dataDir, datasetName)
      const datasetCleanDir: string = `${datasetRawDir}_nettoye`

      await processFiles<Reunion>(
        datasetRawDir,
        datasetCleanDir,
        [],
        cleanReunion,
        {
          fromJson: AgendasConvert.toReunion,
          toJson: AgendasConvert.reunionToJson,
        },
        options,
      )

      exitCode = commitGit(datasetCleanDir, options, exitCode, ["."])
    }
  }

  if (enabledDatasets & EnabledDatasets.Amendements) {
    for (const dataset of chosenDatasets.amendements) {
      const datasetName = path.basename(dataset.filename, ".json")
      if (options.dataset && datasetName !== options.dataset) {
        continue
      }

      cloneGit(gitGroupUrl, datasetName, dataDir)
      analyzeAndRemove(datasetName, options)
      const datasetRawDir: string = path.join(dataDir, datasetName)
      const datasetCleanDir: string = `${datasetRawDir}_nettoye`

      await processFiles<Amendement>(
        datasetRawDir,
        datasetCleanDir,
        [],
        cleanAmendement,
        {
          fromJson: AmendementsConvert.toAmendement,
          toJson: AmendementsConvert.amendementToJson,
        },
        options,
      )

      exitCode = commitGit(datasetCleanDir, options, exitCode, ["."])
    }
  }

  if (enabledDatasets & EnabledDatasets.DossiersLegislatifs) {
    for (const dataset of chosenDatasets.dossiersLegislatifs) {
      const datasetName = path.basename(dataset.filename, ".json")
      if (options.dataset && datasetName !== options.dataset) {
        continue
      }

      cloneGit(gitGroupUrl, datasetName, dataDir)
      analyzeAndRemove(datasetName, options)
      const datasetRawDir: string = path.join(dataDir, datasetName)
      const datasetCleanDir: string = `${datasetRawDir}_nettoye`

      await processFiles<Document>(
        datasetRawDir,
        datasetCleanDir,
        ["documents"],
        cleanDocumentOrDivision,
        {
          fromJson: DossiersLegislatifsConvert.toDocument,
          toJson: DossiersLegislatifsConvert.documentToJson,
        },
        options,
        downloadAndParse,
      )
      await processFiles<DossierParlementaire>(
        datasetRawDir,
        datasetCleanDir,
        ["dossiers"],
        cleanDossierParlementaire,
        {
          fromJson: DossiersLegislatifsConvert.toDossierParlementaire,
          toJson: DossiersLegislatifsConvert.dossierParlementaireToJson,
        },
        options,
      )

      exitCode = commitGit(datasetCleanDir, options, exitCode, [
        "documents",
        "dossiers",
      ])
    }
  }

  if (enabledDatasets & EnabledDatasets.Scrutins) {
    for (const dataset of chosenDatasets.scrutins) {
      const datasetName = path.basename(dataset.filename, ".json")
      if (options.dataset && datasetName !== options.dataset) {
        continue
      }

      cloneGit(gitGroupUrl, datasetName, dataDir)
      analyzeAndRemove(datasetName, options)
      const datasetRawDir: string = path.join(dataDir, datasetName)
      const datasetCleanDir: string = `${datasetRawDir}_nettoye`

      await processFiles<Scrutin>(
        datasetRawDir,
        datasetCleanDir,
        [],
        cleanScrutin,
        {
          fromJson: ScrutinsConvert.toScrutin,
          toJson: ScrutinsConvert.scrutinToJson,
        },
        options,
      )

      exitCode = commitGit(datasetCleanDir, options, exitCode, ["."])
    }
  }

  if (enabledDatasets & EnabledDatasets.Questions) {
    for (const dataset of datasets.questions) {
      const datasetName = path.basename(dataset.filename, ".json")
      if (options.dataset && datasetName !== options.dataset) {
        continue
      }

      cloneGit(gitGroupUrl, datasetName, dataDir)
      analyzeAndRemove(datasetName, options)
      const datasetRawDir: string = path.join(dataDir, datasetName)
      const datasetCleanDir: string = `${datasetRawDir}_nettoye`

      await processFiles<Question>(
        datasetRawDir,
        datasetCleanDir,
        [],
        cleanQuestion,
        {
          fromJson: QuestionsConvert.toQuestion,
          toJson: QuestionsConvert.questionToJson,
        },
        options,
      )

      exitCode = commitGit(datasetCleanDir, options, exitCode, ["."])
    }
  }

  if (enabledDatasets & EnabledDatasets.ComptesRendusSeances) {
    for (const dataset of datasets.comptesRendusSeances) {
      const datasetName = path.basename(dataset.filename, ".xml")
      if (options.dataset && datasetName !== options.dataset) {
        continue
      }

      cloneGit(gitGroupUrl, datasetName, dataDir)
      analyzeAndRemove(datasetName, options)
      const datasetRawDir: string = path.join(dataDir, datasetName)
      const datasetCleanDir: string = `${datasetRawDir}_nettoye`

      await processFiles<CompteRendu>(
        datasetRawDir,
        datasetCleanDir,
        [],
        cleanCompteRendu,
        {
          fromJson: ComptesRendusConvert.toCompteRendu,
          toJson: ComptesRendusConvert.compteRenduToJson,
        },
        options,
      )

      exitCode = commitGit(datasetCleanDir, options, exitCode, ["."])
    }
  }

  // Restore standard conversion of dates to JSON.
  Date.prototype.toJSON = existingDateToJson

  if (!options.silent) {
    console.log("Finished !")
  }

  return exitCode
}

function analyzeAndRemove(datasetName: string, options: any): any {
  const dataDir = options.dataDir
  const datasetRawDir: string = path.join(dataDir, datasetName)
  const datasetCleanDir: string = `${datasetRawDir}_nettoye`
  if (!options.silent) {
    console.log(`Cleaning directory: ${datasetCleanDir}…`)
  }
  fs.ensureDirSync(datasetCleanDir)
  for (const filename of fs.readdirSync(datasetCleanDir)) {
    if (filename[0] === ".") {
      continue
    }
    const pathname = path.join(datasetCleanDir, filename)
    fs.removeSync(pathname)
  }
}

function commitAndPushUid(
  repositoryDir: string,
  options: any,
  subdirectories: any,
): boolean {
  let error = false
  if (options.commit) {
    git.uidBranchUpdate(
      repositoryDir,
      subdirectories,
      options["no-reset-after-commit"],
    )
    for (const remote of options.remote || []) {
      try {
        git.run(repositoryDir, `push --quiet ${remote} uid:uid`)
      } catch (childProcess) {
        error = true
      }
    }
  }
  return error
}

function main(argv: any): Promise<number> {
  const options = parseArgs(argv)
  return cleanReorganizedData(options)
}

/* istanbul ignore if */
if (process.argv[1].endsWith("clean_reorganized_data.ts")) {
  main(process.argv).then(process.exit)
}
