import fs from "fs-extra"
import path from "path"
import stream from "stream"
import StreamZip from "node-stream-zip"
import util from "util"

import { walkDir } from "../file_systems"
import { XMLParser } from "fast-xml-parser"

const pipeline = util.promisify(stream.pipeline)

// Function to fetch zip files with retry mechanism
async function fetchZipFile(
  url: string,
  zipFilePath: string,
  retries = 3,
): Promise<void> {
  for (let attempt = 0; attempt < retries; attempt++) {
    const response = await fetch(url)
    if (response.ok) {
      await pipeline(
        response.body as unknown as NodeJS.ReadableStream,
        fs.createWriteStream(zipFilePath),
      )
      return
    } else if (attempt < retries - 1) {
      console.warn(`Retrying fetch for ${url}: ${response.statusText}`)
    } else {
      console.error(`Failed to fetch ${url} after ${retries} attempts`)
      throw new Error(`Fetch failed: ${url}`)
    }
  }
}

async function extractZip(
  zipFilePath: string,
  outputDir: string,
): Promise<void> {
  return new Promise<void>((resolve, reject) => {
    const zip = new StreamZip({
      file: zipFilePath,
      storeEntries: true,
    })

    zip.on("ready", () => {
      zip.extract(null!, outputDir, (err) => {
        zip.close()
        if (err) {
          reject(err)
        } else {
          resolve()
        }
      })
    })
  })
}

function reindentJsonFiles(dataDirOrFilePath: string): void {
  if (fs.statSync(dataDirOrFilePath).isDirectory()) {
    for (const dataFileSplitPath of walkDir(dataDirOrFilePath)) {
      const dataFilePath = path.join(dataDirOrFilePath, ...dataFileSplitPath)
      if (dataFilePath.endsWith(".json")) {
        const data = JSON.parse(
          fs.readFileSync(dataFilePath, { encoding: "utf-8" }),
        )
        fs.writeFileSync(dataFilePath, JSON.stringify(data, null, 2))
      }
    }
  } else {
    const data = JSON.parse(
      fs.readFileSync(dataDirOrFilePath, { encoding: "utf-8" }),
    )
    fs.writeFileSync(dataDirOrFilePath, JSON.stringify(data, null, 2))
  }
}

type ProcessDatasetArguments = {
  dataset: any
  dataDir: string
  options: any
}

const xmlParser = new XMLParser({
  ignoreDeclaration: true,
  ignoreAttributes: false,
  ignorePiTags: true,
  attributeNamePrefix: "",
  textNodeName: "_",
  processEntities: false,
  trimValues: true,
  stopNodes: ["*.texte", "*.intitule"],
  parseTagValue: false,
  parseAttributeValue: false,
  transformTagName: (tagName: string) =>
    tagName === "Para" ? "para" : tagName,
})

/**
 * Processes a dataset by fetching, unzipping, and converting files from XML to JSON format.
 *
 * @param {any} dataset - The dataset object containing metadata and processing instructions.
 * @param {string} dataDir - The directory where the dataset files will be stored.
 * @param {any} options - Options to control the processing behavior, such as fetching and silent mode.
 * @returns {Promise<void>} A promise that resolves when the dataset processing is complete.
 */
export async function processDataset(
  args: ProcessDatasetArguments,
): Promise<void> {
  const { dataset, dataDir, options } = args
  const zipFilePath = path.join(dataDir, `${dataset.filename}.zip`)
  const dataDirOrFilePath = path.join(dataDir, dataset.filename)
  fs.removeSync(dataDirOrFilePath)

  if (options.fetch) {
    // Fetch & save ZIP file.
    if (!options.silent) {
      console.log(`Loading ${dataset.title}: ${dataset.filename}.zip`)
    }
    await fetchZipFile(dataset.url, zipFilePath)
  }
  if (!options.silent) {
    console.log(`Unzipping ${dataset.title}: ${dataset.filename}.zip`)
  }
  await extractZip(zipFilePath, dataDir)

  if (dataset.repairZip !== undefined) {
    if (!options.silent) {
      console.log(`Repairing ${dataset.title}: ${dataset.filename}`)
    }
    dataset.repairZip(dataset, dataDir)
  }

  // Convert xml to JSON.
  if (dataset.filename.endsWith(".xml")) {
    for (const dataFileSplitPath of walkDir(dataDirOrFilePath)) {
      const dataFilePath = path.join(dataDirOrFilePath, ...dataFileSplitPath)
      if (dataFilePath.endsWith(".xml")) {
        fs.readFile(dataFilePath, function (err, data) {
          if (err) return console.error(err)
          let result = xmlParser.parse(data)
          const newDataFilePath = dataFilePath.replace(/\.[^.]+$/, ".json")
          fs.copy(dataFilePath, newDataFilePath, (err) => {
            fs.writeFileSync(newDataFilePath, JSON.stringify(result, null, 2))
            if (err) return console.error(err)
          })
        })
      }
    }
  }

  // Reindent JSON file.
  if (!options.silent) {
    console.log(`Reidenting ${dataset.title}: ${dataset.filename}`)
  }
  reindentJsonFiles(dataDirOrFilePath)
}
