import assert from "assert"
import commandLineArgs from "command-line-args"
import fs from "fs-extra"
import path from "path"
import {
  datasets,
  DatasetStructure,
  EnabledDatasets,
  getEnabledDatasets,
  filterByLegislature,
} from "../datasets"
import { existingDateToJson, patchedDateToJson } from "../dates"
import * as git from "../git"
import { walkDir } from "../file_systems"
import {
  pathFromCompteRenduUid,
  pathFromDocumentUid,
  pathFromDossierParlementaireUid,
  pathFromQuestionUid,
  pathFromReunionUid,
  pathFromScrutinUid,
} from "../loaders"
import { Convert as ActeursEtOrganesRawConvert } from "../raw_types/acteurs_et_organes"
import { Convert as AgendasRawConvert } from "../raw_types/agendas"
import { Convert as AmendementsRawConvert } from "../raw_types/amendements"
import { Convert as DossiersLegislatifsRawConvert } from "../raw_types/dossiers_legislatifs"
import { Convert as ScrutinsRawConvert } from "../raw_types/scrutins"
import { Convert as QuestionsRawConvert } from "../raw_types/questions"
import { Convert as ComptesRendusRawConvert } from "../raw_types/debats"
import {
  cloneOption,
  commitOption,
  categoriesOption,
  legislatureOption,
  remoteOption,
  silentOption,
  verboseOption,
  dataDirDefaultOption,
  pullOption,
  fetchDocumentsOption,
  parseDocumentsOption,
  documentFormatOption,
} from "./shared/cli_helpers"

function parseArgs(argv: string[]): any {
  const optionsDefinitions = [
    cloneOption,
    commitOption,
    categoriesOption,
    legislatureOption,
    remoteOption,
    silentOption,
    verboseOption,
    dataDirDefaultOption,
    fetchDocumentsOption,
    parseDocumentsOption,
    documentFormatOption,
    pullOption,
    {
      alias: "V",
      help: "don't validate raw data",
      name: "no-validate-raw",
      type: Boolean,
    },
  ]
  const options = commandLineArgs(optionsDefinitions, {
    argv: argv,
  })
  return options
}

function clearDirectory(dirPath: string): void {
  fs.ensureDirSync(dirPath)
  for (const filename of fs.readdirSync(dirPath)) {
    if (!filename.startsWith(".")) {
      fs.removeSync(path.join(dirPath, filename))
    }
  }
}

function cloneGit(
  gitGroupUrl: string | undefined,
  datasetName: string,
  dataDir: string,
): void {
  git.clone(gitGroupUrl, datasetName, dataDir)
}

function reorganizeData(options: any): number {
  const dataDir = options.dataDir
  Date.prototype.toJSON = patchedDateToJson
  let exitCode = options.commit ? 10 : 0
  const gitGroupUrl = options.clone?.trim().replace(/\/+$/, "")
  const validateRawData = !options["no-validate-raw"]
  const enabledDatasets = getEnabledDatasets(options.categories)
  let chosenDatasets = filterByLegislature(datasets, options.legislature)

  function commitGit(datasetReorganizedDir: string, options: any): void {
    if (options.commit) {
      const errorCode = git.commitAndPush(
        datasetReorganizedDir,
        "Nouvelle moisson",
        options.remote,
      )
      if (
        (exitCode === 10 && errorCode !== 10) ||
        (exitCode === 0 && errorCode !== 0 && errorCode !== 10)
      ) {
        exitCode = errorCode
      }
    }
  }

  if (enabledDatasets & EnabledDatasets.ActeursEtOrganes) {
    for (const dataset of chosenDatasets.acteursEtOrganes) {
      const originalJsonDirectoryOrFilePath: string = path.join(
        dataDir,
        dataset.filename,
      )
      if (!options.silent) {
        console.log(`Reorganizing ${originalJsonDirectoryOrFilePath}`)
      }

      const datasetName = path.basename(
        originalJsonDirectoryOrFilePath,
        ".json",
      )
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetReorganizedDir = path.join(dataDir, datasetName)
      if (options.pull) {
        git.resetAndPull(datasetReorganizedDir)
      }
      clearDirectory(datasetReorganizedDir)

      const acteursReorganizedDir = path.join(datasetReorganizedDir, "acteurs")
      fs.mkdirSync(acteursReorganizedDir)
      const organesReorganizedDir = path.join(datasetReorganizedDir, "organes")
      fs.mkdirSync(organesReorganizedDir)

      switch (dataset.structure) {
        case DatasetStructure.SingleFile:
          {
            const originalJson: string = fs.readFileSync(
              originalJsonDirectoryOrFilePath,
              {
                encoding: "utf8",
              },
            )
            const originalData: any = validateRawData
              ? ActeursEtOrganesRawConvert.toActeursEtOrganes(originalJson)
              : JSON.parse(originalJson)

            for (const acteur of originalData.export.acteurs.acteur) {
              const reorganizedJsonFilePath: string = path.join(
                acteursReorganizedDir,
                acteur.uid["#text"] + ".json",
              )
              fs.writeFileSync(
                reorganizedJsonFilePath,
                JSON.stringify(acteur, null, 2),
                {
                  encoding: "utf8",
                },
              )
            }

            for (const organe of originalData.export.organes.organe) {
              const reorganizedJsonFilePath: string = path.join(
                organesReorganizedDir,
                organe.uid + ".json",
              )
              fs.writeFileSync(
                reorganizedJsonFilePath,
                JSON.stringify(organe, null, 2),
                {
                  encoding: "utf8",
                },
              )
            }
          }
          break
        case DatasetStructure.SegmentedFiles:
          {
            const deportsByActeurUid: {
              [acteurUid: string]: { uid: string }[]
            } = {}
            if (
              fs.existsSync(
                path.join(originalJsonDirectoryOrFilePath, "deport"),
              )
            ) {
              for (const deportSplitPath of walkDir(
                originalJsonDirectoryOrFilePath,
                ["deport"],
              )) {
                const deportFilename =
                  deportSplitPath[deportSplitPath.length - 1]
                if (!deportFilename.endsWith(".json")) {
                  continue
                }
                const deportOriginalFilePath = path.join(
                  originalJsonDirectoryOrFilePath,
                  ...deportSplitPath,
                )
                if (options.verbose) {
                  console.log(
                    `  Loading "deport" file: ${deportOriginalFilePath}…`,
                  )
                }
                const deportOriginalJson: string = fs.readFileSync(
                  deportOriginalFilePath,
                  {
                    encoding: "utf8",
                  },
                )
                const deportOriginal: any = validateRawData
                  ? ActeursEtOrganesRawConvert.toDeport(deportOriginalJson)
                  : JSON.parse(deportOriginalJson)
                const deport = deportOriginal.deport
                let deports = deportsByActeurUid[deport.acteurRef]
                if (deports === undefined) {
                  deports = deportsByActeurUid[deport.acteurRef] = []
                }
                deports.push(deport)
              }

              for (const deports of Object.values(deportsByActeurUid)) {
                deports.sort((a, b) =>
                  a.uid.length === b.uid.length
                    ? a.uid.localeCompare(b.uid)
                    : a.uid.length - b.uid.length,
                )
              }
            }

            const mandatsByActeurUid: {
              [acteurUid: string]: { uid: string }[]
            } = {}
            if (
              fs.existsSync(
                path.join(originalJsonDirectoryOrFilePath, "mandat"),
              )
            ) {
              for (const mandatSplitPath of walkDir(
                originalJsonDirectoryOrFilePath,
                ["mandat"],
              )) {
                const mandatFilename =
                  mandatSplitPath[mandatSplitPath.length - 1]
                if (!mandatFilename.endsWith(".json")) {
                  continue
                }
                const mandatOriginalFilePath = path.join(
                  originalJsonDirectoryOrFilePath,
                  ...mandatSplitPath,
                )
                if (options.verbose) {
                  console.log(
                    `  Loading "mandat" file: ${mandatOriginalFilePath}…`,
                  )
                }
                const mandatOriginalJson: string = fs.readFileSync(
                  mandatOriginalFilePath,
                  {
                    encoding: "utf8",
                  },
                )
                const mandatOriginal: any = validateRawData
                  ? ActeursEtOrganesRawConvert.toMandatWrapper(
                      mandatOriginalJson,
                    )
                  : JSON.parse(mandatOriginalJson)
                const mandat = mandatOriginal.mandat
                let mandats = mandatsByActeurUid[mandat.acteurRef]
                if (mandats === undefined) {
                  mandats = mandatsByActeurUid[mandat.acteurRef] = []
                }
                mandats.push(mandat)
              }

              for (const mandats of Object.values(mandatsByActeurUid)) {
                mandats.sort((a, b) =>
                  a.uid.length === b.uid.length
                    ? a.uid.localeCompare(b.uid)
                    : a.uid.length - b.uid.length,
                )
              }
            }

            for (const acteurSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
              ["acteur"],
            )) {
              const acteurFilename = acteurSplitPath[acteurSplitPath.length - 1]
              if (!acteurFilename.endsWith(".json")) {
                continue
              }
              const acteurOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...acteurSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "acteur" file: ${acteurOriginalFilePath}…`,
                )
              }
              const acteurOriginalJson: string = fs.readFileSync(
                acteurOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const acteurOriginal: any = validateRawData
                ? ActeursEtOrganesRawConvert.toActeur(acteurOriginalJson)
                : JSON.parse(acteurOriginalJson)
              const acteur = acteurOriginal.acteur
              const acteurUid = acteur.uid["#text"]
              const deports = deportsByActeurUid[acteurUid]
              if (deports !== undefined) {
                assert.strictEqual(acteur.deports, undefined)
                acteur.deports = { deport: deports }
              }
              const mandats = mandatsByActeurUid[acteurUid]
              if (mandats !== undefined) {
                assert.strictEqual(acteur.mandats, undefined)
                acteur.mandats = { mandat: mandats }
              }
              const acteurReorganizedJsonFilePath: string = path.join(
                acteursReorganizedDir,
                acteurUid + ".json",
              )
              fs.writeFileSync(
                acteurReorganizedJsonFilePath,
                JSON.stringify(acteur, null, 2),
                {
                  encoding: "utf8",
                },
              )
            }

            for (const organeSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
              ["organe"],
            )) {
              const organeFilename = organeSplitPath[organeSplitPath.length - 1]
              if (!organeFilename.endsWith(".json")) {
                continue
              }
              const organeOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...organeSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "organe" file: ${organeOriginalFilePath}…`,
                )
              }
              const organeOriginalJson: string = fs.readFileSync(
                organeOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const organeOriginal: any = validateRawData
                ? ActeursEtOrganesRawConvert.toOrgane(organeOriginalJson)
                : JSON.parse(organeOriginalJson)
              const organe = organeOriginal.organe
              const organeReorganizedJsonFilePath: string = path.join(
                organesReorganizedDir,
                organe.uid + ".json",
              )
              fs.writeFileSync(
                organeReorganizedJsonFilePath,
                JSON.stringify(organe, null, 2),
                {
                  encoding: "utf8",
                },
              )
            }
          }
          break
      }

      commitGit(datasetReorganizedDir, options)
    }
  }

  if (enabledDatasets & EnabledDatasets.Agendas) {
    for (const dataset of chosenDatasets.agendas) {
      const originalJsonDirectoryOrFilePath: string = path.join(
        dataDir,
        dataset.filename,
      )
      if (!options.silent) {
        console.log(`Reorganizing ${originalJsonDirectoryOrFilePath}`)
      }

      const datasetName = path.basename(
        originalJsonDirectoryOrFilePath,
        ".json",
      )
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetReorganizedDir = path.join(dataDir, datasetName)
      clearDirectory(datasetReorganizedDir)

      switch (dataset.structure) {
        case DatasetStructure.SingleFile:
          {
            const originalJson: string = fs.readFileSync(
              originalJsonDirectoryOrFilePath,
              {
                encoding: "utf8",
              },
            )
            const originalData: any = validateRawData
              ? AgendasRawConvert.toAgendas(originalJson)
              : JSON.parse(originalJson)

            for (const reunion of originalData.reunions.reunion) {
              writeReorganizedJson(
                datasetReorganizedDir,
                reunion,
                pathFromReunionUid,
              )
            }
          }
          break
        case DatasetStructure.SegmentedFiles:
          {
            for (const reunionSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
              ["reunion"],
            )) {
              const reunionFilename =
                reunionSplitPath[reunionSplitPath.length - 1]
              if (!reunionFilename.endsWith(".json")) {
                continue
              }
              const reunionOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...reunionSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "reunion" file: ${reunionOriginalFilePath}…`,
                )
              }
              const reunionOriginalJson: string = fs.readFileSync(
                reunionOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const reunionOriginal: any = validateRawData
                ? AgendasRawConvert.toReunion(reunionOriginalJson)
                : JSON.parse(reunionOriginalJson)
              const reunion = reunionOriginal.reunion
              writeReorganizedJson(
                datasetReorganizedDir,
                reunion,
                pathFromReunionUid,
              )
            }
          }
          break
      }

      commitGit(datasetReorganizedDir, options)
    }
  }

  if (enabledDatasets & EnabledDatasets.Amendements) {
    for (const dataset of chosenDatasets.amendements) {
      const originalJsonDirectoryOrFilePath: string = path.join(
        dataDir,
        dataset.filename,
      )
      if (!options.silent) {
        console.log(`Reorganizing ${originalJsonDirectoryOrFilePath}`)
      }

      const datasetName = path.basename(
        originalJsonDirectoryOrFilePath,
        ".json",
      )
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetReorganizedDir = path.join(dataDir, datasetName)
      clearDirectory(datasetReorganizedDir)

      switch (dataset.structure) {
        case DatasetStructure.SingleFile:
          {
            const originalJson: string = fs.readFileSync(
              originalJsonDirectoryOrFilePath,
              {
                encoding: "utf8",
              },
            )
            const originalData: any = validateRawData
              ? AmendementsRawConvert.toAmendements(originalJson)
              : JSON.parse(originalJson)

            for (const texteLegislatif of originalData.textesEtAmendements
              .texteleg) {
              for (const amendement of texteLegislatif.amendements) {
                writeAmendementReorganizedJson(
                  datasetReorganizedDir,
                  amendement,
                )
              }
            }
          }
          break
        case DatasetStructure.SegmentedFiles:
          {
            for (const amendementSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
            )) {
              const amendementFilename =
                amendementSplitPath[amendementSplitPath.length - 1]
              if (!amendementFilename.endsWith(".json")) {
                continue
              }
              const amendementOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...amendementSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "amendement" file: ${amendementOriginalFilePath}…`,
                )
              }
              const amendementOriginalJson: string = fs.readFileSync(
                amendementOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const amendementOriginal: any = validateRawData
                ? AmendementsRawConvert.toAmendementWrapper(
                    amendementOriginalJson,
                  )
                : JSON.parse(amendementOriginalJson)
              const amendement = amendementOriginal.amendement
              writeAmendementReorganizedJson(datasetReorganizedDir, amendement)
            }
          }
          break
      }

      commitGit(datasetReorganizedDir, options)
    }
  }

  if (enabledDatasets & EnabledDatasets.DossiersLegislatifs) {
    for (const dataset of chosenDatasets.dossiersLegislatifs) {
      const originalJsonDirectoryOrFilePath: string = path.join(
        dataDir,
        dataset.filename,
      )
      if (!options.silent) {
        console.log(`Reorganizing ${originalJsonDirectoryOrFilePath}`)
      }

      const datasetName = path.basename(
        originalJsonDirectoryOrFilePath,
        ".json",
      )
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetReorganizedDir = path.join(dataDir, datasetName)
      clearDirectory(datasetReorganizedDir)

      const documentsDir = path.join(datasetReorganizedDir, "documents")
      fs.mkdirSync(documentsDir)
      const dossiersDir = path.join(datasetReorganizedDir, "dossiers")
      fs.mkdirSync(dossiersDir)

      switch (dataset.structure) {
        case DatasetStructure.SingleFile:
          {
            const originalJson: string = fs.readFileSync(
              originalJsonDirectoryOrFilePath,
              {
                encoding: "utf8",
              },
            )
            const originalData: any = validateRawData
              ? DossiersLegislatifsRawConvert.toDossiersLegislatifs(
                  originalJson,
                )
              : JSON.parse(originalJson)

            for (const document of originalData.export.textesLegislatifs
              .document) {
              writeReorganizedJson(documentsDir, document, pathFromDocumentUid)
            }

            for (const { dossierParlementaire } of originalData.export
              .dossiersLegislatifs.dossier) {
              writeReorganizedJson(
                dossiersDir,
                dossierParlementaire,
                pathFromDossierParlementaireUid,
              )
            }
          }
          break
        case DatasetStructure.SegmentedFiles:
          {
            for (const documentSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
              ["document"],
            )) {
              const documentFilename =
                documentSplitPath[documentSplitPath.length - 1]
              if (!documentFilename.endsWith(".json")) {
                continue
              }
              const documentOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...documentSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "document" file: ${documentOriginalFilePath}…`,
                )
              }
              const documentOriginalJson: string = fs.readFileSync(
                documentOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const documentOriginal: any = validateRawData
                ? DossiersLegislatifsRawConvert.toDocument(documentOriginalJson)
                : JSON.parse(documentOriginalJson)
              const document = documentOriginal.document
              writeReorganizedJson(documentsDir, document, pathFromDocumentUid)
            }

            for (const dossierParlementaireSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
              ["dossierParlementaire"],
            )) {
              const dossierParlementaireFilename =
                dossierParlementaireSplitPath[
                  dossierParlementaireSplitPath.length - 1
                ]
              if (!dossierParlementaireFilename.endsWith(".json")) {
                continue
              }
              const dossierParlementaireOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...dossierParlementaireSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "dossierParlementaire" file: ${dossierParlementaireOriginalFilePath}…`,
                )
              }
              const dossierParlementaireOriginalJson: string = fs.readFileSync(
                dossierParlementaireOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const dossierParlementaireOriginal: any = validateRawData
                ? DossiersLegislatifsRawConvert.toDossierParlementaire(
                    dossierParlementaireOriginalJson,
                  )
                : JSON.parse(dossierParlementaireOriginalJson)
              const dossierParlementaire =
                dossierParlementaireOriginal.dossierParlementaire
              writeReorganizedJson(
                dossiersDir,
                dossierParlementaire,
                pathFromDossierParlementaireUid,
              )
            }
          }
          break
      }

      commitGit(datasetReorganizedDir, options)
    }
  }
  if (enabledDatasets & EnabledDatasets.Scrutins) {
    for (let dataset of chosenDatasets.scrutins) {
      const originalJsonDirectoryOrFilePath: string = path.join(
        dataDir,
        dataset.filename,
      )
      if (!options.silent) {
        console.log(`Reorganizing ${originalJsonDirectoryOrFilePath}`)
      }

      const datasetName = path.basename(
        originalJsonDirectoryOrFilePath,
        ".json",
      )
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetReorganizedDir = path.join(dataDir, datasetName)
      clearDirectory(datasetReorganizedDir)

      switch (dataset.structure) {
        case DatasetStructure.SingleFile:
          {
            const originalJson: string = fs.readFileSync(
              originalJsonDirectoryOrFilePath,
              {
                encoding: "utf8",
              },
            )
            const originalData: any = validateRawData
              ? ScrutinsRawConvert.toScrutins(originalJson)
              : JSON.parse(originalJson)

            for (const scrutin of originalData.scrutins.scrutin) {
              writeReorganizedJson(
                datasetReorganizedDir,
                scrutin,
                pathFromScrutinUid,
              )
            }
          }
          break
        case DatasetStructure.SegmentedFiles:
          {
            for (const scrutinSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
            )) {
              const scrutinFilename =
                scrutinSplitPath[scrutinSplitPath.length - 1]
              if (!scrutinFilename.endsWith(".json")) {
                continue
              }
              const scrutinOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...scrutinSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "scrutin" file: ${scrutinOriginalFilePath}…`,
                )
              }
              const scrutinOriginalJson: string = fs.readFileSync(
                scrutinOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const scrutinOriginal: any = validateRawData
                ? ScrutinsRawConvert.toScrutinWrapper(scrutinOriginalJson)
                : JSON.parse(scrutinOriginalJson)
              const scrutin = scrutinOriginal.scrutin
              writeReorganizedJson(
                datasetReorganizedDir,
                scrutin,
                pathFromScrutinUid,
              )
            }
          }
          break
      }

      commitGit(datasetReorganizedDir, options)
    }
  }

  if (enabledDatasets & EnabledDatasets.Questions) {
    for (const dataset of datasets.questions) {
      const originalJsonDirectoryOrFilePath: string = path.join(
        dataDir,
        dataset.filename,
      )
      if (!options.silent) {
        console.log(`Reorganizing ${originalJsonDirectoryOrFilePath}`)
      }

      const datasetName = path.basename(
        originalJsonDirectoryOrFilePath,
        ".json",
      )
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetReorganizedDir = path.join(dataDir, datasetName)
      clearDirectory(datasetReorganizedDir)

      switch (dataset.structure) {
        case DatasetStructure.SegmentedFiles:
          {
            for (const questionSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
            )) {
              const questionFilename =
                questionSplitPath[questionSplitPath.length - 1]
              if (!questionFilename.endsWith(".json")) {
                continue
              }
              const questionOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...questionSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "question" file: ${questionOriginalFilePath}…`,
                )
              }
              const questionOriginalJson: string = fs.readFileSync(
                questionOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )

              const questionOriginal: any = validateRawData
                ? QuestionsRawConvert.toQuestionWrapper(questionOriginalJson)
                : JSON.parse(questionOriginalJson)
              const question = questionOriginal.question
              writeReorganizedJson(
                datasetReorganizedDir,
                question,
                pathFromQuestionUid,
              )
            }
          }
          break
      }

      commitGit(datasetReorganizedDir, options)
    }
  }

  if (enabledDatasets & EnabledDatasets.ComptesRendusSeances) {
    for (const dataset of datasets.comptesRendusSeances) {
      const originalJsonDirectoryOrFilePath: string = path.join(
        dataDir,
        dataset.filename,
      )
      if (!options.silent) {
        console.log(`Reorganizing ${originalJsonDirectoryOrFilePath}`)
      }

      const datasetName = path.basename(originalJsonDirectoryOrFilePath, ".xml")
      cloneGit(gitGroupUrl, datasetName, dataDir)

      const datasetReorganizedDir = path.join(dataDir, datasetName)
      clearDirectory(datasetReorganizedDir)

      switch (dataset.structure) {
        case DatasetStructure.SegmentedFiles:
          {
            for (const compteRenduSeanceSplitPath of walkDir(
              originalJsonDirectoryOrFilePath,
            )) {
              const compteRenduSeanceFilename =
                compteRenduSeanceSplitPath[
                  compteRenduSeanceSplitPath.length - 1
                ]
              if (!compteRenduSeanceFilename.endsWith(".json")) {
                continue
              }
              const compteRenduSeanceOriginalFilePath = path.join(
                originalJsonDirectoryOrFilePath,
                ...compteRenduSeanceSplitPath,
              )
              if (options.verbose) {
                console.log(
                  `  Reorganizing "compteRenduSeance" file: ${compteRenduSeanceOriginalFilePath}…`,
                )
              }
              const compteRenduSeanceOriginalJson: string = fs.readFileSync(
                compteRenduSeanceOriginalFilePath,
                {
                  encoding: "utf8",
                },
              )
              const compteRenduSeanceOriginal: any = validateRawData
                ? ComptesRendusRawConvert.toCompteRenduWrapper(
                    compteRenduSeanceOriginalJson,
                  )
                : JSON.parse(compteRenduSeanceOriginalJson)

              const compteRenduSeance = compteRenduSeanceOriginal.compteRendu
              writeReorganizedJson(
                datasetReorganizedDir,
                compteRenduSeance,
                pathFromCompteRenduUid,
              )
            }
          }
          break
      }

      commitGit(datasetReorganizedDir, options)
    }
  }

  // Restore standard conversion of dates to JSON.
  Date.prototype.toJSON = existingDateToJson

  return exitCode
}

function writeAmendementReorganizedJson(
  datasetReorganizedDir: string,
  amendement: {
    uid: string
    texteLegislatifRef: string
  },
): void {
  const uidTexteLegislatif = amendement.texteLegislatifRef
  assert.notStrictEqual(uidTexteLegislatif, undefined)
  const texteLegislatifDir = path.join(
    datasetReorganizedDir,
    uidTexteLegislatif,
  )
  fs.ensureDirSync(texteLegislatifDir)
  const reorganizedJsonFilePath = path.join(
    texteLegislatifDir,
    amendement.uid + ".json",
  )
  fs.writeFileSync(
    reorganizedJsonFilePath,
    JSON.stringify(amendement, null, 2),
    {
      encoding: "utf8",
    },
  )
}

function writeReorganizedJson<T extends { uid: string }>(
  dir: string,
  obj: T,
  pathFromUid: (dir: string, uid: string) => string,
): void {
  const basename = pathFromUid(dir, obj.uid)
  fs.writeFileSync(`${basename}.json`, JSON.stringify(obj, null, 2), {
    encoding: "utf8",
  })
}

function main(argv: any): number {
  const options = parseArgs(argv)
  return reorganizeData(options)
}

/* istanbul ignore if */
if (process.argv[1].endsWith("reorganize_data.ts"))
  process.exit(main(process.argv))
