import { execSync } from "child_process"
import commandLineArgs from "command-line-args"
import fs from "fs-extra"
import path from "path"
import stream from "stream"
import util from "util"
import * as git from "../git"

import { iterLoadAssembleeActeurs } from "../loaders"
import {
  Acteur,
  CodeTypeOrgane,
  Mandat,
  Photo,
  TypeMandat,
} from "../types/acteurs_et_organes"
import {
  cloneOption,
  commitOption,
  remoteOption,
  silentOption,
  dataDirDefaultOption,
  pullOption,
  legislatureOption,
  fetchOption,
} from "./shared/cli_helpers"

const optionsDefinitions = [
  cloneOption,
  commitOption,
  remoteOption,
  silentOption,
  dataDirDefaultOption,
  pullOption,
  fetchOption,
  legislatureOption,
  {
    alias: "S",
    defaultValue: false,
    name: "square",
    type: Boolean,
  },
  {
    alias: "w",
    defaultValue: 150,
    name: "width",
    type: Number,
  },
  {
    alias: "h",
    defaultValue: 192,
    name: "height",
    type: Number,
  },
]
const options = commandLineArgs(optionsDefinitions)

export function cloneAndUpdateGitRepository(
  photosDirName: string,
  options: any,
): void {
  const gitGroupUrl = options.clone
    ? options.clone.trim().replace(/\/+$/, "")
    : undefined
  const photosDir = path.join(options.dataDir, photosDirName)
  git.clone(gitGroupUrl, photosDirName, options.dataDir)

  if (options.pull) {
    git.resetAndPull(photosDir)
  }
}

export function resizePhoto(
  acteur: Acteur,
  width: number,
  height: number,
  photosDir: string,
  photoFilename: string,
  options: any,
): void {
  const ident = acteur.etatCivil.ident
  if (!options.silent) {
    console.log(
      `Resizing photo ${photoFilename} for ${ident.prenom} ${ident.nom}…`,
    )
  }
  execSync(
    `gm convert -resize ${width}x${height}! ${photoFilename}.jpg ${photoFilename}_${width}x${height}.jpg`,
    {
      cwd: photosDir,
    },
  )
}

export async function downloadPhoto(
  urlPhoto: string,
  photoTempFilePath: string,
  photoFilePath: string,
  missingPhotoFilePath: string,
  retries: number = 3,
): Promise<void> {
  const pipeline = util.promisify(stream.pipeline)

  for (let attempt = 1; attempt <= retries; attempt++) {
    try {
      const response = await fetch(urlPhoto)

      if (response.ok) {
        await pipeline(
          response.body as unknown as NodeJS.ReadableStream,
          fs.createWriteStream(photoTempFilePath),
        )
        fs.renameSync(photoTempFilePath, photoFilePath)
        return
      }

      console.warn(
        `Fetch failed: ${urlPhoto} (Attempt ${attempt}) - Status: ${response.status} ${response.statusText}`,
      )
    } catch (error) {
      console.log(error)
      console.warn(`Attempt ${attempt} failed for ${urlPhoto}`, error)
    }
  }

  if (fs.existsSync(photoFilePath)) {
    console.warn(`Reusing existing image: ${photoFilePath}`)
  } else {
    console.warn(`Using blank image: ${missingPhotoFilePath}`)
    fs.copyFileSync(missingPhotoFilePath, photoFilePath)
  }
}

export function createMosaic(
  acteurUids: string[],
  photosDir: string,
  photosWidth: number,
  photosHeight: number,
  mosaicFilename: string,
): void {
  const photoByIdActeur: { [uid: string]: Photo } = {}
  const rowsFilenames: string[] = []

  if (acteurUids.length === 0) return

  for (
    let acteurIndex = 0, rowIndex = 0;
    acteurIndex < acteurUids.length;
    acteurIndex += 25, rowIndex++
  ) {
    const row = acteurUids.slice(acteurIndex, acteurIndex + 25)
    const photosFilenames: string[] = row.map((acteurUid, columnIndex) => {
      const photoFilename = `${acteurUid}_${photosWidth}x${photosHeight}.jpg`
      photoByIdActeur[acteurUid] = {
        chemin: `${photosDir}/${photoFilename}`,
        cheminMosaique: `${photosDir}/${mosaicFilename}.jpg`,
        largeur: photosWidth,
        hauteur: photosHeight,
        xMosaique: columnIndex * photosWidth,
        yMosaique: rowIndex * photosHeight,
      }
      return photoFilename
    })

    const rowFilename = `row-${rowIndex}.jpg`
    execSync(`gm convert ${photosFilenames.join(" ")} +append ${rowFilename}`, {
      cwd: photosDir,
    })
    rowsFilenames.push(rowFilename)
  }

  // Combine all rows into one image
  execSync(
    `gm convert ${rowsFilenames.join(" ")} -append ${mosaicFilename}.jpg`,
    { cwd: photosDir },
  )

  // Cleanup individual row images
  rowsFilenames.forEach((rowFilename) => {
    fs.unlinkSync(path.join(photosDir, rowFilename))
  })

  const jsonFilePath = path.join(photosDir, `${mosaicFilename}.json`)
  fs.writeFileSync(jsonFilePath, JSON.stringify(photoByIdActeur, null, 2))
}

async function retrievePhotosDeputes(): Promise<number> {
  const dataDir = options.dataDir
  const photosPrefix = "photos_deputes"
  const photosDirName = `${photosPrefix}_${options.legislature}`
  const photosDir = path.join(dataDir, photosDirName)
  const photosWidth = options.width
  const photosHeight = options.height
  const missingPhotoFilePath = `images/transparent_${photosWidth}x${photosHeight}.jpg`

  cloneAndUpdateGitRepository(photosDirName, options)
  fs.ensureDirSync(photosDir)

  const acteurUids = []
  for (const { acteur } of iterLoadAssembleeActeurs(
    dataDir,
    options.legislature,
    { noValidation: true },
  )) {
    const { mandats } = acteur
    if (!mandats) continue

    const mandatDepute = mandats.find(
      (mandat: Mandat) =>
        mandat.xsiType === TypeMandat.MandatParlementaireType &&
        mandat.legislature === options.legislature?.toString() &&
        mandat.typeOrgane === CodeTypeOrgane.Assemblee,
    )
    if (!mandatDepute) continue

    const numeroDepute = acteur.uid.substring(2)

    // Download photos.
    if (options.fetch) {
      const ident = acteur.etatCivil.ident
      const photoFilename = `${numeroDepute}.jpg`
      const photoFilePath = path.join(photosDir, photoFilename)
      const photoTempFilename = `${numeroDepute}_temp.jpg`
      const photoTempFilePath = path.join(photosDir, photoTempFilename)
      let urlPhoto = `http://www2.assemblee-nationale.fr/static/tribun/${options.legislature}/photos/${photoFilename}`
      if (options.square) {
        urlPhoto = `https://www2.assemblee-nationale.fr/static/tribun/${options.legislature}/photos/carre/${photoFilename}`
      }

      if (!options.silent) {
        console.log(
          `Loading photo ${urlPhoto} for ${ident.prenom} ${ident.nom}…`,
        )
      }

      await downloadPhoto(
        urlPhoto,
        photoTempFilePath,
        photoFilePath,
        missingPhotoFilePath,
      )
      resizePhoto(
        acteur,
        photosWidth,
        photosHeight,
        photosDir,
        numeroDepute,
        options,
      )
    }

    acteurUids.push(numeroDepute)
  }

  createMosaic(acteurUids, photosDir, photosWidth, photosHeight, "deputes")

  if (options.commit) {
    return git.commitAndPush(photosDir, "Nouvelle moisson", options.remote)
  }
  return 0
}

retrievePhotosDeputes()
  .then((exitCode) => process.exit(exitCode))
  .catch((error) => {
    console.log(error)
    process.exit(1)
  })
