import assert from "assert"
import commandLineArgs from "command-line-args"
import { createHash } from "crypto"
import { differenceInDays } from "date-fns"
import fs from "fs-extra"
import path from "path"

import {
  DocumentFileInfos,
  DocumentFilesIndex,
  walkDocumentAndDivisions,
} from "../dossiers_legislatifs"
import * as git from "../git"
import { iterLoadAssembleeDocuments, pathFromDocumentUid } from "../loaders"
import { parseTexte } from "../parsers"
import { DocumentUrlFormat, iterDocumentOrDivisionUrls } from "../urls"
import {
  commitOption,
  dataDirDefaultOption,
  legislatureOption,
  pullOption,
  remoteOption,
  silentOption,
  verboseOption,
  documentFormatOption,
} from "./shared/cli_helpers"

function parseArgs(argv: string[]): any {
  const optionsDefinitions = [
    commitOption,
    legislatureOption,
    remoteOption,
    silentOption,
    verboseOption,
    dataDirDefaultOption,
    pullOption,
    {
      alias: "f",
      help: "retrieve all documents, even already retrieved ones",
      name: "full",
      type: Boolean,
    },
    {
      alias: "n",
      help: "try to also retrieve documents that were previously not found",
      name: "not-found",
      type: Boolean,
    },
    documentFormatOption,
    {
      alias: "T",
      help: "type of documents to retrieve (for example: PION)",
      multiple: true,
      name: "document-type",
      type: String,
    },
  ]
  const options = commandLineArgs(optionsDefinitions, {
    argv: argv,
  })
  return options
}

const today = new Date()

async function fetchWithRetry(
  url: string,
  retries: number = 3,
  backoff: number = 300,
): Promise<Response> {
  for (let attempt = 0; attempt < retries; attempt++) {
    try {
      return await fetch(url)
    } catch (error) {
      if (attempt === retries - 1) {
        throw error
      }
      console.warn(
        `Fetch attempt ${attempt + 1} for ${url} failed. Retrying in ${backoff}ms...`,
      )
      await new Promise((resolve) => setTimeout(resolve, backoff))
      backoff *= 2
    }
  }
  throw new Error(`Failed to fetch ${url} after ${retries} attempts`)
}

export async function downloadAndParse(
  document: any,
  documentsDir: string,
  options: any,
): Promise<void> {
  if (document.uid.substring(4, 6) === "SN") {
    return
  }

  if (["RAPP", "RINF"].includes(document.uid.substring(0, 4))) {
    return
  }

  if (options.legislature?.toString() !== document.legislature) {
    return
  }

  if (options.fetchDocuments) {
    const dataDir = options.dataDir
    const documentsFilesDir = path.join(dataDir, "Documents")
    const documentDir = pathFromDocumentUid(
      `${documentsDir}/documents`,
      document.uid,
    )
    const documentFileDir = pathFromDocumentUid(documentsFilesDir, document.uid)

    fs.ensureDirSync(documentsFilesDir)
    await processDocumentOrDivision(document, documentsFilesDir, options)

    if (options.parseDocuments) {
      const documentPath = `${documentDir}.json`
      const htmlPath = path.join(documentFileDir, `dyn-opendata.html`)

      if (fs.existsSync(htmlPath)) {
        const html = fs.readFileSync(htmlPath, { encoding: "utf-8" })
        try {
          const parsedDocument = parseTexte(
            "https://www.assemblee-nationale.fr",
            html,
          )
          if (parsedDocument && fs.existsSync(documentPath)) {
            const json = fs.readFileSync(documentPath, { encoding: "utf-8" })
            const documentJson = JSON.parse(json)
            documentJson.subdivisions = parsedDocument?.subdivisions
            fs.writeFileSync(
              documentPath,
              JSON.stringify(documentJson, null, 2),
            )
          }
        } catch (e) {
          console.log(`Unable to parse document ${document.uid}`)
          return
        }
      }
    }
  }
}

async function retrieveDocuments(options: any): Promise<number> {
  assert(
    !options.commit || !options.uid,
    'Options "commit" & "uid" are incompatible',
  )

  const dataDir = options.dataDir
  const documentsDir = path.join(dataDir, "Documents")
  if (options.pull) {
    git.resetAndPull(documentsDir)
  }
  fs.ensureDirSync(documentsDir)

  if (options.full && !options.uid) {
    for (const filename of fs.readdirSync(documentsDir)) {
      if (filename[0] === ".") {
        continue
      }
      fs.removeSync(path.join(documentsDir, filename))
    }
  }

  const firstUid = options.uid
  let skip = Boolean(firstUid)
  for (const { document } of iterLoadAssembleeDocuments(
    dataDir,
    options.legislature,
  )) {
    for (const documentOrDivision of walkDocumentAndDivisions(document)) {
      // Ignore documents from Sénat.
      if (documentOrDivision.uid.substring(4, 6) === "SN") {
        continue
      }

      if (skip) {
        if (documentOrDivision.uid === firstUid) {
          skip = false
        } else {
          continue
        }
      }

      await processDocumentOrDivision(documentOrDivision, documentsDir, options)
    }
  }

  if (options.commit) {
    return git.commitAndPush(documentsDir, "Nouvelle moisson", options.remote)
  }
  return 0
}

async function processDocumentOrDivision(
  documentOrDivision: any,
  documentsDir: string,
  options: any,
): Promise<void> {
  const documentDir = pathFromDocumentUid(documentsDir, documentOrDivision.uid)
  fs.ensureDirSync(documentDir)

  const filenameBySha256: { [digest: string]: string } = {}
  const indexPath = path.join(documentDir, "index.json")
  const index = (
    fs.pathExistsSync(indexPath) ? fs.readJsonSync(indexPath) : {}
  ) as DocumentFilesIndex

  for (const { format, type, url } of iterDocumentOrDivisionUrls(
    documentOrDivision,
  )) {
    const filename = `${type}.${format === DocumentUrlFormat.Pdf ? "pdf" : "html"}`

    // Filter by file extension if option is passed
    const fileExtension = url.split(".").pop()
    if (options.format && fileExtension && options.format !== fileExtension) {
      continue
    }

    if (
      options.legislature &&
      options.legislature.toString() !== documentOrDivision.legislature
    ) {
      continue
    }

    // Filter by document type if option is passed
    const documentType = documentOrDivision.classification?.type?.code
    if (
      options["document-type"] !== undefined &&
      !options["document-type"].includes(documentType)
    ) {
      continue
    }

    let formatFilesInfos = index[format] ?? (index[format] = [])
    let fileInfos =
      formatFilesInfos.find((file) => file.url === url) ??
      ({} as DocumentFileInfos)
    if (!formatFilesInfos.includes(fileInfos)) {
      formatFilesInfos.push(fileInfos)
    }
    fileInfos.url = url

    if (fileInfos.status === 200 && !options.full) {
      filenameBySha256[fileInfos.sha256 as string] = filename
      continue
    }
    if (
      fileInfos.status === 404 &&
      !options["not-found"] &&
      differenceInDays(
        today,
        documentOrDivision.cycleDeVie.chrono.dateCreation ??
          (documentOrDivision.cycleDeVie.chrono.dateDepot as Date | string),
      ) > 10
    ) {
      continue
    }

    if (!options.silent) {
      console.log(
        `Retrieving document or division ${documentOrDivision.uid} at ${url}…`,
      )
    }

    const response = await fetchWithRetry(url)
    const filePath = path.join(documentDir, filename)

    if (response.ok) {
      const arrayBuffer = await response.arrayBuffer()
      const buffer = Buffer.from(arrayBuffer)
      if (
        format === DocumentUrlFormat.Pdf &&
        !buffer.subarray(0, 4).toString().startsWith("%PDF")
      ) {
        // Instead of a PDF, the received data may be an HTML page with a message like
        // "Document non encore publié".
        if (!options.silent) {
          console.warn(`  PDF "${url}" not found.`)
        }
        fs.removeSync(filePath)
        delete fileInfos.filename
        delete fileInfos.sha256
        fileInfos.status = 404
      } else {
        const sha256 = createHash("sha256").update(buffer).digest("hex")
        const existingFilename = filenameBySha256[sha256]
        if (existingFilename === undefined) {
          fs.writeFileSync(filePath, buffer, {
            encoding: "utf8",
          })
          fileInfos.filename = filename
          filenameBySha256[sha256] = filename
        } else {
          fileInfos.filename = existingFilename
        }
        fileInfos.sha256 = sha256
        fileInfos.status = response.status
      }
    } else {
      if (response.status === 404) {
        if (!options.silent) {
          console.warn(`  Page "${url}" not found.`)
        }
      } else {
        console.error(
          `  Error:\n${JSON.stringify(
            { code: response.status, message: response.statusText },
            null,
            2,
          )}`,
        )
      }
      fs.removeSync(filePath)
      delete fileInfos.filename
      delete fileInfos.sha256
      fileInfos.status = response.status
    }
  }

  fs.writeJsonSync(indexPath, index, { encoding: "utf-8", spaces: 2 })
}

function main(argv: any): Promise<number> {
  const options = parseArgs(argv)
  return retrieveDocuments(options)
}

/* istanbul ignore if */
if (process.argv[1].endsWith("retrieve_documents.ts")) {
  main(process.argv)
    .then((exitCode) => process.exit(exitCode))
    .catch((error) => {
      console.log(error)
      process.exit(1)
    })
}
