import commandLineArgs from "command-line-args"
import deepEqual from "deep-equal"
import fs from "fs-extra"
import path from "path"

import * as git from "../git"
import { iterRechercheRawAmendements } from "../parsers"
import {
  cloneOption,
  commitOption,
  dataDirDefaultOption,
  legislatureOption,
  pullOption,
  remoteOption,
  silentOption,
  verboseOption,
} from "./shared/cli_helpers"

const optionsDefinitions = [
  cloneOption,
  commitOption,
  legislatureOption,
  remoteOption,
  silentOption,
  verboseOption,
  dataDirDefaultOption,
  pullOption,
  {
    alias: "i",
    help: "incremental mode: retrieve only amendments whose position in search has changed",
    name: "incremental",
    type: Boolean,
  },
]
const options = commandLineArgs(optionsDefinitions)

async function main(): Promise<number> {
  const gitGroupUrl = options.clone
    ? options.clone.trim().replace(/\/+$/, "")
    : undefined

  git.clone(gitGroupUrl, "Amendements_en_attente", options.dataDir)

  const amendementsDir = path.join(options.dataDir, "Amendements_en_attente")
  if (options.pull) {
    git.resetAndPull(amendementsDir)
  }

  await fs.ensureDir(amendementsDir)
  const cacheDir = path.join(options.dataDir, "Amendements_en_attente_cache")
  for (const etat of ["En recevabilité", "En traitement"]) {
    const url = `https://www.assemblee-nationale.fr/dyn/${
      options.legislature
    }/amendements?etat=${encodeURIComponent(etat)}`
    for await (const [
      amendementUrlPath,
      amendement,
    ] of iterRechercheRawAmendements(
      cacheDir,
      url,
      options.incremental,
      null,
      null,
      { silent: options.silent, verbose: options.verbose },
    )) {
      const amendementUrlPathSplitted = amendementUrlPath!.split("/")
      const amendementFilePath =
        path.join(amendementsDir, ...amendementUrlPathSplitted.slice(2)) +
        ".json"
      await fs.ensureDir(path.dirname(amendementFilePath))
      const existingAmendement = (await fs.pathExists(amendementFilePath))
        ? await fs.readJson(amendementFilePath)
        : null
      if (!deepEqual(amendement, existingAmendement)) {
        if (!options.silent) {
          if (existingAmendement === null) {
            console.log(`Adding amendement JSON: ${amendementFilePath}…`)
          } else {
            console.log(`Updating amendement JSON: ${amendementFilePath}…`)
          }
        }
        await fs.writeJson(amendementFilePath, amendement, { spaces: 2 })
      }
    }
  }

  if (options.commit) {
    return git.commitAndPush(amendementsDir, "Nouvelle moisson", options.remote)
  }
  return 0
}

main()
  .then((exitCode) => process.exit(exitCode))
  .catch((error) => {
    console.log(error)
    process.exit(1)
  })
