import commandLineArgs from "command-line-args"
import fs from "fs-extra"
import path from "path"
import * as git from "../git"
import { execSync } from "child_process"
import { iterLoadAssembleeActeurs } from "../loaders"
import {
  cloneAndUpdateGitRepository,
  resizePhoto,
  createMosaic,
} from "./retrieve_deputes_photos"
import {
  Acteur,
  CodeTypeOrgane,
  Mandat,
  TypeMandat,
} from "../types/acteurs_et_organes"
import {
  cloneOption,
  commitOption,
  dataDirDefaultOption,
  fetchOption,
  pullOption,
  remoteOption,
  silentOption,
} from "./shared/cli_helpers"
import { Legislature } from "../types/legislatures"

const optionsDefinitions = [
  cloneOption,
  commitOption,
  pullOption,
  remoteOption,
  silentOption,
  dataDirDefaultOption,
  fetchOption,
  {
    alias: "w",
    defaultValue: 155,
    name: "width",
    type: Number,
  },
  {
    alias: "h",
    defaultValue: 225,
    name: "height",
    type: Number,
  },
  {
    alias: "S",
    help: "fetch square picture",
    name: "square",
    type: Boolean,
  },
]
const options = commandLineArgs(optionsDefinitions)

async function retrievePhotosSenateurs(): Promise<number> {
  const dataDir = options.dataDir
  const photosDirName = "photos_senateurs"
  const photosDir = path.join(dataDir, photosDirName)
  const photosWidth = options.width
  const photosHeight = options.height
  const missingPhotoFilePath = `images/transparent_${photosWidth}x${photosHeight}.jpg`

  cloneAndUpdateGitRepository(photosDirName, options)
  fs.ensureDirSync(photosDir)

  const acteurUids = []
  for (const { acteur } of iterLoadAssembleeActeurs(dataDir, Legislature.All, {
    noValidation: true,
  })) {
    const { mandats } = acteur
    if (!mandats) continue

    const mandatDepute = mandats.find(
      (mandat: Mandat) =>
        mandat.xsiType === TypeMandat.MandatParlementaireType &&
        mandat.typeOrgane === CodeTypeOrgane.Senat,
    )
    if (!mandatDepute) continue

    const { adresses } = acteur
    if (!adresses) continue

    let url = null
    for (const adresse of adresses) {
      if (
        adresse.xsiType === "AdresseSiteWeb_Type" &&
        adresse.typeLibelle === "Url sénateur"
      ) {
        url = adresse.valElec
      }
    }

    if (!url) continue

    const senateurStem = `${stemFromSenateur(acteur)}${options.square ? "_carre" : ""}`

    if (options.fetch) {
      const ident = acteur.etatCivil.ident
      const photoFilename = `${senateurStem}.jpg`
      const photoFilePath = path.join(photosDir, photoFilename)
      const photoTempFilename = `${senateurStem}_temp.jpg`
      const photoTempFilePath = path.join(photosDir, photoTempFilename)
      const urlPhoto = `https://www.senat.fr/senimg/${photoFilename}`
      if (!options.silent) {
        console.log(
          `Loading photo ${urlPhoto} for ${ident.prenom} ${ident.nom}…`,
        )
      }

      try {
        execSync(`wget --quiet -O ${photoTempFilename} ${urlPhoto}`, {
          cwd: photosDir,
          env: process.env,
          encoding: "utf-8",
          // stdio: ["ignore", "ignore", "pipe"],
        })
        fs.renameSync(photoTempFilePath, photoFilePath)
      } catch (error) {
        console.error(error)
        if (fs.existsSync(photoFilePath)) {
          console.warn(`Reusing existing image: ${photoFilePath}`)
        } else {
          console.warn(`Using blank image: ${missingPhotoFilePath}`)
          fs.copyFileSync(missingPhotoFilePath, photoFilePath)
        }
      }

      resizePhoto(
        acteur,
        photosWidth,
        photosHeight,
        photosDir,
        senateurStem,
        options,
      )
    }

    acteurUids.push(senateurStem)
  }

  createMosaic(acteurUids, photosDir, photosWidth, photosHeight, "senateurs")

  if (options.commit) {
    return git.commitAndPush(photosDir, "Nouvelle moisson", options.remote)
  }
  return 0
}

function stemFromSenateur(senateur: Acteur) {
  const { adresses } = senateur
  let url = null
  for (const adresse of adresses!) {
    if (
      adresse.xsiType === "AdresseSiteWeb_Type" &&
      adresse.typeLibelle === "Url sénateur"
    ) {
      url = adresse.valElec
    }
  }
  const match = url!.match(/^https:\/\/www\.senat\.fr\/senateur\/(.*)\.html$/)
  const senateurStem = match![1]
  return senateurStem
}

retrievePhotosSenateurs()
  .then((exitCode) => process.exit(exitCode))
  .catch((error) => {
    console.log(error)
    process.exit(1)
  })
