import Ajv from "ajv"
import commandLineArgs from "command-line-args"
import commandLineUsage from "command-line-usage"
import { load, getFiles } from "../file_systems"
import * as git from "../git"
import temp from "temp"
import fs from "fs-extra"
import path from "path"
import {
  validDataset,
  getDatasets,
  getSchemas,
  datasetDirectorySchema,
} from "../../src/datasets"

// Automatically track and cleanup files at exit
temp.track()

function parseArgs(argv: string[]): any {
  const optionsDefinitions = [
    {
      description: "pathname to the directory containing the dataset",
      defaultOption: true,
      name: "data",
      type: String,
    },
    {
      description: "URL or path to the tricoteuses-assemblee git repository",
      defaultValue:
        "https://git.en-root.org/tricoteuses/tricoteuses-assemblee.git",
      name: "repository",
      type: String,
    },
    {
      description: `dataset (possible values must include exactly one of ${getDatasets()})`,
      name: "dataset",
      type: String,
    },
    {
      description: `schema (possible values are ${getSchemas()})`,
      name: "schema",
      type: String,
    },
    {
      description:
        "ignore schemaVersion from JSON files and use the current directory instead of the --repository",
      name: "dev",
      type: Boolean,
    },
    {
      description: "increase verbosity",
      name: "verbose",
      type: Boolean,
    },
    {
      name: "help",
      description: "Print this usage guide.",
    },
  ]
  const sections = [
    {
      header: "Validate JSON files",
      content: "Validate JSON files",
    },
    {
      header: "Options",
      optionList: optionsDefinitions,
    },
  ]
  const options = commandLineArgs(optionsDefinitions, {
    argv: argv,
  })

  if (options.dataset) {
    if (!validDataset(options.dataset)) {
      console.error(`--dataset ${options.dataset}`)
      options.help = true
    }
  }

  if (options.schema) {
    if (!getSchemas().includes(options.schema)) {
      console.error(`--schema ${options.schema}`)
      options.help = true
    }
  }

  if ("help" in options) {
    const usage = commandLineUsage(sections)
    console.warn(usage)
    return null
  }
  return options
}

class Validate {
  options: any
  version2schema: any
  tmpDir: string

  constructor(options: any) {
    this.options = options
    this.version2schema = {}
    this.tmpDir = temp.mkdirSync("validation")
  }

  getSchemaDir(tag: string) {
    if (this.options.dev == true) {
      if (this.options.verbose) console.log("--dev: use src/schemas")
      return "src/schemas"
    }
    const repositoryDir = path.join(this.tmpDir, "tricoteuses-assemblee")
    const tagDir = path.join(this.tmpDir, tag)
    if (!fs.existsSync(tagDir)) {
      if (!fs.existsSync(repositoryDir)) {
        git.run(
          this.tmpDir,
          `clone --quiet ${this.options.repository} ${repositoryDir}`,
        )
      }
      git.run(
        repositoryDir,
        `checkout tags/${tag} -- src/schemas > /dev/null 2>&1`,
      )
      fs.renameSync(path.join(repositoryDir, "src/schemas"), tagDir)
    }
    return tagDir
  }

  getValidator(version: string): any {
    if (!(version in this.version2schema)) {
      const ajv = new Ajv()
      const dir = this.getSchemaDir(`schema-${version}`)
      let references = []
      for (const reference of getFiles([
        `${dir}/${this.options.schema}/*.json`,
      ]))
        references.push(reference)
      for (const file of getFiles(references)) {
        if (this.options.verbose) console.log(`reference: ${file}`)
        ajv.addSchema(load(file))
      }
      const schema = this.options.schema
      const Schema = schema.charAt(0).toUpperCase() + schema.slice(1)
      const s = `${dir}/${schema}/${Schema}.json`
      if (this.options.verbose) console.log(`schema ${s}`)
      this.version2schema[version] = ajv.compile(load(s))
    }
    return this.version2schema[version]
  }

  getVersion(content: any): string {
    if (content.schemaVersion != undefined) return content.schemaVersion
    return `${this.options.schema}-1.0`
  }

  getSchemaURL(version: string): any {
    if (this.options.dev == true) return "src/schemas"
    else {
      if (fs.existsSync(this.options.repository)) {
        return `${this.options.repository}/src/schemas at tag schema-${version}`
      } else {
        const repository = this.options.repository.replace(/\.git$/, "")
        return `${repository}/tree/schema-${version}/src/schemas`
      }
    }
  }

  validate(content: any, filename: string): number {
    const tag = this.getVersion(content)
    const validator = this.getValidator(tag)
    const valid = validator(content)
    const schemaURL = this.getSchemaURL(tag)
    if (!valid) {
      console.error(`${schemaURL} finds ${filename} is invalid`)
      console.error(validator.errors)
      return 1
    } else {
      if (this.options.verbose)
        console.log(`${schemaURL} finds ${filename} is valid`)
      return 0
    }
  }

  async process(): Promise<number> {
    let status = 0
    for (const file of getFiles([this.options.data])) {
      const content = load(file)
      status |= this.validate(content, file)
    }
    return status
  }
}

async function main(argv: any): Promise<number> {
  const options = parseArgs(argv)
  if (options === null) return 1
  let directorySchema = []
  if (options.schema) {
    directorySchema = [[options.data, options.schema]]
  } else if (options.dataset) {
    directorySchema = datasetDirectorySchema(options.dataset)
  } else {
    console.error("either --schema or --dataset is required")
    return 1
  }
  for (const d of directorySchema) {
    options.schema = d[1]
    options.data = d[0]
    if (options.verbose)
      console.log(
        `working on files ${options.data} with schema ${options.schema}`,
      )
    const validate = new Validate(options)
    const status = await validate.process()
    if (status != 0) return status
  }
  return 0
}

/* istanbul ignore if */
if (process.argv[1].endsWith("validate_json.ts"))
  main(process.argv)
    .then((exitCode) => {
      process.exit(exitCode)
    })
    .catch((error) => {
      console.error(error)
      process.exit(1)
    })
