// To parse this data:
//
//   import { Convert, Questions } from "./file";
//
//   const questions = Convert.toQuestions(json);
//
// These functions will throw an error if the JSON doesn't
// match the expected interface, even if the JSON is valid.

export interface Questions {
  questions: Question[]
}

export interface Question {
  schemaVersion?: string
  "@xsi:type": QuestionXsiType
  uid: string
  identifiant: Identifiant
  type: Type
  indexationAN: IndexationAn
  auteur: Auteur
  minInt: Groupe
  minAttribs: MinAttribs
  textesQuestion?: TextesQuestion | null
  textesReponse?: TextesReponse | null
  cloture?: Cloture | null
  signalement?: Signalement | null
  renouvellements?: Renouvellements | null
}

export enum QuestionXsiType {
  QuestionEcrite = "QuestionEcrite_Type",
  QuestionGouvernement = "QuestionGouvernement_Type",
  QuestionOrale = "QuestionOrale_Type",
}

export interface Auteur {
  identite: Identite
  groupe: Groupe
}

export interface Groupe {
  organeRef: string
  abrege: string
  developpe: string
}

export interface Identite {
  acteurRef: string
  mandatRef: string
}

export interface Cloture {
  codeCloture: string
  libelleCloture: string
  dateCloture: string
  infoJO: InfoJo | null
}

export enum TypeJo {
  JoQuestion = "JO_QUESTION",
  JoLoiDecret = "JO_LOI_DECRET",
  JoDebate = "JO_DEBAT",
}

export interface Identifiant {
  numero: string
  regime: string
  legislature: string
}

export interface IndexationAn {
  rubrique: string | null
  teteAnalyse: null
  analyses: Analyse
}

export interface Analyse {
  analyse: string | null
}

export interface MinAttribs {
  minAttrib: MinAttribElement[] | MinAttribElement
}

export interface MinAttribElement {
  infoJO: InfoJo
  denomination: Groupe
}

export interface TextesQuestion {
  texteQuestion: Texte[] | Texte
}

export interface Texte {
  infoJO: InfoJo
  texte: string | null
  erratum?: string
}

export interface InfoJo {
  typeJO: TypeJo | null
  dateJO: string
  pageJO: string | null
  numJO: string | null
  urlLegifrance: string | null
  referenceNOR: string | null
}

export interface TextesReponse {
  texteReponse: Texte[] | Texte
}

export interface Signalement {
  dateDepotSignal: string
  infoJO: InfoJo
}

export interface Renouvellements {
  renouvellement: Renouvellement[] | Renouvellement
}

export interface Renouvellement {
  infoJO: InfoJo
}

export enum Type {
  Qe = "QE",
  Qg = "QG",
  Qosd = "QOSD",
}

// Converts JSON strings to/from your types
// and asserts the results of JSON.parse at runtime
export class Convert {
  public static toQuestions(json: string): Questions {
    return cast(JSON.parse(json), r("Questions"))
  }

  public static questionsToJson(value: Questions): string {
    return JSON.stringify(uncast(value, r("Questions")), null, 2)
  }

  public static toQuestion(json: string): Question {
    return cast(JSON.parse(json), r("Question"))
  }

  public static questionToJson(value: Question): string {
    return JSON.stringify(uncast(value, r("Question")), null, 2)
  }
}

function invalidValue(typ: any, val: any, key: any = ""): never {
  if (key) {
    throw Error(
      `Invalid value for key "${key}". Expected type ${JSON.stringify(
        typ,
      )} but got ${JSON.stringify(val)}`,
    )
  }
  throw Error(
    `Invalid value ${JSON.stringify(val)} for type ${JSON.stringify(typ)}`,
  )
}

function jsonToJSProps(typ: any): any {
  if (typ.jsonToJS === undefined) {
    const map: any = {}
    typ.props.forEach((p: any) => (map[p.json] = { key: p.js, typ: p.typ }))
    typ.jsonToJS = map
  }
  return typ.jsonToJS
}

function jsToJSONProps(typ: any): any {
  if (typ.jsToJSON === undefined) {
    const map: any = {}
    typ.props.forEach((p: any) => (map[p.js] = { key: p.json, typ: p.typ }))
    typ.jsToJSON = map
  }
  return typ.jsToJSON
}

function transform(val: any, typ: any, getProps: any, key: any = ""): any {
  function transformPrimitive(typ: string, val: any): any {
    if (typeof typ === typeof val) return val
    return invalidValue(typ, val, key)
  }

  function transformUnion(typs: any[], val: any): any {
    // val must validate against one typ in typs
    const l = typs.length
    for (let i = 0; i < l; i++) {
      const typ = typs[i]
      try {
        return transform(val, typ, getProps)
      } catch (_) {}
    }
    return invalidValue(typs, val)
  }

  function transformEnum(cases: string[], val: any): any {
    if (cases.indexOf(val) !== -1) return val
    return invalidValue(cases, val)
  }

  function transformArray(typ: any, val: any): any {
    // val must be an array with no invalid elements
    if (!Array.isArray(val)) return invalidValue("array", val)
    return val.map((el) => transform(el, typ, getProps))
  }

  function transformDate(val: any): any {
    if (val === null) {
      return null
    }
    const d = new Date(val)
    if (isNaN(d.valueOf())) {
      return invalidValue("Date", val)
    }
    return d
  }

  function transformObject(
    props: { [k: string]: any },
    additional: any,
    val: any,
  ): any {
    if (val === null || typeof val !== "object" || Array.isArray(val)) {
      return invalidValue("object", val)
    }
    const result: any = {}
    Object.getOwnPropertyNames(props).forEach((key) => {
      const prop = props[key]
      const v = Object.prototype.hasOwnProperty.call(val, key)
        ? val[key]
        : undefined
      result[prop.key] = transform(v, prop.typ, getProps, prop.key)
    })
    Object.getOwnPropertyNames(val).forEach((key) => {
      if (!Object.prototype.hasOwnProperty.call(props, key)) {
        result[key] = transform(val[key], additional, getProps, key)
      }
    })
    return result
  }

  if (typ === "any") return val
  if (typ === null) {
    if (val === null) return val
    return invalidValue(typ, val)
  }
  if (typ === false) return invalidValue(typ, val)
  while (typeof typ === "object" && typ.ref !== undefined) {
    typ = typeMap[typ.ref]
  }
  if (Array.isArray(typ)) return transformEnum(typ, val)
  if (typeof typ === "object") {
    return typ.hasOwnProperty("unionMembers")
      ? transformUnion(typ.unionMembers, val)
      : typ.hasOwnProperty("arrayItems")
        ? transformArray(typ.arrayItems, val)
        : typ.hasOwnProperty("props")
          ? transformObject(getProps(typ), typ.additional, val)
          : invalidValue(typ, val)
  }
  // Numbers can be parsed by Date but shouldn't be.
  if (typ === Date && typeof val !== "number") return transformDate(val)
  return transformPrimitive(typ, val)
}

function cast<T>(val: any, typ: any): T {
  return transform(val, typ, jsonToJSProps)
}

function uncast<T>(val: T, typ: any): any {
  return transform(val, typ, jsToJSONProps)
}

function a(typ: any) {
  return { arrayItems: typ }
}

function u(...typs: any[]) {
  return { unionMembers: typs }
}

function o(props: any[], additional: any) {
  return { props, additional }
}

// function m(additional: any) {
//     return { props: [], additional };
// }

function r(name: string) {
  return { ref: name }
}

const typeMap: any = {
  Questions: o(
    [{ json: "questions", js: "questions", typ: a(r("Question")) }],
    false,
  ),
  Question: o(
    [
      { json: "schemaVersion", js: "schemaVersion", typ: u(undefined, "") },
      { json: "@xsi:type", js: "@xsi:type", typ: r("QuestionXsiType") },
      { json: "uid", js: "uid", typ: "" },
      { json: "identifiant", js: "identifiant", typ: r("Identifiant") },
      { json: "type", js: "type", typ: r("Type") },
      { json: "indexationAN", js: "indexationAN", typ: r("IndexationAn") },
      { json: "auteur", js: "auteur", typ: r("Auteur") },
      { json: "minInt", js: "minInt", typ: r("Groupe") },
      { json: "minAttribs", js: "minAttribs", typ: r("MinAttribs") },
      {
        json: "textesQuestion",
        js: "textesQuestion",
        typ: u(r("TextesQuestion"), null, undefined),
      },
      {
        json: "textesReponse",
        js: "textesReponse",
        typ: u(r("TextesReponse"), null, undefined),
      },
      { json: "cloture", js: "cloture", typ: u(r("Cloture"), null, undefined) },
      {
        json: "signalement",
        js: "signalement",
        typ: u(r("Signalement"), null, undefined),
      },
      {
        json: "renouvellements",
        js: "renouvellements",
        typ: u(r("Renouvellements"), null, undefined),
      },
    ],
    false,
  ),
  Auteur: o(
    [
      { json: "identite", js: "identite", typ: r("Identite") },
      { json: "groupe", js: "groupe", typ: r("Groupe") },
    ],
    false,
  ),
  Groupe: o(
    [
      { json: "organeRef", js: "organeRef", typ: "" },
      { json: "abrege", js: "abrege", typ: "" },
      { json: "developpe", js: "developpe", typ: "" },
    ],
    false,
  ),
  Identite: o(
    [
      { json: "acteurRef", js: "acteurRef", typ: "" },
      { json: "mandatRef", js: "mandatRef", typ: "" },
    ],
    false,
  ),
  Cloture: o(
    [
      { json: "codeCloture", js: "codeCloture", typ: "" },
      {
        json: "libelleCloture",
        js: "libelleCloture",
        typ: "",
      },
      { json: "dateCloture", js: "dateCloture", typ: "" },
      { json: "infoJO", js: "infoJO", typ: u(r("InfoJo"), null) },
    ],
    false,
  ),
  Identifiant: o(
    [
      { json: "numero", js: "numero", typ: "" },
      { json: "regime", js: "regime", typ: "" },
      { json: "legislature", js: "legislature", typ: "" },
    ],
    false,
  ),
  IndexationAn: o(
    [
      { json: "rubrique", js: "rubrique", typ: u("", null) },
      { json: "teteAnalyse", js: "teteAnalyse", typ: null },
      { json: "analyses", js: "analyses", typ: r("Analyse") },
    ],
    false,
  ),
  Analyse: o([{ json: "analyse", js: "analyse", typ: u("", null) }], false),
  MinAttribs: o(
    [
      {
        json: "minAttrib",
        js: "minAttrib",
        typ: u(a(r("MinAttribElement")), r("MinAttribElement")),
      },
    ],
    false,
  ),
  MinAttribElement: o(
    [
      { json: "infoJO", js: "infoJO", typ: r("InfoJo") },
      { json: "denomination", js: "denomination", typ: r("Groupe") },
    ],
    false,
  ),
  InfoJo: o(
    [
      { json: "typeJO", js: "typeJO", typ: u(r("TypeJo"), null) },
      { json: "dateJO", js: "dateJO", typ: "" },
      { json: "pageJO", js: "pageJO", typ: u("", null) },
      { json: "numJO", js: "numJO", typ: u("", null) },
      { json: "urlLegifrance", js: "urlLegifrance", typ: u("", null) },
      { json: "referenceNOR", js: "referenceNOR", typ: u("", null) },
    ],
    false,
  ),
  TextesQuestion: o(
    [
      {
        json: "texteQuestion",
        js: "texteQuestion",
        typ: u(a(r("Texte")), r("Texte")),
      },
    ],
    false,
  ),
  Texte: o(
    [
      { json: "infoJO", js: "infoJO", typ: r("InfoJo") },
      { json: "texte", js: "texte", typ: u("", null) },
      { json: "erratum", js: "erratum", typ: u(undefined, "") },
    ],
    false,
  ),
  TextesReponse: o(
    [
      {
        json: "texteReponse",
        js: "texteReponse",
        typ: u(a(r("Texte")), r("Texte")),
      },
    ],
    false,
  ),
  Signalement: o(
    [
      { json: "infoJO", js: "infoJO", typ: r("InfoJo") },
      { json: "dateDepotSignal", js: "dateDepotSignal", typ: "" },
    ],
    false,
  ),
  Renouvellements: o(
    [
      {
        json: "renouvellement",
        js: "renouvellement",
        typ: u(a(r("Renouvellement")), r("Renouvellement")),
      },
    ],
    false,
  ),
  Renouvellement: o(
    [{ json: "infoJO", js: "infoJO", typ: r("InfoJo") }],
    false,
  ),
  QuestionXsiType: [
    "QuestionEcrite_Type",
    "QuestionGouvernement_Type",
    "QuestionOrale_Type",
  ],
  TypeJo: ["JO_QUESTION", "JO_LOI_DECRET", "JO_DEBAT"],
  Type: ["QE", "QG", "QOSD"],
}
