import type { Document, Division } from "./types/dossiers_legislatifs"
import type { Scrutin } from "./types/scrutins"

export enum DocumentUrlFormat {
  Html = "html",
  Pdf = "pdf",
  RawHtml = "raw-html", // The version of the document that can be parsed to extract divisions & articles
}

const documentUrlFragmentsByType: {
  [type: string]: {
    directory: string
    prefix: string
    suffix: string
  }
} = {
  ACIN: {
    directory: "projets",
    prefix: "pl",
    suffix: "-ai",
  },
  AVCEPRJLOI: {
    directory: "projets",
    prefix: "pl",
    suffix: "-ace",
  },
  AVCELETTRE: {
    directory: "projets",
    prefix: "pl",
    suffix: "-ace",
  },
  AVCEPIONLOI: {
    directory: "propositions",
    prefix: "pion",
    suffix: "-ace",
  },
  AVIS: {
    directory: "rapports",
    prefix: "r",
    suffix: "",
  },
  ETDI: {
    directory: "projets",
    prefix: "pl",
    suffix: "-ei",
  },
  LETT: {
    directory: "projets",
    prefix: "pl",
    suffix: "-l",
  },
  PION: {
    directory: "propositions",
    prefix: "pion",
    suffix: "",
  },
  PNRE: {
    directory: "propositions",
    prefix: "pion",
    suffix: "",
  },
  PNREAPPART341: {
    directory: "propositions",
    prefix: "pion",
    suffix: "",
  },
  PNRECOMENQ: {
    directory: "propositions",
    prefix: "pion",
    suffix: "",
  },
  PNREMODREGLTAN: {
    directory: "propositions",
    prefix: "pion",
    suffix: "",
  },
  PNRETVXINSTITEUROP: {
    directory: "europe/resolutions",
    prefix: "ppe",
    suffix: "",
  },
  PRJL: {
    directory: "projets",
    prefix: "pl",
    suffix: "",
  },
  RAPP: {
    directory: "rapports",
    prefix: "r",
    suffix: "",
  },
  RINF: {
    directory: "rap-info",
    prefix: "i",
    suffix: "",
  },
  RION: {
    directory: "",
    prefix: "",
    suffix: "",
  },
  TADO: {
    directory: "ta",
    prefix: "ta",
    suffix: "",
  },
  TCOM: {
    directory: "ta-commission",
    prefix: "r",
    suffix: "-a0",
  },
  TCOMCOMENQ: {
    directory: "ta-commission",
    prefix: "r",
    suffix: "-a0",
  },
  TCOMMODREGLTAN: {
    directory: "ta-commission",
    prefix: "r",
    suffix: "-a0",
  },
  TCOMTVXINSTITEUROP: {
    directory: "ta-commission",
    prefix: "r",
    suffix: "-a0",
  },
}

function extensionFromDocumentUrlFormat(format: DocumentUrlFormat): string {
  switch (format) {
    case DocumentUrlFormat.Html:
      return ".asp"
    case DocumentUrlFormat.Pdf:
      return ".pdf"
    case DocumentUrlFormat.RawHtml:
      return ".asp"
  }
}

export function* iterDocumentOrDivisionUrls(
  documentOrDivision: Document | Division,
  format: DocumentUrlFormat | undefined = undefined,
): Generator<
  { format: DocumentUrlFormat; type: string; url: string },
  void,
  unknown
> {
  // Raw HTML
  if (format === undefined || format === DocumentUrlFormat.RawHtml) {
    // Note: Both URLs should work and return the same result.
    yield {
      format: DocumentUrlFormat.RawHtml,
      type: "dyn-opendata",
      url: `https://www.assemblee-nationale.fr/dyn/opendata/${documentOrDivision.uid}.html`,
    }
    yield {
      format: DocumentUrlFormat.RawHtml,
      type: "dyn-docs",
      url: `https://www.assemblee-nationale.fr/dyn/docs/${documentOrDivision.uid}.raw`,
    }
    const urlInfos = urlInfosFromDocumentOrDivision(
      documentOrDivision,
      DocumentUrlFormat.RawHtml,
    )
    if (urlInfos !== null) {
      yield {
        format: DocumentUrlFormat.RawHtml,
        type: urlInfos.directory.replaceAll("/", "-"),
        url: urlInfos.url,
      }
    }
  }

  // HTML
  if (format === undefined || format === DocumentUrlFormat.Html) {
    const urlInfos = urlInfosFromDocumentOrDivision(
      documentOrDivision,
      DocumentUrlFormat.Html,
    )
    if (urlInfos !== null) {
      yield {
        format: DocumentUrlFormat.Html,
        type: urlInfos.directory.replaceAll("/", "-"),
        url: urlInfos.url,
      }
    }
  }

  // PDF
  if (format === undefined || format === DocumentUrlFormat.Pdf) {
    // Note: Both URLs should work and return the same result.
    yield {
      format: DocumentUrlFormat.RawHtml,
      type: "dyn-opendata",
      url: `https://www.assemblee-nationale.fr/dyn/opendata/${documentOrDivision.uid}.html`,
    }
    yield {
      format: DocumentUrlFormat.Pdf,
      type: "dyn-docs",
      url: `https://www.assemblee-nationale.fr/dyn/docs/${documentOrDivision.uid}.pdf`,
    }
    const urlInfos = urlInfosFromDocumentOrDivision(
      documentOrDivision,
      DocumentUrlFormat.Pdf,
    )
    if (urlInfos !== null) {
      yield {
        format: DocumentUrlFormat.Pdf,
        type: urlInfos.directory.replaceAll("/", "-"),
        url: urlInfos.url,
      }
    }
  }
}

/// Note: This function doesn't use the "modern" Assemblée URLs for documents.
/// Use generator `iterDocumentOrDivisionUrls` instead.
export function urlFromDocument(
  documentOrDivision: Document | Division,
  format: DocumentUrlFormat,
): string | null {
  const urlInfos = urlInfosFromDocumentOrDivision(documentOrDivision, format)
  return urlInfos === null ? null : urlInfos.url
}

export function urlFromScrutin(scrutin: Scrutin) {
  // Sample UIDs: VTANR5L15V389
  const match = /^VTANR([0-9]+)L([0-9]+)V([0-9]+)$/.exec(scrutin.uid)
  if (match === null) {
    console.log(
      `Unable to generate URL from scrutin with unexpected UID: "${scrutin.uid}"`,
    )
    return null
  }
  const legislature = match[2]
  const numero = match[3]
  return `http://www2.assemblee-nationale.fr/scrutins/detail/(legislature)/${legislature}/(num)/${numero}`
}

/// Note: This function doesn't use the modern Assemblée URLs for documents.
/// Use generator `iterDocumentOrDivisionUrls` instead.
function urlInfosFromDocumentOrDivision(
  documentOrDivision: Document | Division,
  format: DocumentUrlFormat,
): { directory: string; url: string } | null {
  // Code taken from function `an_text_url` in project anpy:
  // https://github.com/regardscitoyens/anpy/blob/master/anpy/dossier_from_opendata.py
  // See https://www.assemblee-nationale.fr/opendata/Implementation_Referentiel/Identifiants_Referentiels.html
  // for a description of the format of an UID.

  const extension = extensionFromDocumentUrlFormat(format)
  const match = /^(.{4})([ANS]*)(R[0-9])([LS]*)([0-9]*)([BTACP]*)(.*)$/.exec(
    documentOrDivision.uid,
  )
  if (match === null) {
    console.log(
      `Unable to generate URL from document or division with unexpected UID: "${documentOrDivision.uid}"`,
    )
    return null
  }
  const chambre = match[2]
  if (chambre === "SN") {
    console.log(
      `Unable to generate URL for document or division in Sénat. UID: "${documentOrDivision.uid}"`,
    )
    return null
  }
  const legislature = match[5]
  let number = match[7]
  if (format === DocumentUrlFormat.RawHtml) {
    return {
      directory: "textes",
      url: `https://www.assemblee-nationale.fr/${legislature}/textes/${number}${extension}`,
    }
  }
  let documentType =
    (
      {
        BTC: "TCOM",
        BTA: "TADO",
        TAP: "TADO",
      } as { [key: string]: string }
    )[match[6]] || match[1]
  if (
    documentOrDivision.classification !== undefined &&
    documentOrDivision.classification.sousType !== undefined &&
    documentOrDivision.classification.sousType.code === "COMPA"
  ) {
    // Annexe au rapport : Texte comparatif COMPA
    number = number.replace(/-COMPA/, "-aCOMPA")
  }
  if (documentType == "AVCE") {
    documentType = `AVCE${documentOrDivision.classification.famille?.classe.code}`
  }
  const fragments = documentUrlFragmentsByType[documentType]
  if (fragments === undefined) {
    // For example: ALCNANR5L15B0002 (allocution du président)
    console.log(
      `Unable to generate URL from document or division of UID "${documentOrDivision.uid}" and type "${documentType}"`,
    )
    return null
  }
  const directory =
    format === DocumentUrlFormat.Pdf
      ? `pdf/${fragments.directory}`
      : fragments.directory
  return {
    directory,
    url: `https://www.assemblee-nationale.fr/${legislature}/${directory}/${fragments.prefix}${number}${fragments.suffix}${extension}`,
  }
}
