import fs from "fs-extra"
import temp from "temp"
let sinon = require("sinon")
import { assert } from "chai"
import * as git from "../src/git"
import { write } from "../src/file_systems"

import {
  BugsHelper,
  checkoutFix,
  mergeFix,
  main,
} from "../src/scripts/bugs_helper"

// Automatically track and cleanup files at exit
temp.track()

suite("bugs_helper")

test("#fixInRepository", async function () {
  const tmpDir = temp.mkdirSync("check")
  git.run(tmpDir, "init")
  git.run(tmpDir, "config commit.gpgsign false")
  git.run(tmpDir, "config user.email 'tricoteuses@tricoteuses.fr'")
  git.run(tmpDir, "config user.name 'Tricoteuses données Assemblée'")
  const file = "RUANR5L15S2020IDS22089.json"
  fs.copySync(`tests/check/${file}`, `${tmpDir}/${file}`)
  git.run(tmpDir, "add .")
  git.run(tmpDir, "commit -m 'initial commit'")

  git.run(tmpDir, "checkout --quiet -b upstream/uid master")
  write("A", `${tmpDir}/upstreamfile`)
  git.run(tmpDir, "add .")
  git.run(tmpDir, "commit -m 'upstream commit'")

  git.run(tmpDir, "checkout --quiet -b bugfixes master")

  const h = new BugsHelper({
    //    debug: true,
    "bugs-dir": "tests/bugs_helper",
    bug: "00002",
    commit: tmpDir,
    schema: "agenda",
    data: `${tmpDir}/${file}`,
  })
  checkoutFix({ commit: tmpDir })
  await h.fix()
  mergeFix({ commit: tmpDir })
  assert.include(git.run(tmpDir, "log --oneline"), "Fix agenda")
  assert.include(git.run(tmpDir, "log --oneline"), "upstream commit")
  assert.notInclude(git.run(tmpDir, "branch"), "bugfixes")
  assert.notInclude(git.run(tmpDir, "branch"), "Revert")
  assert.include(git.run(tmpDir, "log --oneline -1"), "Merge bugfixes")
  const hash = git.run(tmpDir, "rev-parse HEAD")

  // nothing changed, do nothing
  checkoutFix({ commit: tmpDir })
  await h.fix()
  mergeFix({ commit: tmpDir })
  assert.equal(hash, git.run(tmpDir, "rev-parse HEAD"))

  // upstream changes, revert and re-apply fixes
  git.run(tmpDir, "checkout --quiet upstream/uid")
  write("B", `${tmpDir}/upstreamfile`)
  git.run(tmpDir, "add .")
  git.run(tmpDir, "commit -m 'another upstream commit'")
  git.run(tmpDir, "checkout --quiet master")

  checkoutFix({ commit: tmpDir })
  await h.fix()
  mergeFix({ commit: tmpDir })
  assert.notEqual(hash, git.run(tmpDir, "rev-parse HEAD"))
  assert.equal(
    'Merge bugfixes\nFix agenda-00002\nMerge upstream\nRevert "Merge bugfixes"',
    git.run(tmpDir, "log --date-order -4 --format=%s"),
  )
})

test("#checkWillBeFixed", async function () {
  const log = sinon.stub(console, "log")
  const warn = sinon.stub(console, "warn")
  const error = sinon.stub(console, "error")

  let status = await main([
    "--verbose",
    "--bugs-dir",
    "tests/bugs_helper",
    "--bug",
    "00002",
    "--check",
    "--schema=agenda",
    "--cr=tests/check",
    "--data=tests/check/RUANR5L15S2020IDS22089.json",
  ])
  assert.equal(status, 0)
  assert.deepEqual(warn.args, [])
  warn.reset()

  assert.deepEqual(error.args, [])
  error.reset()

  assert.include(log.args[2][0], "will-be-fixed")
  log.reset()

  log.restore()
  warn.restore()
  error.restore()
})

test("#checkPluginThrows", async function () {
  const log = sinon.stub(console, "log")
  const warn = sinon.stub(console, "warn")
  const error = sinon.stub(console, "error")

  let status
  // catch an exception if the plugin throws
  await main([
    "--verbose",
    "--bugs-dir=tests/bugs_helper",
    "--bug",
    "00001",
    "--check",
    "--dataset=tests/check/Agenda_XV",
  ])
    .then((_s) => {
      status = "exception not caught"
    })
    .catch((error) => {
      status = error
    })

  assert.deepEqual(status, "EXCEPTION")

  log.restore()
  warn.restore()
  error.restore()
})

test("#checkDataset", async function () {
  const log = sinon.stub(console, "log")
  const warn = sinon.stub(console, "warn")
  const error = sinon.stub(console, "error")

  let status = await main([
    "--verbose",
    "--bugs-dir=tests/bugs_helper",
    "--bug=00002",
    "--check",
    "--cr=tests/check",
    "--dataset=tests/check/Agenda_XV",
  ])

  assert.deepEqual(warn.args, [])
  warn.reset()

  assert.deepEqual(error.args, [])
  error.reset()

  assert.equal(status, 0)

  assert.include(log.args[2][0], "will-be-fixed")
  log.reset()

  // it's ok if there are no bugs plugins
  status = await main([
    "--verbose",
    "--check",
    "--bugs-dir=/unlikely",
    "--dataset=tests/check/Scrutins_XV",
  ])
  assert.equal(status, 0)

  log.restore()
  warn.restore()
  error.restore()
})

test("#wrongArgument", async function () {
  const log = sinon.stub(console, "log")
  const warn = sinon.stub(console, "warn")
  const error = sinon.stub(console, "error")

  for (const argument of ["schema", "dataset"]) {
    const status = await main([
      "--verbose",
      "--check",
      `--${argument}=INVALID`,
      "--cr=tests/check",
      "--data=tests/check/RUANR5L15S2020IDS22089.json",
    ])
    assert.equal(status, 1)
    assert.deepEqual(error.args, [[`--${argument} INVALID`]])
    error.reset()
  }

  log.restore()
  warn.restore()
  error.restore()
})

test("#main", async function () {
  const log = sinon.stub(console, "log")
  const warn = sinon.stub(console, "warn")
  const error = sinon.stub(console, "error")

  // good run
  let status = await main([
    "--verbose",
    "--check",
    "--bugs-dir=tests/bugs_helper",
    "--bug=00002",
    "--schema=agenda",
    "--cr=tests/check",
    "--data=tests/check/RUANR5L15S2020IDS22089.json",
  ])
  assert.equal(status, 0)
  assert.deepEqual(warn.args, [])
  warn.reset()

  // missing mandatory --data option
  assert.equal(await main([]), 1)

  log.restore()
  warn.restore()
  error.restore()
})
