let sinon = require("sinon")
import fs from "fs-extra"
import { assert } from "chai"
import * as git from "../src/git"
import temp from "temp"

// Automatically track and cleanup files at exit
temp.track()

suite("git")

test("#commit", async function () {
  const log = sinon.stub(console, "log")
  const error = sinon.stub(console, "error")

  const tmpDir = temp.mkdirSync("git")
  fs.writeFileSync(`${tmpDir}/A`, "AA")
  git.run(tmpDir, "init")
  git.run(tmpDir, "config commit.gpgsign false")
  git.run(tmpDir, "config user.email 'tricoteuses@tricoteuses.fr'")
  git.run(tmpDir, "config user.name 'Tricoteuses données Assemblée'")

  const message = "MESSAGE"
  assert.equal(git.commit(tmpDir, message), true)
  const gitLog = git.run(tmpDir, "log --oneline -1")
  assert.include(gitLog, message)
  assert.equal(git.commit(tmpDir, message), false)

  assert.throws(
    () => git.commit(tmpDir, 'A" --invalid-option "A'),
    "unknown option `invalid-option",
  )

  log.restore()
  error.restore()
})

test("#runGit", async function () {
  const tmpDir = temp.mkdirSync("git")
  git.run(tmpDir, "init")
  git.run(tmpDir, "config commit.gpgsign false")
  git.run(tmpDir, "config user.email 'tricoteuses@tricoteuses.fr'")
  git.run(tmpDir, "config user.name 'Tricoteuses données Assemblée'")

  assert(git.test(tmpDir, "rev-parse"))
  assert.equal(git.test(tmpDir, "rev-parse unknown"), false)

  const error = sinon.stub(console, "error")
  assert.throws(
    () => git.run(tmpDir, "rev-parse ; echo FAIL ; false"),
    "Command failed: git rev-parse ; echo FAIL ; false",
  )
  error.restore()
})

test("#uidBranchUpdateFlat", async function () {
  const tmpDir = temp.mkdirSync("git")
  git.run(tmpDir, "init")
  git.run(tmpDir, "config commit.gpgsign false")
  git.run(tmpDir, "config user.email 'tricoteuses@tricoteuses.fr'")
  git.run(tmpDir, "config user.name 'Tricoteuses données Assemblée'")

  fs.ensureDirSync(`${tmpDir}/files`)
  fs.writeFileSync(`${tmpDir}/files/a.json`, "aa")
  git.run(tmpDir, "add .")
  git.run(tmpDir, "commit -m initial")

  // a new branch is created with the uids
  // the working tree is left as is
  assert(fs.existsSync(`${tmpDir}/files/a.json`))
  assert.strictEqual(git.uidBranchUpdate(tmpDir, ["."], true), true)
  assert.equal(git.run(tmpDir, "rev-parse --abbrev-ref HEAD"), "master")
  assert(!fs.existsSync(`${tmpDir}/files/a.json`))
  assert(fs.existsSync(`${tmpDir}/uid/a.json`))
  assert(/ uid$/.test(git.run(tmpDir, "branch")))
  const uidHash = git.run(tmpDir, "rev-parse uid")

  // trying again re-uses the same local branch
  // the working tree is reset --hard
  assert.strictEqual(git.uidBranchUpdate(tmpDir, ["."], false), false)
  assert.equal(uidHash, git.run(tmpDir, "rev-parse uid"), "uid branch hash")
  assert(fs.existsSync(`${tmpDir}/files/a.json`))
  assert(!fs.existsSync(`${tmpDir}/uid/a.json`))

  git.run(tmpDir, "checkout --quiet uid")
  assert(!fs.existsSync(`${tmpDir}/files/a.json`), `${tmpDir}/files/a.json`)
  assert(fs.existsSync(`${tmpDir}/uid/a.json`))
  git.run(tmpDir, "checkout --quiet master")

  // trying again re-uses the same local branch
  // the working tree is left as is
  assert.strictEqual(git.uidBranchUpdate(tmpDir, ["."], true), false)
  assert.equal(git.run(tmpDir, "rev-parse --abbrev-ref HEAD"), "master")
  assert.equal(uidHash, git.run(tmpDir, "rev-parse uid"), "uid branch hash")
  assert(!fs.existsSync(`${tmpDir}/files/a.json`))
  assert(fs.existsSync(`${tmpDir}/uid/a.json`))

  // trying again with a remote branch and no local branch succeeds
  git.run(tmpDir, "reset --hard")
  git.run(tmpDir, "remote add origin http://somewhere.far.far.away")
  git.run(tmpDir, `update-ref refs/remotes/origin/uid ${uidHash}`)
  git.run(tmpDir, "branch -D uid")
  assert.strictEqual(git.uidBranchUpdate(tmpDir, ["."], false), false)
  assert.equal(git.run(tmpDir, "rev-parse --abbrev-ref HEAD"), "master")
})

test("#uidBranchUpdateSubdirectories", async function () {
  const tmpDir = temp.mkdirSync("git")
  git.run(tmpDir, "init")
  git.run(tmpDir, "config commit.gpgsign false")
  git.run(tmpDir, "config user.email 'tricoteuses@tricoteuses.fr'")
  git.run(tmpDir, "config user.name 'Tricoteuses données Assemblée'")

  const ds = ["A", "B"]
  for (const d of ds) {
    fs.ensureDirSync(`${tmpDir}/${d}/files`)
    fs.writeFileSync(`${tmpDir}/${d}/files/${d}.json`, "aa")
  }
  git.run(tmpDir, "add .")
  git.run(tmpDir, "commit -m initial")

  for (const d of ds) assert(fs.existsSync(`${tmpDir}/${d}/files/${d}.json`))
  assert.strictEqual(git.uidBranchUpdate(tmpDir, ds, true), true)
  for (const d of ds) {
    assert(!fs.existsSync(`${tmpDir}/${d}/files/${d}.json`))
    assert(fs.existsSync(`${tmpDir}/${d}/uid/${d}.json`))
  }
})
