import * as git from "../src/git"
import fs from "fs-extra"
import temp from "temp"
let rewire = require("rewire")
let sinon = require("sinon")
import { assert } from "chai"

// Automatically track and cleanup files at exit
temp.track()

let retrieve_documents = rewire("../src/scripts/retrieve_documents")

suite("retrieve_documents")

test("#urlFromUid", async function () {
  const Retrieve = retrieve_documents.__get__("Retrieve")
  const url = Retrieve.urlFromUid("RINFANR5L15B0385")
  assert.include(url, "uid/RINFANR5L15B0385")
  assert.include(url, "Dossiers_Legislatifs")
})

test("#needsDownload", async function () {
  const log = sinon.stub(console, "log")
  const Retrieve = retrieve_documents.__get__("Retrieve")
  let retrieve = new Retrieve({ verbose: true })

  retrieve.needsDownload(
    "tests/retrieve_documents/does-not-need-download.html",
    "UID",
  )
  assert.include(log.args[0][0], "UID: Cached OK")
  assert.include(log.args[0][0], "http://")
  log.reset()

  retrieve.needsDownload(
    "tests/retrieve_documents/does-not-need-download-but-has-no-url.html",
    "UID",
  )
  assert.include(log.args[0][0], "UID: Cached OK")
  assert.notInclude(log.args[0][0], "http://")
  assert.include(log.args[0][0], "unknown url")
  log.reset()

  retrieve.needsDownload("tests/retrieve_documents/needs-download.html", "UID")
  assert.include(log.args[0][0], "UID: Cached FAIL")
  log.reset()

  retrieve.needsDownload(" ENOENT ", "UID")
  assert.deepEqual(log.args, [])
  log.reset()

  log.restore()
})

test("#sanityCheckFormat", async function () {
  const Retrieve = retrieve_documents.__get__("Retrieve")
  let r
  r = Retrieve.sanityCheckFormat(Buffer.from("%PDF"), "pdf")
  assert.equal(r[1], "OK")
  r = Retrieve.sanityCheckFormat(Buffer.from("  <!DOCTYPE html>"), "html")
  assert.equal(r[1], "OK")
  r = Retrieve.sanityCheckFormat(Buffer.from("<?xml"), "html")
  assert.equal(r[1], "OK")

  let garbage
  garbage = Buffer.from("garbage")

  r = Retrieve.sanityCheckFormat(garbage, "pdf")
  assert(r[1].startsWith("bad pdf garb"), r[1])
  r = Retrieve.sanityCheckFormat(garbage, "html")
  assert(r[1].startsWith("bad html garb"), r[1])
})

test("#getUrls", async function () {
  const Retrieve = retrieve_documents.__get__("Retrieve")
  const retrieve = new Retrieve({})
  {
    const urls = retrieve.getUrls({ uid: "RINFANR5L15B0385" }, "pdf")
    assert.equal(urls.length, 2)
  }
  {
    const urls = retrieve.getUrls({ uid: "RINFANR5L15B0385" }, "html")
    assert.equal(urls.length, 3)
    assert.equal(
      urls[0],
      "http://www.assemblee-nationale.fr/dyn/docs/RINFANR5L15B0385.raw",
    )
  }
})

test("#retrieveDocuments", async function () {
  const log = sinon.stub(console, "log")
  const error = sinon.stub(console, "error")

  function fetch(url: string) {
    return {
      buffer: () => Buffer.from(`%PDF DOCTYPE html ${url}`),
      ok: true,
    }
  }
  retrieve_documents.__set__("fetch", fetch)
  const main = retrieve_documents.__get__("main")
  const tmpDir = temp.mkdirSync("retrieve")
  git.run(tmpDir, "init")

  let status = await main([
    "--verbose",
    `--textes=${tmpDir}`,
    "--data=tests/retrieve_documents/RINFANR5L15B0385.json",
  ])
  assert.equal(status, 0)

  const filename = `${tmpDir}/15/uid/RINFANR5L15B0385.html`
  const url = fs.readFileSync(filename)
  assert.include(
    url.toString(),
    "http://www.assemblee-nationale.fr/dyn/docs/RINFANR5L15B0385.raw",
  )
  assert(fs.existsSync(filename), filename)

  log.restore()
  error.restore()
})

test("#main", async function () {
  const main = retrieve_documents.__get__("main")
  const log = sinon.stub(console, "log")
  const error = sinon.stub(console, "error")

  const tmpDir = temp.mkdirSync("retrieve")
  git.run(tmpDir, "init")

  let status = await main([
    "--verbose",
    "--dry-run",
    `--textes=${tmpDir}`,
    "--data=tests/retrieve_documents/RINFANR5L15B0385.json",
  ])
  assert.deepEqual(error.args, [])
  assert.equal(status, 0)

  log.restore()
  error.restore()
})
