let rewire = require("rewire")
let sinon = require("sinon")
import { assert } from "chai"
import fs from "fs-extra"
import path from "path"
import * as git from "../src/git"
import { load } from "../src/file_systems"

let validate_json = rewire("../src/scripts/validate_json")

suite("validate_json")

test("#getSchemaDir", function () {
  const log = sinon.stub(console, "log")
  const error = sinon.stub(console, "error")

  const Validate = validate_json.__get__("Validate")
  const v = new Validate({
    verbose: true,
    repository: git.run(".", "rev-parse --show-toplevel"),
  })
  const tag = "schema-acteur-1.0"
  const tagDir = v.getSchemaDir(tag)
  assert(fs.existsSync(tagDir), tagDir)
  const acteurDir = path.join(tagDir, "acteur")
  assert(fs.existsSync(acteurDir), acteurDir)
  assert.equal(tagDir, v.getSchemaDir(tag))

  // assert.throws(
  //   () => v.getSchemaDir("FAKETAG"),
  //   "pathspec 'tags/FAKETAG' did not match",
  // )

  v.options.dev = true
  const devDir = v.getSchemaDir(tag)
  assert.equal(devDir, "src/schemas")

  log.restore()
  error.restore()
})

test("#getValidator", function () {
  const Validate = validate_json.__get__("Validate")
  const v = new Validate({
    schema: "document",
    repository: git.run(".", "rev-parse --show-toplevel"),
  })

  const validator = v.getValidator("document-1.0")
  assert.equal(validator, v.getValidator("document-1.0"))

  let valid = validator({
    schemaVersion: "document-1.0",
  })
  assert.equal(false, valid)
  assert.equal(validator.errors[0].params.missingProperty, "xsiType")

  const p = "tests/validate_json/RAPPANR5L15B1302.json"
  valid = validator(load(p))
  assert(valid, p)
})

test("#validate", function () {
  const Validate = validate_json.__get__("Validate")
  const v = new Validate({
    schema: "document",
    repository: git.run(".", "rev-parse --show-toplevel"),
  })

  const error = sinon.stub(console, "error")

  const status = v.validate({
    schemaVersion: "document-1.0",
  })
  assert.equal(status, 1)
  const errors = error.args[1][0]
  assert.equal(errors[0].params.missingProperty, "xsiType")

  error.restore()
})

test("#getVersion", function () {
  const Validate = validate_json.__get__("Validate")
  const v = new Validate({ schema: "acteur" })
  assert.equal(v.getVersion({}), "acteur-1.0")
  const version = "document-2.0"
  assert.equal(v.getVersion({ schemaVersion: version }), version)
})

test("#getSchemaURL", function () {
  const Validate = validate_json.__get__("Validate")
  const v = new Validate({ schema: "acteur" })
  const version = "document-2.0"

  v.options.repository = "http://site/repo.git"
  assert.include(v.getSchemaURL(version), `repo/tree/schema-${version}`)

  v.options.repository = "http://site/repo"
  assert.include(v.getSchemaURL(version), `repo/tree/schema-${version}`)

  v.options.repository = "." // current directory always exists
  assert.include(v.getSchemaURL(version), `at tag schema-${version}`)

  v.options.dev = true
  assert.equal(v.getSchemaURL(version), `src/schemas`)
})

test("#wrongArgument", async function () {
  const main = validate_json.__get__("main")
  const log = sinon.stub(console, "log")
  const warn = sinon.stub(console, "warn")
  const error = sinon.stub(console, "error")

  for (const argument of ["schema", "dataset"]) {
    const status = await main([
      "--verbose",
      `--${argument}=INVALID`,
      "--data=tests/check/RUANR5L15S2020IDS22089.json",
    ])
    assert.equal(status, 1)
    assert.deepEqual(error.args, [[`--${argument} INVALID`]])
    error.reset()
  }

  log.restore()
  warn.restore()
  error.restore()
})

test("#main", async function () {
  const main = validate_json.__get__("main")
  const log = sinon.stub(console, "log")
  const warn = sinon.stub(console, "warn")
  const error = sinon.stub(console, "error")

  let status = await main([
    "--verbose",
    "--repository",
    git.run(".", "rev-parse --show-toplevel"),
    "--schema=document",
    "--data=tests/validate_json/RAPPANR5L15B1302.json",
  ])
  assert.deepEqual(warn.args, [])
  warn.reset()

  assert.deepEqual(error.args, [])
  error.reset()

  assert.equal(status, 0)
  assert.deepEqual(warn.args, [])
  warn.reset()

  status = await main([])
  assert.equal(status, 1)
  assert.deepEqual(error.args, [["either --schema or --dataset is required"]])

  log.restore()
  warn.restore()
  error.restore()
})
